@@ -97,6 +97,9 @@ extern /* RETSIGTYPE */ int loginit __((int));
 /* nobody.c */
 extern int getnobody __((void));
 
+/* parseintv.c */
+extern unsigned long parse_interval __((char *str, char **retp));
+
 /* prversion.c */
 extern void prversion __((const char *prgname));
 
@@ -119,9 +122,15 @@ extern int  hdr_status __((const char *cp, const char *lbuf, int n, int octo));
 #ifdef Z_TOKEN_H
 extern token822 * scan822 __((const char **cpp, size_t n, int c1, int c2,
 				  int allowcomments, token822 **tlistp));
+extern token822 * scan822utext __((const char **cpp, size_t n,
+				   token822 **tlistp));
 #endif
 
 /* selfaddrs.c */
+
+struct sockaddr; /* a "forward" declaration */
+struct addrinfo; /* a "forward" declaration */
+
 extern void stashmyaddresses  __((const char *host));
 #ifdef SOCK_STREAM
 extern int  loadifaddresses   __((struct sockaddr ***));
@@ -131,6 +140,9 @@ extern int  matchmyaddress    __((struct sockaddr *));
 extern int  matchmyaddresses  __((struct addrinfo *));
 #endif
 
+/* zgetifaddress.c */
+extern int zgetifaddress __((int af, const char *ifname, struct sockaddr *));
+
 /* splay.c */
 /* .... much ... in  "splay.h" */
 
