@@ -34,7 +34,8 @@ struct taddress {
 #define _DSN_NOTIFY_DELAY	0x004
 #define _DSN_NOTIFY_NEVER	0x008
 
-#define _DSN__DIAGDELAYMODE	0x800 /* Internal magic */
+#define _DSN__DIAGDELAYMODE	 0x800 /* Internal magic for MAILBOX/SIEVE */
+#define _DSN__TEMPFAIL_NO_UNLOCK 0x400 /* Internal magic for SMTP */
 
 /* `convertmode' controls the behaviour of the message conversion:
      _CONVERT_NONE (0): send as is
@@ -153,6 +154,8 @@ extern void 	       notaryreport __((const char*, const char*, const char*, cons
 extern void            notary_setxdelay __((int));
 extern void            notary_setwtt __(( const char *host ));
 extern void            notary_setwttip __(( const char *ip ));
+extern void            notary_settaid __(( const char *name, int ));
+extern void            notary_setcvtmode __(( CONVERTMODE ));
 #if defined(HAVE_STDARG_H)
 extern void	       diagnostic __((struct rcpt *rp, int rc, int timeout, const char *fmt, ... ));
 #else
