@@ -138,7 +138,8 @@ eunlink(file)
 {
 	int	r;
 
-	r = unlink(file);
+	while ((r = unlink(file)) < 0 && (errno == EBUSY || errno == EINTR))
+	  ;
 	if (r < 0) {
 	  int serrno = errno;
 	  fprintf(stderr, "%s: unlink(%s): %s\n", progname,
@@ -241,26 +242,52 @@ erename(from, to)
 {
 
 #ifdef	HAVE_RENAME
-	if (rename(from, to) < 0) {
+	while (rename(from, to) < 0) {
 	  int serrno = errno;
+	  if (errno == EBUSY || errno == EINTR) {
+	    /* Solaris says EBUSY, we clean up.. */
+	    while (unlink(to) < 0) {
+	      if (errno == EBUSY || errno == EINTR)
+		continue; /* Crazy Solaris 2.x (including 2.6!) */
+	      /* Anything else is considered ok for the unlink(),
+		 things like  ENOENT, etc.. */
+	      break;
+	    }
+	    /* Solaris says EBUSY, we retry.. */
+	    continue;
+	  }
 	  fprintf(stderr, "%s: rename(%s,%s): %s\n", progname,
-		  from, to, strerror(errno));
+		  from, to, strerror(serrno));
 	  errno = serrno;
 	  return -1;
 	}
 
 #else	/* !HAVE_RENAME */
 	
-	if ((unlink(to) < 0 && errno != ENOENT) || (link(from, to) < 0)) {
+	while (unlink(to) < 0) {
 	  int serrno = errno;
+	  if (errno == EINTR || errno == EBUSY)
+	    continue; /* Solaris 2.x previous to 2.7 may yield EBUSY.. */
+	  if (errno == ENOENT) break; /* Ok! */
+	  fprintf(stderr, "%s: rename(%s,%s): %s\n", progname,
+		  from, to, strerror(serrno));
+	  errno = serrno;
+	  return -1;
+	}
+	while (link(from, to) < 0) {
+	  int serrno = errno;
+	  if (errno == EINTR || errno == EBUSY)
+	    continue; /* Solaris may yield this */
 	  fprintf(stderr, "%s: rename(%s,%s): %s\n", progname,
 		  from, to, strerror(errno));
 	  errno = serrno;
 	  return -1;
 	}
 
-	if (unlink(from) < 0) {
+	while (unlink(from) < 0) {
 	  int serrno = errno;
+	  if (errno == EINTR || errno == EBUSY)
+	    continue; /* Solaris ... */
 	  if (errno != ENOENT)
 	    fprintf(stderr, "%s: rename(%s,%s): %s\n", progname,
 		    from, to, strerror(errno));
@@ -277,9 +304,21 @@ int
 eqrename(from, to)
 	const char *from, *to;
 {
-#ifdef	HAVE_RENAME_we_use_link_unlink_to_avoid_races_in_routers
-
-	if (rename(from, to) < 0) {
+#ifdef	HAVE_RENAME
+	while (rename(from, to) < 0) {
+	  int serrno = errno;
+	  if (errno == EBUSY || errno == EINTR) {
+	    /* Solaris says EBUSY, we clean up.. */
+	    while (unlink(to) < 0) {
+	      if (errno == EBUSY || errno == EINTR)
+		continue; /* Crazy Solaris 2.x (including 2.6!) */
+	      /* Actually Solaris reports only EBUSY, but .. */
+	      break;
+	    }
+	    /* Solaris says EBUSY, we retry.. */
+	    continue;
+	  }
+	  errno = serrno;
 	  return -1;
 	}
 
