@@ -27,7 +27,7 @@
     by Matti Aarnio <mea@nic.funet.fi> 1997
 */
 
-#include "mailer.h"
+#include "hostenv.h"
 #include <sys/types.h>
 #ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
@@ -67,6 +67,8 @@ union sockaddr_uni {
 #endif
 };
 
+#include "libc.h"
+#include "zmalloc.h"
 #include "libz.h"
 
 extern char  *getzenv     __((const char *));
@@ -94,173 +96,364 @@ struct mbuf;
 #include <arpa/inet.h>
 #include <net/if.h>
 
+/* #include "l-if.h" --- just some fake test stuff for SIOCGLIF*** */
+
+#ifdef HAVE_IFADDRS_H
+#include <ifaddrs.h>
+#endif
 
 int
 loadifaddresses(sockaddrp)
 struct sockaddr ***sockaddrp;
 {
-#ifdef SIOCGIFCONF
-	int s;
 	int i;
-	int ifcount = 0;
-	int pf = PF_INET;
-        struct ifconf ifc;
-	int ifbufsize = 4 * sizeof(struct ifreq) + 4;
-	char *interfacebuf = (void*)malloc(ifbufsize);
+	int sapcount = -1;
+	int sapspace = 2;
 	union sockaddr_uni **sap;
 
-	sap = (void*)malloc(sizeof(union sockaddr_uni*) * (ifcount + 2));
-	if (! sap || !interfacebuf) {
-	  if (interfacebuf)
-	    free(interfacebuf);
-	  if (sap)
-	    free(sap);
+	sap = (void*)malloc(sizeof(union sockaddr_uni*) * (sapspace + 2));
+
+	if (! sap)
 	  return -3; /* UAARGH! */
-	}
+
+#ifdef HAVE_GETIFADDRS
+	{
+	  struct ifaddrs *ifar = NULL, *ifa;
+
+	  i = getifaddrs( &ifar );
+	  if (i < 0) {
+	    free(sap);
+	    return i;
+	  }
+
+	  for (i = 0, ifa = ifar; ifa ; ++i, ifa = ifa->ifa_next) {
+	    if ((ifa->ifa_flags & IFF_UP) &&
+		(ifa->ifa_addr != NULL)) {
+
+	      struct sockaddr *sa = ifa->ifa_addr;
+
+	      if (sapcount +2 >= sapspace) {
+		sapspace <<= 1;
+		sap = (void*)realloc(sap, (sizeof(union sockaddr_uni*) *
+					   (sapspace + 2)));
+	      }
+
+	      if (! sap) {
+		return -5; /* UAARGH! */
+	      }
+
+	      if (sa->sa.sa_family == AF_INET) {
+		struct sockaddr_in *si4 = (void*)malloc(sizeof(*si4));
+		if (si4 == NULL)
+		  break;
+
+		/* pick the whole sockaddr package! */
+		memcpy(si4, sa, sizeof(*si4));
+		sap[++sapcount] = (union sockaddr_uni *)si4;
+	      }
 
 #if defined(AF_INET6) && defined(INET6)
-other_socktype:
+	      if (sa->sa.sa_family == AF_INET6) {
+		struct sockaddr_in6 *si6 = (void*)malloc(sizeof(*si6));
+		if (si6 == NULL)
+		  break;
+
+		/* pick the whole sockaddr package! */
+		memcpy(si6, sa, sizeof(*si6));
+		sap[++sapcount] = (union sockaddr_uni *)si6;
+	      }
+#endif
+	    }
+	  }
+
+
+#ifdef HAVE_FREEIFADDRS
+	  freeifaddrs(ifar);
+#else
+	  free(ifar);
+#endif
+	}
+#else /* not HAVE_GETIFADDRS */
+
+#if defined(SIOCGLIFCONF) && defined(AF_INET6) && defined(INET6)
+	/*#warning "experimental INET6 related SIOCGLIFCONF code activated!"*/
+	{
+	  struct lifconf lifc;
+	  struct ifconf ifc;
+	  int ifbufsize = 4 * sizeof(struct lifreq) + 4;
+	  char *interfacebuf = (void*)malloc(ifbufsize);
+	  int s;
+
+	  if (!interfacebuf) {
+	    free(sap);
+	    return -2;
+	  }
+
+	  s = socket(PF_INET6, SOCK_DGRAM, 0);
+	  if (s < 0)
+	    goto done_this_ipv6;
+
+
+	  /* Some utilities seem to do this probing also with sockets of
+	     AF_X25, AF_IPX, AF_AX25, AF_INET6, etc. -address families,
+	     but on Linux (very least) it can be done with any of them,
+	     thus we use the one that is most likely available: AF_INET */
+
+	  for (;;) {
+
+	    /* get the list of known IP address from the kernel */
+	    ifbufsize <<= 1;
+	    interfacebuf = (void*)realloc(interfacebuf,ifbufsize);
+	    memset(&lifc, 0, sizeof(lifc));
+	    lifc.lifc_buf    = interfacebuf;
+	    lifc.lifc_len    = ifbufsize;
+	    lifc.lifc_family = AF_UNSPEC;
+	    lifc.lifc_flags  = 0;
+	    if (ioctl(s, SIOCGLIFCONF, (char *)&lifc) < 0) {
+	      if (errno == EINTR)
+		continue;
+	      if (errno == EINVAL)
+		continue;
+	      goto done_this_ipv6; /* HUH!! ??? */
+	    }
+
+	    if (lifc.lifc_len < (ifbufsize - 2*sizeof(struct lifreq)))
+	      break;
+
+	    /* Redo the query, perhaps didn't get them all.. */
+	  }
+
+	  /* Likely got them all.. */
+
+	  /* Count how many addresses listed */
+
+	  for (i = 0; i < lifc.lifc_len; ) {
+
+	    struct lifreq *lifr = (struct lifreq *) &lifc.lifc_buf[i];
+	    union sockaddr_uni *sa = (union sockaddr_uni *) &lifr->lifr_addr;
+#ifdef SIOCGLIFFLAGS
+	    struct lifreq lifrf;
 #endif
 
-	s = socket(pf, SOCK_DGRAM, 0);
-	if (s < 0) {
+#if defined(SA_LEN)
+	    if (SA_LEN(((struct sockaddr *)sa)) > sizeof(lifr->lifr_addr))
+	      i += sizeof lifr->lifr_name + SA_LEN(((struct sockaddr *)sa));
+	    else
+#elif defined(HAVE_SA_LEN)
+	      if (sa->sa.sa_len > sizeof lifr->lifr_addr)
+		i += sizeof lifr->lifr_name + sa->sa.sa_len;
+	      else
+#endif
+		i += sizeof *lifr;
+
+	    /* Known address families ?
+	       The one we scanned for ??*/
+
+	    if ( sa->sa.sa_family != AF_INET &&
+		 sa->sa.sa_family != AF_INET6 )
+		/* Skip if not either of used protocols.. */
+		continue;
+
+	    /* Now, what do the flags say ? Are they alive ? */
+
+#ifdef SIOCGLIFFLAGS
+
+	    memset(&lifrf, 0, sizeof(struct lifreq));
+	    strncpy(lifrf.lifr_name, lifr->lifr_name, sizeof(lifrf.lifr_name));
+
+	    if (ioctl(s, SIOCGLIFFLAGS, (char *) &lifrf) < 0)
+	      continue; /* Failed.. */
+
+#if 0
+	    printf("name='%s'  lifrf_flags=0x%x\n",
+		   lifr->lifr_name,lifrf.lifr_flags);
+#endif
+	    if (!(IFF_UP & lifrf.lifr_flags))
+	      continue;
+#else
+
+	    /* printf("lifr_flags=0x%x\n",lifr->lifr_flags); */
+
+	    if (!(IFF_UP & lifr->lifr_flags))
+	      continue;
+#endif
+
+	    if (sapcount +2 >= sapspace) {
+	      sapspace <<= 1;
+	      sap = (void*)realloc(sap, (sizeof(union sockaddr_uni*) *
+					 (sapspace + 2)));
+	    }
+
+	    if (! sap) {
+	      close(s);
+	      free(interfacebuf);
+	      return -5; /* UAARGH! */
+	    }
+
+	    if (sa->sa.sa_family == AF_INET) {
+	      struct sockaddr_in *si4 = (void*)malloc(sizeof(*si4));
+	      if (si4 == NULL)
+		break;
+
+	      /* pick the whole sockaddr package! */
+	      memcpy(si4, sa, sizeof(*si4));
+	      sap[++sapcount] = (union sockaddr_uni *)si4;
+	    } else if (sa->sa.sa_family == AF_INET6) {
+	      struct sockaddr_in6 *si6 = (void*)malloc(sizeof(*si6));
+	      if (si6 == NULL)
+		break;
+
+	      /* pick the whole sockaddr package! */
+	      memcpy(si6, sa, sizeof(*si6));
+	      sap[++sapcount] = (union sockaddr_uni *)si6;
+	    }
+	  }
+
+done_this_ipv6:
+
+	  if (s >= 0) close(s);
 	  free(interfacebuf);
-	  free(sap);
-	  return -1;
+
 	}
+#else /* SIOCGLIFCONF && IPv6 */
 
-	/* Redo the buffer size increase until we get response size to
-	   be something of LESS THAN the buffer size minus two-times
-	   the sizeof(struct ifreq) -- because then we don't have
-	   a potential case of having larger block of addresses in
-	   system, but us being unable to get them all..
-	   Usually system has TWO interfaces -- loopback, and the LAN,
-	   thus the following loop is executed exactly once! */
+#ifdef SIOCGIFCONF
+	{
+	  struct ifconf ifc;
+	  int ifbufsize = 4 * sizeof(struct ifreq) + 4;
+	  char *interfacebuf = (void*)malloc(ifbufsize);
+	  int s;
 
-	/* Some utilities seem to do this probing also with sockets of
-	   AF_X25, AF_IPX, AF_AX25, AF_INET6, etc. -address families,
-	   but on Linux (very least) it can be done with any of them,
-	   thus we use the one that is most likely available: AF_INET */
+	  if (!interfacebuf) {
+	    free(sap);
+	    return -2;
+	  }
 
-	for (;;) {
+	  s = socket(PF_INET, SOCK_DGRAM, 0);
 
-	  /* get the list of known IP address from the kernel */
-	  ifbufsize <<= 1;
-	  interfacebuf = (void*)realloc(interfacebuf,ifbufsize);
-	  memset(&ifc, 0, sizeof(ifc));
-	  ifc.ifc_buf = interfacebuf;
-	  ifc.ifc_len = ifbufsize;
-	  if (ioctl(s, SIOCGIFCONF, (char *)&ifc) < 0)
-	    if (errno == EINVAL)
+	  if (s < 0) {
+	    free(interfacebuf);
+	    free(sap);
+	    return -1;
+	  }
+
+	  /* Redo the buffer size increase until we get response size to
+	     be something of LESS THAN the buffer size minus two-times
+	     the sizeof(struct ifreq) -- because then we don't have
+	     a potential case of having larger block of addresses in
+	     system, but us being unable to get them all..
+	     Usually system has TWO interfaces -- loopback, and the LAN,
+	     thus the following loop is executed exactly once! */
+
+	  /* Some utilities seem to do this probing also with sockets of
+	     AF_X25, AF_IPX, AF_AX25, AF_INET6, etc. -address families,
+	     but on Linux (very least) it can be done with any of them,
+	     thus we use the one that is most likely available: AF_INET */
+
+	  for (;;) {
+
+	    /* get the list of known IP address from the kernel */
+	    ifbufsize <<= 1;
+	    interfacebuf = (void*)realloc(interfacebuf,ifbufsize);
+	    memset(&ifc, 0, sizeof(ifc));
+	    ifc.ifc_buf = interfacebuf;
+	    ifc.ifc_len = ifbufsize;
+	    if (ioctl(s, SIOCGIFCONF, (char *)&ifc) < 0)
+	      if (errno == EINVAL)
+		continue;
+
+	    if (ifc.ifc_len < (ifbufsize - 2*sizeof(struct ifreq)))
+	      break;
+
+	    /* Redo the query, perhaps didn't get them all.. */
+	  }
+
+
+	  /* Count how many addresses listed */
+
+	  for (i = 0; i < ifc.ifc_len; ) {
+
+	    struct ifreq *ifr = (struct ifreq *) &ifc.ifc_buf[i];
+	    union sockaddr_uni *sa = (union sockaddr_uni *) &ifr->ifr_addr;
+#ifdef SIOCGIFFLAGS
+	    struct ifreq ifrf;
+#endif
+
+#if defined(SA_LEN)
+	    if (SA_LEN(((struct sockaddr*)sa)) > sizeof ifr->ifr_addr)
+	      i += sizeof ifr->ifr_name + SA_LEN(((struct sockaddr *)sa));
+	    else
+#elif defined(HAVE_SA_LEN)
+	      if (sa->sa.sa_len > sizeof ifr->ifr_addr)
+		i += sizeof ifr->ifr_name + sa->sa.sa_len;
+	      else
+#endif
+		i += sizeof *ifr;
+
+	    /* Known address families ?
+	       The one we scanned for ??*/
+
+	    if (ifr->ifr_addr.sa_family != PF_INET)
 	      continue;
 
-	  if (ifc.ifc_len < (ifbufsize - 2*sizeof(struct ifreq)))
-	    break;
-
-	  /* Redo the query, perhaps didn't get them all.. */
-	}
-
-
-	/* Count how many addresses listed */
-
-	for (i = 0; i < ifc.ifc_len; ) {
-
-	  struct ifreq *ifr = (struct ifreq *) &ifc.ifc_buf[i];
-	  union sockaddr_uni *sa = (union sockaddr_uni *) &ifr->ifr_addr;
-#ifdef SIOCGIFFLAGS
-	  struct ifreq ifrf;
-#endif
-
-#ifdef HAVE_SA_LEN
-	  if (sa->sa.sa_len > sizeof ifr->ifr_addr)
-	    i += sizeof ifr->ifr_name + sa->sa.sa_len;
-	  else
-#endif
-	    i += sizeof *ifr;
-
-	  /* Known address families ? */
-
-	  if (ifr->ifr_addr.sa_family != AF_INET
-#if defined(AF_INET6) && defined(INET6)
-	      &&
-	      ifr->ifr_addr.sa_family != AF_INET6
-#endif
-	      )
-	    continue; /* Not IPv4, nor IPv6 */
-
-
-	  /* Now, what do the flags say ? Are they alive ? */
+	    /* Now, what do the flags say ? Are they alive ? */
 
 #ifdef SIOCGIFFLAGS
 
-	  memset(&ifrf, 0, sizeof(struct ifreq));
-	  strncpy(ifrf.ifr_name, ifr->ifr_name, sizeof(ifrf.ifr_name));
+	    memset(&ifrf, 0, sizeof(struct ifreq));
+	    strncpy(ifrf.ifr_name, ifr->ifr_name, sizeof(ifrf.ifr_name));
+
+	    if (ioctl(s, SIOCGIFFLAGS, (char *) &ifrf) < 0)
+	      continue; /* Failed.. */
 
-	  if (ioctl(s, SIOCGIFFLAGS, (char *) &ifrf) < 0)
-	    continue; /* Failed.. */
 #if 0
-	   printf("name='%s'  ifrf_flags=0x%x\n",
-	     ifr->ifr_name,ifrf.ifr_flags);
+	    printf("name='%s'  ifrf_flags=0x%x\n",
+		   ifr->ifr_name,ifrf.ifr_flags);
 #endif
-	  if (!(IFF_UP & ifrf.ifr_flags))
-	    continue;
+	    if (!(IFF_UP & ifrf.ifr_flags))
+	      continue;
 #else
 
-	  /* printf("ifr_flags=0x%x\n",ifr->ifr_flags); */
+	    /* printf("ifr_flags=0x%x\n",ifr->ifr_flags); */
 
-	  if (!(IFF_UP & ifr->ifr_flags))
-	    continue;
+	    if (!(IFF_UP & ifr->ifr_flags))
+	      continue;
 #endif
 
-	  sap = (void*)realloc(sap, sizeof(union sockaddr_uni*) * (ifcount + 2));
-	  if (! sap) {
-	    close(s);
-	    free(interfacebuf);
-	    return -4; /* UAARGH! */
-	  }
+	    if (sapcount +2 >= sapspace) {
+	      sapspace <<= 1;
+	      sap = (void*)realloc(sap, (sizeof(union sockaddr_uni*) *
+					 (sapspace + 2)));
+	    }
+	    if (! sap) {
+	      close(s);
+	      free(interfacebuf);
+	      return -4; /* UAARGH! */
+	    }
 
-	  /* XX: Use sa_len on BSD44 ??? */
-	  if (sa->sa.sa_family == AF_INET) {
-	    struct sockaddr_in *si4 = (void*)malloc(sizeof(*si4));
-	    if (si4 == NULL)
-	      break;
-	    /* pick the whole sockaddr package! */
-	    memcpy(si4, &ifr->ifr_addr, sizeof(struct sockaddr_in));
-	    sap[ifcount] = (union sockaddr_uni *)si4;
+	    if (sa->sa.sa_family == AF_INET) {
+	      struct sockaddr_in *si4 = (void*)malloc(sizeof(*si4));
+	      if (si4 == NULL)
+		break;
+	      /* pick the whole sockaddr package! */
+	      memcpy(si4, &ifr->ifr_addr, sizeof(*si4));
+	      sap[++sapcount] = (union sockaddr_uni *)si4;
+	    }
 	  }
-#if defined(AF_INET6) && defined(INET6)
-	  if (sa->sa.sa_family == AF_INET6) {
-	    struct sockaddr_in6 *si6 = (void*)malloc(sizeof(*si6));
-	    if (si6 == NULL)
-	      break;
-
-	    /* XX: MUST DO AN  ioctl(s, SIOCGIFADDR, ...) HERE ! */
-
-	    /* pick the whole sockaddr package! */
-	    memcpy(si6, &ifr->ifr_addr, sizeof(struct sockaddr_in6));
-	    sap[ifcount] = (union sockaddr_uni *)si6;
-	  }
-#endif
-	  ++ifcount;
+	  close(s);
+	  free(interfacebuf);
 	}
-	sap[ifcount] = NULL;
-	close(s);
+#endif /* defined(SIOCGIFCONF) */
+#endif /* SIOCGLIFCONF / SIOCGIFCONF */
 
-#if defined(AF_INET6) && defined(INET6)
-	if (pf != PF_INET6) {
-	  pf = PF_INET6;
-	  goto other_socktype;
+#endif /* not HAVE_GETIFADDRS */
+
+	  *sockaddrp = (struct sockaddr **)sap;
+
+	  sap[++sapcount] = NULL;
+
+	  return sapcount;
 	}
-#endif
-
-	free(interfacebuf);
-	*sockaddrp = (struct sockaddr **)sap;
-
-	return ifcount;
-#else
-	return -1;
-#endif
-}
 
 #ifndef TESTMODE /* We test ONLY of  loadifaddresses() routine! */
 
@@ -406,10 +599,11 @@ const char *host;
 	  myaddrs = sa;
 	  nmyaddrs = sacnt;
 
-	  return;
 	}
 
-	/* Didn't get any by probeing interfaces ?! Lets use environment .. */
+	/* Independent of getting (or not) of interface data, stash in
+	   also whatever we can get from our environment.  Thus was
+	   we can block CLUSTER input addresses, for example. */
 
 	zenv = getzenv("SELFADDRESSES");
 
@@ -451,7 +645,7 @@ matchmyaddress(_sa)
 	  net = (ntohl(sau->v4.sin_addr.s_addr) >> 24) & 0xFF;
 	  if (net == 127)
 	    return 2; /* Loopback network */
-	  if (net == 0 || net == 127 || net > 223)
+	  if (net == 0 || net > 223)
 	    return 3;
 	}
 
@@ -488,7 +682,7 @@ matchmyaddresses(ai)
 
 #ifdef TESTMODE
 
-char *progname = "selfaddrstest";
+const char *progname = "selfaddrstest";
 
 int main(argc,argv)
 int argc;
