@@ -22,7 +22,9 @@
 #include <sys/time.h>
 #include <sys/file.h>
 #include <fcntl.h> /*XXX*/
+#ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
+#endif
 #include <netinet/in.h>
 #ifdef HAVE_NETINET_IN6_H
 # include <netinet/in6.h>
@@ -43,6 +45,9 @@ extern int errno;
 
 #ifdef _AIX
 #include <sys/types.h>
+#endif
+
+#ifdef HAVE_SYS_SELECT_H
 #include <sys/select.h>
 #endif
 
@@ -158,7 +163,6 @@ static volatile const char *ident_sockuser2(s,local,remote,realbuf,realbuflen)
   register int buflen;
   register int w;
   register int saveerrno;
-  char ch;
   int rlocal;
   int rremote;
   register int fl;
@@ -192,7 +196,7 @@ static volatile const char *ident_sockuser2(s,local,remote,realbuf,realbuflen)
       saveerrno = errno;
       close(s);
       if (errno == ECONNREFUSED || errno == EPIPE)
-	return "NO-IDENT-SERVICE";
+	return "NO-IDENT-SERVICE[2]";
       else {
 	sprintf(realbuf,"SOCKWRITE-%d", errno);
 	return realbuf;
@@ -253,8 +257,8 @@ static void sig_alrm __((int));
 static void sig_alrm (sig)
 	int sig;
 {
-	SIGNAL_HANDLE(SIGALRM, sig_alrm);
-	SIGNAL_RELEASE(SIGALRM);
+	SIGNAL_RELEASE(sig);
+	SIGNAL_HANDLE(sig, sig_alrm);
 	longjmp(jmpalarm,1);
 }
 
@@ -274,46 +278,54 @@ volatile const char *ident_tcpuser9(af,len,inlocal,inremote,local,remote,timeout
   int saveerrno;
   void (*old_sig)__((int));
   void (*old_alrm)__((int));
-  unsigned int oldival;
+  volatile unsigned int oldival = 0;
   volatile const char *retval;
   
   SIGNAL_HANDLESAVE(SIGPIPE, SIG_IGN, old_sig);
-  oldival  = alarm(10); /* 10 seconds until... */
   SIGNAL_HANDLESAVE(SIGALRM, sig_alrm, old_alrm);
   s = -1;
 
   if (setjmp(jmpalarm) == 0) {
+    oldival  = alarm(timeout+5);
     r = ident_tcpsock5(af, len, inlocal, inremote, &s);
+
+    if (r < 0) {
+      SIGNAL_HANDLE(SIGPIPE, old_sig);
+      SIGNAL_HANDLE(SIGALRM, old_alrm);
+      alarm(oldival);
+      if (errno == ECONNREFUSED)
+	return "NO-IDENT-SERVICE";
+      return "SOCKFAULT1";
+    }
+    tv.tv_sec = timeout;
+    tv.tv_usec = 0;
+    FD_ZERO(&wfds);
+    FD_SET(s,&wfds);
+    r = select(s + 1, NULL, &wfds, NULL,&tv);
+    /* XXX: how to handle EINTR? */
+    if (r == -1) {
+      SIGNAL_HANDLE(SIGPIPE, old_sig);
+      SIGNAL_HANDLE(SIGALRM, old_alrm);
+      alarm(oldival);
+      CLORETS("SOCKSELERR");
+    }
+    if (!FD_ISSET(s,&wfds)) {
+      close(s);
+      SIGNAL_HANDLE(SIGALRM, old_alrm);
+      SIGNAL_HANDLE(SIGPIPE, old_sig);
+      alarm(oldival);
+      errno = ETIMEDOUT;
+      return "TIMEDOUT";
+    }
+    retval = ident_sockuser2(s,local,remote,buf,buflen);
   } else {
     /* We reach here in case of alarm timer chime... */
-    if (s >= 0) close(s);
-    r = -1;
-  }
+    retval = "TIMEDOUT2";
+  }  
+  if (s >= 0) close(s);
   SIGNAL_HANDLE(SIGALRM, old_alrm);
-  alarm(oldival);
-  if (r < 0) {
-    SIGNAL_HANDLE(SIGPIPE, old_sig);
-    return "SOCKFAULT1";
-  }
-  tv.tv_sec = timeout;
-  tv.tv_usec = 0;
-  FD_ZERO(&wfds);
-  FD_SET(s,&wfds);
-  r = select(s + 1, NULL, &wfds, NULL,&tv);
-  /* XXX: how to handle EINTR? */
-  if (r == -1) {
-    SIGNAL_HANDLE(SIGPIPE, old_sig);
-    CLORETS("SOCKSELERR");
-  }
-  if (!FD_ISSET(s,&wfds)) {
-    close(s);
-    errno = ETIMEDOUT;
-    SIGNAL_HANDLE(SIGPIPE, old_sig);
-    return "TIMEDOUT";
-  }
-  retval = ident_sockuser2(s,local,remote,buf,buflen);
   SIGNAL_HANDLE(SIGPIPE, old_sig);
-  
+  alarm(oldival);
   return retval;
 }
 
@@ -566,7 +578,7 @@ register struct in_addr *inremote;
   register int fl;
   register int saveerrno;
 
-  if ((s = socket(AF_INET,SOCK_STREAM,0)) == -1)
+  if ((s = socket(PF_INET,SOCK_STREAM,0)) == -1)
     return -1;
   if (inlocal->s_addr) {
     memset(&sa,0,sizeof(sa));
