@@ -75,8 +75,8 @@ static int run_basename  ARGCV;
 static int run_recase    ARGCV;
 static int run_squirrel  ARGCV;
 static int run_822syntax ARGCV;
-static int run_dequote   ARGCV;
 static int run_condquote ARGCV;
+static int run_dequote   ARGCV;
 static int run_syslog    ARGCV;
 
 #if	defined(XMEM) && defined(CSRIMALLOC)
@@ -232,7 +232,7 @@ run_trace(argc, argv)
 
 
 int gensym;
-const char *gs_name = "g%d";
+const char * const gs_name = "g%d";
 
 static int
 run_gensym(argc, argv)
@@ -806,6 +806,8 @@ run_listexpand(avl, il)
 		 * For sendmail compatibility, addresses may not cross line
 		 * boundaries, and line boundary is just as good an address
 		 * separator as comma is, sigh.
+		 * Also lines beginning with '#' are comments, and blank
+		 * lines are (sortof) comments too.
 		 */
 		if (zlinebuf[n-1] == '\n')
 			--n;
@@ -1244,6 +1246,8 @@ run_listaddrs(argc, argv)
 		 * For sendmail compatibility, addresses may not cross line
 		 * boundaries, and line boundary is just as good an address
 		 * separator as comma is, sigh.
+		 * Also lines beginning with '#' are comments, and blank
+		 * lines are (sortof) comments too.
 		 */
 		if (zlinebuf[n-1] == '\n')
 			--n;
@@ -1888,99 +1892,125 @@ run_822syntax(argc, argv)
 
 
 static int
-run_dequote(argc, argv)
-	int argc;
+run_condquote_(argc, argv, condq)
+	int argc, condq;
 	const char *argv[];
 {
-	int len;
-	const char *s = argv[1];
+	const char *s;
+	int mustquote = 0;
+	int c, quoted;
+	int spc = 0;
+	int errflg = 0;
+	const char *appstr = NULL;
 
-	if (argc != 2)
-	  return 2; /* Bad bad! Missing/extra arg! */
+	extern int rfc822_mustquote __((const char *, const int));
 
-	len = strlen(s);
+	/* We remove quotes when they are not needed, and add them when
+	   they really are needed! */
+
+	optind = 1;
+	while (1) {
+	  c = getopt(argc, (char*const*)argv, "s:a:");
+	    
+	  if (c == EOF) break;
+	  switch (c) {
+	  case 's':
+	    spc = *optarg; /* First char only */
+	    break;
+	  case 'a':
+	    appstr = optarg;
+	    break;
+	  default:
+	    ++errflg;
+	    break;
+	  }
+	}
+	if (errflg || optind != argc - 1) {
+	  fprintf(stderr,
+		  "Usage: %s [ -s SPCCHR ] [ -a APPENDSTR ] string\n",
+		  argv[0]);
+	  return 1;
+	}
+
+	s = argv[optind];
+
+	mustquote = rfc822_mustquote(s, spc);
+	/* A bitset:
+	   0001  Has Quotes
+	   0002  Ended while inside a quote
+	   0004  Has characters which must be quoted
+	*/
+
+
+	if (!condq || mustquote == 1) {
+	  /* Well, we can actually DEQUOTE the thing just fine! */
+	  quoted = 0;
+	  for (; *s; ++s) {
+	    c = *s;
+	    if (c == ' ' && spc) {
+	      putchar(spc);
+	      quoted = 0;
+	      continue;
+	    }
+	    if (quoted) {
+	      putchar(c);
+	      quoted = 0;
+	      continue;
+	    }
+	    if (c == '\\') {
+	      /* A quoted pair! */
+	      quoted = 1;
+	      putchar(c);
+	      c = *++s;
+	    } else if (c == '"') {
+	      continue; /* Drop it! */
+	    } else
+	      putchar(c);
+	  }
+	} else if (mustquote > 1 && !(mustquote & 1) && condq) {
+	  /* Has things needing quotes, but no quotes in place! */
+	  putchar('"');
+	  for (; *s; ++s) {
+	    c = *s;
+	    if (c == ' ' && spc)
+	      putchar(spc);
+	    else if (c == '"') {
+	      putchar('\\');
+	      putchar(c);
+	    } else
+	      putchar(c);
+	  }
+	  putchar('"');
+	} else {
+	  /* The original one.. */
+	  for (; *s; ++s) {
+	    c = *s;
+	    if (c == ' ' && spc)
+	      putchar(spc);
+	    else if (c == '\\') {
+	      putchar(c);
+	    } else
+	      putchar(c);
+	  }
+	}
+	if (appstr)
+	  puts(appstr);
 
-	if (len > 1 &&
-	    ((*s == '"'  && s[len-1] == '"' ) ||
-	     (*s == '\'' && s[len-1] == '\''))) {
-	  fwrite(s+1,1,len-2,stdout);
-	} else
-	  fwrite(s,1,len,stdout);
-	putc('\n',stdout);
 	return 0;
 }
 
-
 static int
 run_condquote(argc, argv)
 	int argc;
 	const char *argv[];
 {
-	int len;
-	const char *s = argv[1];
-	int mustquote = 0;
-	int candequote = 0;
-	int c;
-
-	/* We remove quotes when they are not needed, and add them when
-	   they really are needed! */
-
-	if (argc != 2)
-	  return 2; /* Bad bad! Missing/extra arg! */
-
-	c = *s;
-	if (c == '"') {
-	  ++s; /* Starting quote */
-	  while (*s && *s != '"') {
-	    /* While within quoted string */
-	    if (*s == '\\') {
-	      ++s;
-	    }
-	    if (*s != 0)
-	      ++s;
-	  }
-	  if (*s != '"')
-	    mustquote = 1;
-	  else if (*s == 0)
-	    candequote = 1;
-	}
-	while (*s) {
-	  if (c == '\\') {
-	    ++s;
-	    c = *s;
-	    if (c == '!') /* fooo\!baar */
-	      mustquote = 1;
-	  } else if (c == ' ' || c == '\t')
-	    mustquote = 1; /* Unquoted spaces! */
-	  if (c != 0)
-	    ++s;
-	}
-
-	s = argv[1];
-	len = strlen(s);
-
-	/* Quoted, and without a need for quotes */
-	if (!mustquote && candequote) {
-	  /* XXX: THIS SHOULD REALLY USE SOME SYNTAX SCANNER -- LIKE THAT ONE
-	          FOR SMTP: RFC821SCN()    */
-	  fwrite(argv[1] + 1, 1, len -2, stdout); /* Dequoted! */
-	  putc('\n', stdout);
-	  return 0;
-	}
-	if (mustquote) {
-	  /* We need to quote this */
-	  s = argv[1];
-	  putchar('"'); /* First quote */
-	  while (*s) {
-	    if (*s == '"' || *s == '\\')
-	      putchar('\\');
-	    putchar(*s);
-	    ++s;
-	  }
-	  printf("\"\n"); /* The last quote */
-	  return 0;
-	}
-	/* The original one.. */
-	printf("%s\n",s);
-	return 0;
+  return run_condquote_(argc,argv,1);
+}
+
+static int
+run_dequote(argc, argv)
+	int argc;
+	const char *argv[];
+{
+  return run_condquote_(argc,argv,0);
 }
