@@ -2,7 +2,7 @@
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
  *
- *	Copyright 1996-1997 Matti Aarnio
+ *	Copyright 1996-1999 Matti Aarnio
  */
 
 /* LINTLIBRARY */
@@ -12,7 +12,11 @@
 #ifdef HAVE_FCNTL_H
 # include <fcntl.h>
 #endif
-#include <db.h>
+#if defined(HAVE_DB_185_H) && !defined(HAVE_DB_OPEN2)
+# include <db_185.h>
+#else
+# include <db.h>
+#endif
 #include <sys/file.h>
 #include "search.h"
 #include "io.h"
@@ -23,22 +27,24 @@
 extern int errno;
 extern int deferit;
 
-
+#ifndef HAVE_DB_OPEN2
 static BTREEINFO BINFO = { 0, 2560, 0, 0, 0, NULL,  NULL, 0 };
+#endif
 
 /*
  * Flush buffered information from this database, close any file descriptors.
  */
 
 void
-close_btree(sip)
+close_btree(sip,comment)
 	search_info *sip;
+	const char *comment;
 {
 	DB *db;
 	struct spblk *spl = NULL;
 	spkey_t symid;
 
-	if (sip->file == NULL || spt_files->symbols == NULL)
+	if (sip->file == NULL)
 		return;
 	symid = symbol_lookup_db(sip->file, spt_files->symbols);
 	if ((spkey_t)0 != symid)
@@ -46,9 +52,20 @@ close_btree(sip)
 	if (spl != NULL)
 	  sp_delete(spl, spt_modcheck);
 	spl = sp_lookup(symid, spt_files);
+#if 0
+{
+  char bb[2000];
+  sprintf(bb,"/tmp/-mark2- file='%s' symid=%lx spl=%p db=%p cmt=%s",sip->file,symid,spl, (spl) ? spl->data : NULL, comment );
+  unlink(bb);
+}
+#endif
 	if (spl == NULL || (db = (DB *)spl->data) == NULL)
 		return;
+#ifdef HAVE_DB_OPEN2
+	(db->close)(db,0);
+#else
 	(db->close)(db);
+#endif
 	symbol_free_db(sip->file, spt_files->symbols);
 	sp_delete(spl, spt_files);
 }
@@ -61,25 +78,38 @@ open_btree(sip, flag, comment)
 	int flag;
 	const char *comment;
 {
-	DB *db;
+	DB *db = NULL;
 	struct spblk *spl;
 	spkey_t symid;
 	int i;
 
 	if (sip->file == NULL)
 		return NULL;
-	if (spt_files == NULL)
-	  spt_files = sp_init();
-	if (spt_files->symbols == NULL)
-	  spt_files->symbols = sp_init();
 
 	symid = symbol_db(sip->file, spt_files->symbols);
 	spl = sp_lookup(symid, spt_files);
-	if (spl != NULL && flag == O_RDWR && spl->mark != O_RDWR)
-		close_btree(sip);
+#if 0
+{
+  char bb[2000];
+  sprintf(bb,"/tmp/-mark1- file='%s' symid=%lx spl=%p cmt=%s spl->mark=%lx flag=%x",sip->file,symid,spl,comment,spl?spl->mark:0,flag);
+  unlink(bb);
+}
+#endif
+	if (spl != NULL && flag != spl->mark)
+		close_btree(sip,"open_btree");
 	if (spl == NULL || (db = (DB *)spl->data) == NULL) {
 		for (i = 0; i < 3; ++i) {
+#ifdef HAVE_DB_OPEN2
+		  int err;
+		  db = NULL;
+		  /*unlink("/tmp/ -mark1- ");*/
+		  err = db_open(sip->file, DB_BTREE,
+				DB_NOMMAP|((flag == O_RDONLY) ? DB_RDONLY:DB_CREATE),
+				0644, NULL, NULL, &db);
+		  /*unlink("/tmp/ -mark2- ");*/
+#else
 		  db = dbopen(sip->file, flag, 0, DB_BTREE, &BINFO);
+#endif
 		  if (db != NULL)
 		    break;
 		  sleep(1); /* Open failed, retry after a moment */
@@ -92,9 +122,9 @@ open_btree(sip, flag, comment)
 			return NULL;
 		}
 		if (spl == NULL)
-			sp_install(symid, (void *)db, flag, spt_files);
-		else
-			spl->data = (void *)db;
+			spl = sp_install(symid, (void *)db, flag, spt_files);
+		spl->data = (void *)db;
+		spl->mark = flag;
 	}
 	return db;
 }
@@ -110,31 +140,41 @@ search_btree(sip)
 {
 	DB *db;
 	DBT val, key;
-	conscell *tmp;
-	struct spblk *spl;
 	int retry, rc;
-	spkey_t symid;
-	char *us;
 
 	retry = 0;
+#if 0
 reopen:
+#endif
 	db = open_btree(sip, O_RDONLY, "search_btree");
 	if (db == NULL)
 	  return NULL; /* Huh! */
 
-	key.data = (char*)sip->key;
+	memset(&key, 0, sizeof(key));
+	memset(&val, 0, sizeof(val));
+
+	key.data = (void*)sip->key;
 	key.size = strlen(sip->key) + 1;
+
+#ifdef HAVE_DB_OPEN2
+	rc = (db->get)(db, NULL, &key, &val, 0);
+#else
 	rc = (db->get)(db, &key, &val, 0);
+#endif
 	if (rc != 0) {
+
+#if 0 /* SleepyCat's DB 2.x leaks memory mappings when opening...
+	 at least the version at glibc 2.1.1 */
+
 		if (!retry && rc < 0) {
-			close_btree(sip);
+			close_btree(sip,"search_btree");
 			++retry;
 			goto reopen;
 		}
+#endif
 		return NULL;
 	}
-	us = strnsave(val.data, val.size);
-	return newstring(us);
+	return newstring(dupnstr(val.data, val.size), val.size);
 }
 
 
@@ -155,11 +195,22 @@ add_btree(sip, value)
 	if (db == NULL)
 		return EOF;
 
-	key.data = (char*)sip->key;
+	memset(&key, 0, sizeof(key));
+	memset(&val, 0, sizeof(val));
+
+	key.data = (void*)sip->key;
 	key.size = strlen(sip->key) + 1;
-	val.data = (char*)value;
+
+	val.data = (void*)value;
 	val.size = strlen(value)+1;
+#ifdef HAVE_DB_OPEN2
+	rc = (db->put)(db, NULL, &key, &val, 0);
+	/* Emulate BSD DB 1.85 behaviour */
+	if (rc != 0)
+	  rc = -1;
+#else
 	rc = (db->put)(db, &key, &val, 0);
+#endif
 	if (rc < 0) {
 		++deferit;
 		v_set(DEFER, DEFER_IO_ERROR);
@@ -186,9 +237,15 @@ remove_btree(sip)
 	if (db == NULL)
 		return EOF;
 
-	key.data = (char*)sip->key;
+	memset(&key, 0, sizeof(key));
+
+	key.data = (void*)sip->key;
 	key.size = strlen(sip->key) + 1;
+#ifdef HAVE_DB_OPEN2
+	rc = (db->del)(db, NULL, &key, 0);
+#else
 	rc = (db->del)(db, &key, 0);
+#endif
 	if (rc < 0) {
 		++deferit;
 		v_set(DEFER, DEFER_IO_ERROR);
@@ -211,11 +268,47 @@ print_btree(sip, outfp)
 	DB *db;
 	DBT key, val;
 	int rc;
+#ifdef HAVE_DB_OPEN2
+	DBC *curs;
 
 	db = open_btree(sip, O_RDONLY, "print_btree");
 	if (db == NULL)
 		return;
 
+#ifdef HAVE_DB_CURSOR4
+	rc = (db->cursor)(db, NULL, &curs, 0);
+#else
+	rc = (db->cursor)(db, NULL, &curs);
+#endif
+
+	memset(&val, 0, sizeof(val));
+	memset(&key, 0, sizeof(key));
+
+	if (rc == 0 && curs)
+	  rc = (curs->c_get)(curs, &key, &val, DB_FIRST);
+	for ( ; rc == 0 ; ) {
+		if (val.data == NULL)
+			continue;
+		if (*(char*)val.data == '\0')
+			fprintf(outfp, "%s\n", key.data);
+		else
+			fprintf(outfp, "%s\t%s\n", key.data, val.data);
+
+		memset(&val, 0, sizeof(val));
+		memset(&key, 0, sizeof(key));
+
+		rc = (curs->c_get)(curs, &key, &val, DB_NEXT);
+	}
+	(curs->c_close)(curs);
+#else
+
+	db = open_btree(sip, O_RDONLY, "print_btree");
+	if (db == NULL)
+		return;
+
+	memset(&val, 0, sizeof(val));
+	memset(&key, 0, sizeof(key));
+
 	rc = (db->seq)(db, &key, &val, R_FIRST);
 	for ( ; rc == 0 ; ) {
 		if (val.data == NULL)
@@ -224,8 +317,13 @@ print_btree(sip, outfp)
 			fprintf(outfp, "%s\n", key.data);
 		else
 			fprintf(outfp, "%s\t%s\n", key.data, val.data);
+
+		memset(&val, 0, sizeof(val));
+		memset(&key, 0, sizeof(key));
+
 		rc = (db->seq)(db, &key, &val, R_NEXT);
 	}
+#endif
 	fflush(outfp);
 }
 
@@ -242,17 +340,47 @@ count_btree(sip, outfp)
 	DBT key, val;
 	int cnt = 0;
 	int rc;
+#ifdef HAVE_DB_OPEN2
+	DBC *curs;
 
+	db = open_btree(sip, O_RDONLY, "count_btree");
+
+	if (db != NULL) {
+#ifdef HAVE_DB_CURSOR4
+	  rc = (db->cursor)(db, NULL, &curs, 0);
+#else
+	  rc = (db->cursor)(db, NULL, &curs);
+#endif
+
+	  memset(&val, 0, sizeof(val));
+	  memset(&key, 0, sizeof(key));
+
+	  if (rc == 0 && curs)
+	    rc = (curs->c_get)(curs, &key, &val, DB_FIRST);
+	  while (rc == 0) {
+	    if (val.data == NULL) /* ???? When this would happen ? */
+	      continue;
+	    ++cnt;
+
+	    memset(&val, 0, sizeof(val));
+	    memset(&key, 0, sizeof(key));
+
+	    rc = (curs->c_get)(curs, &key, &val, DB_NEXT);
+	  }
+	}
+	(curs->c_close)(curs);
+#else
 	db = open_btree(sip, O_RDONLY, "count_btree");
 	if (db != NULL) {
 	  rc = (db->seq)(db, &key, &val, R_FIRST);
 	  while (rc == 0) {
-	    if (val.data == NULL)
+	    if (val.data == NULL) /* ???? When this would happen ? */
 	      continue;
 	    ++cnt;
 	    rc = (db->seq)(db, &key, &val, R_NEXT);
 	  }
 	}
+#endif
 	fprintf(outfp,"%d\n",cnt);
 	fflush(outfp);
 }
@@ -270,6 +398,7 @@ owner_btree(sip, outfp)
 {
 	DB *db;
 	struct stat stbuf;
+	int fd;
 
 	db = open_btree(sip, O_RDONLY, "owner_btree");
 	if (db == NULL)
@@ -277,7 +406,12 @@ owner_btree(sip, outfp)
 
 	/* There are timing hazards, when the internal fd is not
 	   available for probing.. */
-	if (fstat((db->fd)(db), &stbuf) < 0) {
+#ifdef HAVE_DB_OPEN2
+	(db->fd)(db, &fd);
+#else
+	fd = (db->fd)(db);
+#endif
+	if (fstat(fd, &stbuf) < 0) {
 		fprintf(stderr, "owner_btree: cannot fstat(\"%s\")!\n",
 				sip->file);
 		return;
@@ -294,13 +428,18 @@ modp_btree(sip)
 	struct stat stbuf;
 	struct spblk *spl;
 	spkey_t symid;
-	int rval;
+	int rval, fd;
 
 	db = open_btree(sip, O_RDONLY, "owner_btree");
 	if (db == NULL)
 		return 0;
 
-	if (fstat((db->fd)(db), &stbuf) < 0) {
+#ifdef HAVE_DB_OPEN2
+	(db->fd)(db, &fd);
+#else
+	fd = (db->fd)(db);
+#endif
+	if (fstat(fd, &stbuf) < 0) {
 		fprintf(stderr, "modp_btree: cannot fstat(\"%s\")!\n",
 				sip->file);
 		return 0;
@@ -312,11 +451,12 @@ modp_btree(sip)
 	spl = sp_lookup(symid, spt_modcheck);
 	if (spl != NULL) {
 		rval = ((long)stbuf.st_mtime != (long)spl->data ||
-			(long)stbuf.st_nlink != (long)spl->mark);
+			(long)stbuf.st_nlink != 1);
 	} else
 		rval = 0;
 	sp_install(symid, (void *)((long)stbuf.st_mtime),
 		   stbuf.st_nlink, spt_modcheck);
+
 	return rval;
 }
 #endif	/* HAVE_DB */
