@@ -18,91 +18,6 @@
 #include "libc.h"
 #include "libsh.h"
 
-/*
- * Search an NDBM format database for a key pair.
- */
-
-conscell *
-search_ndbm(sip)
-	search_info *sip;
-{
-	DBM *db;
-	datum val, key;
-	conscell *tmp;
-	struct spblk *spl;
-	int retry;
-	spkey_t symid;
-	char *us;
-
-	if (sip->file == NULL)
-		return NULL;
-	retry = 0;
-	symid = symbol(sip->file);
-	spl = sp_lookup(symid, spt_files);
-	if (spl == NULL || (db = (DBM *)spl->data) == NULL) {
-reopen:
-		db = dbm_open((char* /* Sigh.. */)sip->file, O_RDONLY, 0); 
-		if (db == NULL) {
-			++deferit;
-			v_set(DEFER, DEFER_IO_ERROR);
-			fprintf(stderr, "search_ndbm: cannot open %s!\n",
-					sip->file);
-			return NULL;
-		}
-		if (spl == NULL)
-			sp_install(symid, (void *)db, O_RDONLY, spt_files);
-		else
-			spl->data = (void *)db;
-	}
-	key.dptr  = (char*) sip->key; /* Sigh.. the cast.. */
-	key.dsize = strlen(sip->key) + 1;
-	val = dbm_fetch(db, key);
-	if (val.dptr == NULL) {
-#ifdef HAVE_DBM_ERROR
-	  if (!retry && dbm_error(db)) {
-	    dbm_close(db);
-	    ++retry;
-	    goto reopen;
-	  }
-#else
-	  if (!retry && errno != 0) {
-	    dbm_close(db);
-	    ++retry;
-	    goto reopen;
-	  }
-#endif
-	  return NULL;
-	}
-	us = strnsave(val.dptr, val.dsize);
-	return newstring(us);
-}
-
-/*
- * Flush buffered information from this database, close any file descriptors.
- */
-
-void
-close_ndbm(sip)
-	search_info *sip;
-{
-	DBM *db;
-	struct spblk *spl;
-	spkey_t symid;
-
-	if (sip->file == NULL)
-		return;
-	symid = symbol(sip->file);
-	spl = sp_lookup(symid, spt_modcheck);
-	if (spl != NULL)
-		sp_delete(spl, spt_modcheck);
-	spl = sp_lookup(symid, spt_files);
-	if (spl == NULL || (db = (DBM *)spl->data) == NULL)
-		return;
-	dbm_close(db);
-	sp_install(symid, (u_char *)NULL, 0, spt_files);
-}
-
-
 static DBM * open_ndbm __((search_info *, int, const char *));
 static DBM *
 open_ndbm(sip, flag, comment)
@@ -110,19 +25,25 @@ open_ndbm(sip, flag, comment)
 	int flag;
 	const char *comment;
 {
-	DBM *db;
+	DBM *db = NULL;
 	struct spblk *spl;
 	spkey_t symid;
+	int i;
 
-	if (sip->file == NULL)
-		return NULL;
-	symid = symbol(sip->file);
+	if (sip->file == NULL)		return NULL;
+
+	symid = symbol_db(sip->file, spt_files->symbols);
 	spl = sp_lookup(symid, spt_files);
 	if (spl != NULL && flag == O_RDWR && spl->mark != O_RDWR)
-		close_ndbm(sip);
+		close_ndbm(sip,"open_ndbm");
 	if (spl == NULL || (db = (DBM *)spl->data) == NULL) {
-		db = dbm_open((char * /* Sigh.. the cast.. */)sip->file,
-			      flag, 0);
+		for (i = 0; i < 3; ++i) {
+		  db = dbm_open(sip->file, flag, 0);
+		  if (db != NULL)
+		    break;
+		  sleep(1);
+		}
+
 		if (db == NULL) {
 			++deferit;
 			v_set(DEFER, DEFER_IO_ERROR);
@@ -138,6 +59,76 @@ open_ndbm(sip, flag, comment)
 	return db;
 }
 
+/*
+ * Search an NDBM format database for a key pair.
+ */
+
+conscell *
+search_ndbm(sip)
+	search_info *sip;
+{
+	DBM *db;
+	datum val, key;
+	conscell *tmp;
+	int retry;
+
+	retry = 0;
+
+reopen:
+
+	db = open_ndbm(sip, O_RDONLY, "search_ndbm");
+	if (db == NULL)
+	  return NULL; /* Failed :-( */
+
+	key.dptr  = (void*) sip->key; /* Sigh.. the cast.. */
+	key.dsize = strlen(sip->key) + 1;
+	val = dbm_fetch(db, key);
+	if (val.dptr == NULL) {
+#ifdef HAVE_DBM_ERROR
+	  if (!retry && dbm_error(db)) {
+	    close_ndbm(sip,"search_ndbm");
+	    ++retry;
+	    goto reopen;
+	  }
+#else
+	  if (!retry && errno != 0) {
+	    close_ndbm(sip,"search_ndbm");
+	    ++retry;
+	    goto reopen;
+	  }
+#endif
+	  return NULL;
+	}
+	return newstring(dupnstr(val.dptr, val.dsize), val.dsize);
+}
+
+/*
+ * Flush buffered information from this database, close any file descriptors.
+ */
+
+void
+close_ndbm(sip,comment)
+	search_info *sip;
+	const char *comment;
+{
+	DBM *db;
+	struct spblk *spl;
+	spkey_t symid;
+
+	if (sip->file == NULL)
+		return;
+	symid = symbol_db(sip->file, spt_files->symbols);
+	spl = sp_lookup(symid, spt_modcheck);
+	if (spl != NULL)
+		sp_delete(spl, spt_modcheck);
+	spl = sp_lookup(symid, spt_files);
+	if (spl == NULL || (db = (DBM *)spl->data) == NULL)
+		return;
+	dbm_close(db);
+	sp_delete(spl, spt_files);
+	symbol_free_db(sip->file, spt_files->symbols);
+}
+
 /*
  * Add the indicated key/value pair to the database.
  */
@@ -153,9 +144,10 @@ add_ndbm(sip, value)
 	db = open_ndbm(sip, O_RDWR, "add_ndbm");
 	if (db == NULL)
 		return EOF;
-	key.dptr  = (char*) sip->key;	/* Sigh.. the cast.. */
+
+	key.dptr  = (void*) sip->key;	/* Sigh.. the cast.. */
 	key.dsize = strlen(sip->key) + 1;
-	val.dptr  = (char*) value;	/* Sigh.. the cast.. */
+	val.dptr  = (void*) value;	/* Sigh.. the cast.. */
 	val.dsize = strlen(value) + 1;
 	if (dbm_store(db, key, val, DBM_REPLACE) < 0) {
 		++deferit;
@@ -181,7 +173,8 @@ remove_ndbm(sip)
 	db = open_ndbm(sip, O_RDWR, "remove_ndbm");
 	if (db == NULL)
 		return EOF;
-	key.dptr  = (char*) sip->key;	/* Sigh.. the cast.. */
+
+	key.dptr  = (void*) sip->key;	/* Sigh.. the cast.. */
 	key.dsize = strlen(sip->key) + 1;
 	if (dbm_delete(db, key) < 0) {
 		++deferit;
@@ -208,6 +201,7 @@ print_ndbm(sip, outfp)
 	db = open_ndbm(sip, O_RDONLY, "print_ndbm");
 	if (db == NULL)
 		return;
+
 	for (key = dbm_firstkey(db); key.dptr != NULL; key = dbm_nextkey(db)) {
 		val = dbm_fetch(db, key);
 		if (val.dptr == NULL)
@@ -237,15 +231,14 @@ count_ndbm(sip, outfp)
 	FILE *outfp;
 {
 	DBM *db;
-	datum key, val;
+	datum key;
 	int cnt = 0;
 
 	db = open_ndbm(sip, O_RDONLY, "count_ndbm");
 	if (db != NULL)
-	  for (key = dbm_firstkey(db); key.dptr != NULL; key = dbm_nextkey(db)) {
-	    val = dbm_fetch(db, key);
-	    if (val.dptr == NULL)
-	      continue;
+	  for (key = dbm_firstkey(db);
+	       key.dptr != NULL;
+	       key = dbm_nextkey(db)) {
 #ifdef HAVE_DBM_ERROR
 	    if (dbm_error(db))
 	      break;
@@ -297,9 +290,10 @@ modp_ndbm(sip)
 	spkey_t symid;
 	int rval;
 
-	if (sip->file == NULL
-	    || (db = open_ndbm(sip, O_RDONLY, "owner_ndbm")) == NULL)
+	db = open_ndbm(sip, O_RDONLY, "owner_ndbm");
+	if (db == NULL)
 		return 0;
+
 	if (fstat(dbm_dirfno(db), &stbuf) < 0) {
 		fprintf(stderr, "modp_ndbm: cannot fstat(\"%s\")!\n",
 			sip->file);
@@ -308,7 +302,7 @@ modp_ndbm(sip)
 	if (stbuf.st_nlink == 0)
 		return 1;	/* Unlinked underneath of us! */
 
-	symid = symbol(sip->file);
+	symid = symbol_db(sip->file, spt_files->symbols);
 	spl = sp_lookup(symid, spt_modcheck);
 	if (spl != NULL) {
 		rval = ((long)stbuf.st_mtime != (long)spl->data ||
