@@ -25,8 +25,8 @@
 
 /* various libraries, standard libc included.. */
 
-extern const char * gs_name;
-extern const char * monthname[];
+extern const char * const gs_name;
+extern const char * const monthname[];
 extern char	*prio_list[];
 #ifndef HAVE_STRERROR /* System has it, and propably has prototype too..
 			 IRIX 6.2 */
@@ -46,7 +46,7 @@ extern int	tolower __((const int));
 /* File: conf.c */
 extern int	 files_gid;
 extern char	*cf_suffix;
-extern char	*default_trusted[];
+extern const char *default_trusted[];
 extern char	*err_prio_list[];
 extern char	*files_group;
 extern char	*logdir;
@@ -66,7 +66,7 @@ extern struct sptree	*spt_databases, *spt_files, *spt_modcheck,
 extern struct sptree_init splaytrees[];
 
 /* File: dateparse.c */
-extern long	dateParse __((struct tm *localtmptr, struct token *t));
+extern long	dateParse __((struct tm *localtmptr, token822 *t));
 
 /* File: db.c */
 extern int	run_relation __((int argc, const char *argv[]));
@@ -116,8 +116,7 @@ extern void	squirrel __((struct envelope *e, const char *keyword, const char *te
 extern struct header	*erraddress __((struct envelope *e));
 extern void	defer __((struct envelope *e, const char *why));
 extern struct header	*mkSender __((struct envelope *e, const char *name, int flag));
-extern struct header	*mkTrace __((struct envelope *e));
-extern int	nullhost __((const char *s));
+extern struct header	*mkTrace __((struct envelope *e, struct header *rcvdhdr));
 extern conscell	*pickaddress __((conscell *l));
 extern int	thesender __((struct envelope *e, struct address *a));
 extern conscell	*makequad __((void));
@@ -138,8 +137,8 @@ extern void	hdr_print __((struct header *h, FILE *fp));
 extern int	hdr_nilp __((struct header *h));
 extern void	pureAddress __((FILE *fp, struct addr *pp));
 extern int	pureAddressBuf __((char *buf, int len, struct addr *pp));
-extern int	printAddress __((FILE *fp, struct addr *pp, int onlylength));
-extern int	printLAddress __((FILE *fp, struct addr *pp, int col));
+extern int	printAddress __((FILE *fp, struct addr *pp, int col));
+extern int	printLAddress __((FILE *fp, struct addr *pp, int col, int foldcol, int nofold));
 extern char    *saveAddress __((struct addr *pp));
 extern void	errprint __((FILE *fp, struct addr *pp, int hdrlen));
 extern HeaderStamp hdr_type __((struct header *h));
@@ -148,7 +147,7 @@ extern void	hdr_errprint __((struct envelope *e, struct header *h, FILE *fp, con
 
 /* File: rfc822walk.c */
 extern struct address	*revaddress __((struct address *ap));
-extern union misc	 parse822 __((HeaderSemantics entry, struct token **tlistp, struct tm *ltm, FILE *tfp));
+extern union misc	 parse822 __((HeaderSemantics entry, token822 **tlistp, struct tm *ltm, FILE *tfp));
 
 
 /* File: router.c */
@@ -163,9 +162,10 @@ extern int   canexit;
 extern int   deferit;
 extern int   router_id;
 extern int   savefile;
+extern int   do_hdr_warning;
 
 /* File: rtsyslog.c */
-extern void rtsyslog __(( time_t msgmtime, const char *msgfile, const char *from, const char *smtprelay, int size, int nrcpts, const char *msgid ));
+extern void rtsyslog __(( time_t msgmtime, long msgino, const char *from, const char *smtprelay, int size, int nrcpts, const char *msgid ));
 
 extern int	main __((int argc, const char *argv[]));
 extern int	login_to_uid __((const char *name));
