@@ -217,8 +217,13 @@ run_rfc822(argc, argv)
 
 	if (D_final > 0)
 		dumpInfo(e);
+
+	/* This will always make some noise in successfull case,
+	   except that it is SILENT code... */
+
 	if (status != PERR_CTRLFILE && status != PERR_DEFERRED && !savefile)
 		(void) zunlink(file); /* SILENT! */
+
 	if (e->e_fp != NULL)
 		fclose(e->e_fp);
 	tfree(MEM_TEMP);
@@ -439,6 +444,8 @@ dumpHeader(h)
 	case nilHeaderSemantics:
 		for (t = h->h_lines; t != NULL; t = t->t_next)
 			printf("\t%s\n", formatToken(t));
+		for (t = h->h_contents.t; t != NULL; t = t->t_next)
+			printf("\t%s\n", formatToken(t));
 		break;
 	case Received:
 		if (h->h_contents.r == NULL)
@@ -496,6 +503,14 @@ int	isRecpntAddr = 0;
 	for (h = e->e_eHeaders; h != NULL; h = h->h_next) \
 	    if (h->h_descriptor->class == X) \
 		break;
+#define FindEnvelopeLast(X)	\
+	{ struct header *__h_1 = NULL;			\
+	  for (h = e->e_eHeaders; h != NULL; h = h->h_next) \
+	    if (h->h_descriptor->class == X)		\
+		__h_1 = h;				\
+	  h = __h_1;					\
+	}
+
 
 #define RESENT_SIZE (sizeof "resent-" - 1)
 
@@ -810,7 +825,9 @@ defer(e, why)
 	s = path;
 	while ((s = strchr(s, ' '))) /* remap blanks to '-':es.. */
 	  *s = '-';
+#ifndef HAVE_RENAME
 	zunlink(path);
+#endif
 
 	/* try renameing every few seconds for a minute */
 	for (i = 0; erename(e->e_file, path) < 0 && i < 10 && errno != ENOENT;++i)
@@ -975,8 +992,9 @@ static const struct headerinfo	traceDesc =
 	{ "received", Received, nilUserType, normal };
 
 struct header *
-mkTrace(e)
+mkTrace(e, rcvdhdr)
 	struct envelope *e;
+	struct header *rcvdhdr;
 {
 	register struct header	*h, *th;
 	struct addr *na;
@@ -990,7 +1008,7 @@ mkTrace(e)
 	th->h_contents.r = (struct received *)tmalloc(sizeof (struct received));
 	th->h_lines = NULL;
 
-	FindEnvelope(eRcvdFrom);
+	h = rcvdhdr; /* FindEnvelopeLast(eRcvdFrom); */
 	/* from */
 	if (h != NULL)
 		th->h_contents.r->r_from = h->h_contents.a;
@@ -1168,6 +1186,7 @@ conscell *rwmappend(rwmroot,info,errtop)
  */
 
 int ReceivedCount = 0;
+static int schedulersubdirhash = -1;
 
 int
 sequencer(e, file)
@@ -1175,6 +1194,7 @@ sequencer(e, file)
 	const char *file;
 {
 	struct header  *h, *ph, *nh = NULL, *oh, *msgidh, **hp;
+	struct header  *rcvdhdr;
 	struct address *a, *ap;
 	struct addr    *p = NULL;
 	char	       *ofpname, *path, *qpath;
@@ -1193,8 +1213,30 @@ sequencer(e, file)
 	const char     *fromaddr = "?from?";
 	const char     *msgidstr = "?msgid?";
 	const char     *smtprelay = NULL;
+	char subdirhash[8];
 	GCVARS5;
 
+	if (schedulersubdirhash < 0) {
+	  char *s = getzenv("SCHEDULERDIRHASH");
+	  if (s && ((s[0] == '1' || s[0] == '2') && s[1] == 0))
+	    schedulersubdirhash = s[0] - '0';
+	  else
+	    schedulersubdirhash = 0;
+	}
+
+	if (schedulersubdirhash) {
+	  long ino = e->e_statbuf.st_ino;
+	  if (schedulersubdirhash > 1) {
+	    int h1 = (ino / 26) % 26;
+	    int h2 = ino % 26;
+	    sprintf(subdirhash, "%c/%c/", h1 + 'A', h2 + 'A');
+	  } else {
+	    int h2 = ino % 26;
+	    sprintf(subdirhash, "%c/", h2 + 'A');
+	  }
+	} else
+	  *subdirhash = 0;
+
 	deferuid = 0;
 
 	errors_to = NULL; /* to be gotten rid off.. */
@@ -1211,10 +1253,11 @@ sequencer(e, file)
 	if (deferuid)
 	  return PERR_DEFERRED;
 
-	FindEnvelope(eRcvdFrom);
+	FindEnvelopeLast(eRcvdFrom);
+	rcvdhdr = h;
 	if (e->e_trusted) {
-	  if (h != NULL)
-	    smtprelay = h->h_lines->t_pname;
+	  if (rcvdhdr != NULL)
+	    smtprelay = rcvdhdr->h_lines->t_pname;
 	  else {
 	    const char *s  = uidpwnam(e->e_statbuf.st_uid);
 	    char *sr = strnsave("",strlen(s)+20);
@@ -1226,7 +1269,7 @@ sequencer(e, file)
 	  char *ts;
 	  int totlen;
 
-	  if (h != NULL) {
+	  if (rcvdhdr != NULL) {
 
 	    /* Ok, we don't trust it!  In fact we might overrule the data
 	       that the message writer coded in... */
@@ -1235,15 +1278,15 @@ sequencer(e, file)
 	    s = uidpwnam(e->e_statbuf.st_uid);
 	    ps = "";
 	    totlen = 10 + strlen(s) + 60;
-	    if (h) {
-	      ps = h->h_lines->t_pname;
+	    if (rcvdhdr) {
+	      ps = rcvdhdr->h_lines->t_pname;
 	      totlen += strlen(ps) + 10;
 	    }
 	    /* ts = strnsave("", totlen); */
 	    ts = tmalloc(totlen); /* Alloc the space */
-	    if (h) {
+	    if (rcvdhdr) {
 	      sprintf(ts, "rcvdfrom STDIN (from localhost user: '%s' uid#%ld fake: %s)",
-		      s, (long) e->e_statbuf.st_uid, h->h_lines->t_pname);
+		      s, (long) e->e_statbuf.st_uid, rcvdhdr->h_lines->t_pname);
 	    } else
 	      sprintf(ts, "rcvdfrom STDIN (from localhost user: '%s' uid#%ld)",
 		      s, (long) e->e_statbuf.st_uid);
@@ -1258,14 +1301,14 @@ sequencer(e, file)
 		    s, (long) e->e_statbuf.st_uid);
 	  }
 
-	  h = makeHeader(spt_eheaders, ts, 8);
-	  h->h_next = e->e_eHeaders;
-	  e->e_eHeaders = h;
+	  rcvdhdr = makeHeader(spt_eheaders, ts, 8);
+	  rcvdhdr->h_next = e->e_eHeaders;
+	  e->e_eHeaders = rcvdhdr;
 
-	  h->h_lines    = makeToken(ts+9, strlen(ts+9));
-	  h->h_lines->t_type = Line;
-	  h->h_contents = hdr_scanparse(e, h, 0, 0);
-	  h->h_stamp    = hdr_type(h);
+	  rcvdhdr->h_lines    = makeToken(ts+9, strlen(ts+9));
+	  rcvdhdr->h_lines->t_type = Line;
+	  rcvdhdr->h_contents = hdr_scanparse(e, rcvdhdr, 0, 0);
+	  rcvdhdr->h_stamp    = hdr_type(rcvdhdr);
 
 	  smtprelay     = ts+9;
 	}
@@ -1278,7 +1321,7 @@ sequencer(e, file)
 	perr = 0;
 	if (myhostname) {	/* appease honeyman et al groupies */
 		dprintf("Stamp it with a trace header\n");
-		h = mkTrace(e);
+		h = mkTrace(e, rcvdhdr);
 		if (h != NULL) {
 			h->h_next = e->e_headers;
 			e->e_headers = h;
@@ -1489,7 +1532,8 @@ sequencer(e, file)
 				optsave(FYI_NOCHANNEL, e);
 			}
 		}
-		FindEnvelope(eRcvdFrom);
+
+		h = rcvdhdr; /* FindEnvelopeLast(eRcvdFrom); */
 		if (h != NULL && h->h_contents.a && h->h_contents.a->a_pname) {
 			/* a previous host was specified */
 			slen = strlen(h->h_contents.a->a_pname);
@@ -1666,16 +1710,11 @@ sequencer(e, file)
 	while (*hp != NULL) {
 	  h = *hp;
 	  if (h->h_descriptor->hdr_name != NULL &&
-	      ((h->h_descriptor->class == normal &&
-		(CISTREQ(h->h_descriptor->hdr_name,"bcc") ||
-		 CISTREQ(h->h_descriptor->hdr_name,"return-path") ||
-		 CISTREQ(h->h_descriptor->hdr_name,"x-orcpt")     ||
-		 CISTREQ(h->h_descriptor->hdr_name,"x-envid")        ))
+	      ((h->h_descriptor->user_type == killUserType)
+	       || (h->h_descriptor->class == normal &&
+		   (CISTREQ(h->h_descriptor->hdr_name,"bcc")))
 	       || (h->h_descriptor->class == Resent &&
-		   (CISTREQ(h->h_descriptor->hdr_name,"resent-BCC")         ||
-		    CISTREQ(h->h_descriptor->hdr_name,"resent-return-path") ||
-		    CISTREQ(h->h_descriptor->hdr_name,"resent-x-orcpt")     ||
-		    CISTREQ(h->h_descriptor->hdr_name,"resent-x-envid")  )))) {
+		   (CISTREQ(h->h_descriptor->hdr_name,"resent-bcc"))))) {
 	    /* Skip this one */
 	    *hp = h->h_next;
 	  } else {
@@ -2330,11 +2369,11 @@ sequencer(e, file)
 	}
 	/* all is nirvana -- link the input file to somewhere safe */
 #ifdef	USE_ALLOCA
-	qpath = (char*)alloca(5+strlen(QUEUEDIR)+strlen(file));
+	qpath = (char*)alloca(12+strlen(QUEUEDIR)+strlen(file));
 #else
-	qpath = (char*)emalloc(5+strlen(QUEUEDIR)+strlen(file));
+	qpath = (char*)emalloc(12+strlen(QUEUEDIR)+strlen(file));
 #endif
-	sprintf(qpath, "../%s/%s", QUEUEDIR, file);
+	sprintf(qpath, "../%s/%s%s", QUEUEDIR, subdirhash, file);
 	fflush(ofp);
 #ifdef HAVE_FSYNC
 	fsync(FILENO(ofp));
@@ -2352,14 +2391,16 @@ sequencer(e, file)
 	}
 
 #ifndef USE_ALLOCA
-	path = (char*)emalloc(5+strlen(TRANSPORTDIR)+strlen(file));
+	path = (char*)emalloc(12+strlen(TRANSPORTDIR)+strlen(file));
 #else
 	/* This actually reallocs more space from stack, but then it
 	   is just stack space and will disappear.. */
-	path = (char*)alloca(5+strlen(TRANSPORTDIR)+strlen(file));
+	path = (char*)alloca(12+strlen(TRANSPORTDIR)+strlen(file));
 #endif
-	sprintf(path, "../%s/%s", TRANSPORTDIR, file);
+	sprintf(path, "../%s/%s%s", TRANSPORTDIR, subdirhash, file);
+#ifndef HAVE_RENAME
 	zunlink(path);	/* Should actually always fail.. */
+#endif
 	if (erename(ofpname, path) < 0) {
 	  zunlink(qpath);
 	  zunlink(path);
@@ -2436,19 +2477,37 @@ prctladdr(info, fp, cfflag, comment)
 	      if (x->string == NULL || *x->string == '\0')
 		putc('-', fp);
 	      else {
-		char *s = strchr(x->string,' ');
-		if (!s) s = strchr(x->string,'\t');
-		if (s && x->string[0] != '"') {
-		  s = x->string;
+		char *s = x->string;
+		int quote = 0;
+		/* Do same scanner as  skip821address()  will do with
+		   our result.  Does it arrive to the end-of-string ?
+		   Does it do it with 'quote' state CLEAR ? */
+		for ( ; *s; ++s) {
+		  int c = *s;
+		  if (c == '\\') {
+		    ++s;
+		    if (*s == 0) break;
+		  }
+		  if (c == quote) /* 'c' is non-zero here */
+		    quote = 0;
+		  else if (c == '"')
+		    quote = c;
+		  else if (!quote && (c == ' ' || c == '\t'))
+		    break;
+		}
+
+		if (*s || quote) {
+		  /* Didn't arrive to the end-of-string with quote clear! */
+
 		  putc('"',fp);
-		  while (*s) {
+		  for (s = x->string; *s; ++s) {
 		    if (*s == '"' || *s == '\\')
 		      putc('\\',fp);
 		    putc(*s,fp);
-		    ++s;
 		  }
 		  putc('"',fp);
 		} else
+		  /* It is clean string not needing extra quotes */
 		  fprintf(fp, "%s", x->string);
 	      }
 	      if (i == 4 && x->string != NULL)
