@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1995
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2000
  */
 
 
@@ -51,6 +51,11 @@ time_t	now;
 extern memtypes stickymem;
 extern conscell *s_value;
 
+#include "shmmib.h"
+
+struct MIB_MtaEntry MIBMtaEntryLcl;
+struct MIB_MtaEntry *MIBMtaEntry = &MIBMtaEntryLcl;
+
 int	mustexit = 0;
 int	canexit = 0;
 int	router_id = 0;
@@ -565,13 +570,15 @@ logmessage(e)
 						to = buf;
 					}
 					if (to + len + 2 >= buf + sizeof buf)
-						continue;
+					  /* Can't fit it in.. */
+					  continue;
 				} else {
 					if (n) *to++ = ',';
 					*to++ = ' ';
 				}
-				strncpy(to, cp, len);
+				memcpy(to, cp, len);
 				to += len;
+				*to = 0;
 				++n;
 			}
 		}
@@ -598,16 +605,15 @@ logit(file, id, from, to)
 	if (id == NULL)
 		id = file;
 	baselen = strlen(file) + strlen(id) + 4;
-	printf("%.200s: file: %.150s %.250s...\n", id, file, from);
+	printf("%.200s: file: %s %s...\n", id, file, from);
 	if (!nosyslog)
 	  zsyslog((LOG_INFO, "%.200s: file: %.150s %.250s...",
 		   id, file, from));
 	flen = strlen(from);
 	if (flen > 0)
-	  printf("%.200s: file: %.150s %.200s => %.200s\n",
-		 id, file, from, to);
+	  printf("%.200s: file: %s %s => %s\n", id, file, from, to);
 	else
-	  printf("%.200s: file: %.150s %.250s\n", id, file, to);
+	  printf("%.200s: file: %s %s\n",       id, file, to);
 	if (!nosyslog) {
 	  if (flen > 0)
 	    zsyslog((LOG_INFO, "%.200s: file: %.150s %.200s => %.200s",
