@@ -12,62 +12,50 @@
 #include "scheduler.h"
 #include "prototypes.h"
 
-time_t qipcretry = 0;
-
 int	/* return non-zero when there are some childs
 	   available for start				*/
 doagenda()
 {
-	int didsomething;
 	struct thread *ncuritem, *nncuritem;
+	int didsomething = 0;
 
 	mytime(&now);
 
 	ncuritem = thread_head;
-	didsomething = 0;
-	do {
-	  didsomething = 0;
-	  if (verbose)
-	    sfprintf(sfstdout,"curitem %p curitem->wakeup %lu now %d\n",
-		     ncuritem, ncuritem ? ncuritem->wakeup : 0, (int)now);
 
-	  /* thread_head -chain should be in time order, thus the while-loop
-	     should be traversed only once, because  thread_start kicks
-	     right away (or reschedules) */
+	if (verbose)
+	  sfprintf(sfstdout,"curitem %p curitem->wakeup %lu now %d\n",
+		   ncuritem, ncuritem ? ncuritem->wakeup : 0, (int)now);
 
-	  while (ncuritem  && ncuritem->wakeup <= now  ) {
-	    /* Object pointed by  ncuritem  may disappear due to
-	       expiration at thread_start() .. */
-	    nncuritem = ncuritem->nexttr;
-	    /* Not running, and wakeup in past ? */
-	    if (ncuritem->proc == NULL && ncuritem->wakeup <= now)
-	      /* Try to start it! */
-	      while (thread_start(ncuritem, 0))
-		didsomething += 1;
+	/* thread_head -chain should be in time order, thus the while-loop
+	   should be traversed only once, because  thread_start kicks
+	   right away (or reschedules) */
+
+	while (ncuritem  && ncuritem->wakeup <= now  ) {
+	  /* Object pointed by  ncuritem  may disappear due to
+	     expiration at thread_start() .. */
+	  nncuritem = ncuritem->nexttr;
+	  /* Not running, and wakeup in past ? */
+	  if (ncuritem->proc == NULL && ncuritem->wakeup <= now)
+	    /* Try to start it! */
+	    while (thread_start(ncuritem, 0))
+	      ++ didsomething;
+
+	  ncuritem = nncuritem;
 
-	    ncuritem = nncuritem;
-	  }
 	  queryipccheck();
-	} while (didsomething && ncuritem);
+	}
 
 	mytime(&now);
 
 	/* if (verbose)
 	   printf("alarmed %d\n", now);  */
 
-	if (qipcretry > 0 && qipcretry <= now) {
-	  qipcretry = 0;
-	  queryipcinit();
-	  /*
-	   * If qipcretry is set here, the value will be ignored, but
-	   * that's ok since sweepretry is active by now
-	   */
-	}
 	return (didsomething);
 }
 
 /* Do immediate scheduling of given host-indentifier. */
-void
+int
 turnme(turnarg)
 const char *turnarg;
 {
@@ -76,6 +64,7 @@ const char *turnarg;
 	char *cp = strchr(turnarg,' ');
 	struct thread *ncuritem, *nncuritem;
 	spkey_t spk;
+	int rc = 0;
 
 	/* caller has done 'strlower()' to our input.. */
 	if (cp) *cp++ = 0;
@@ -85,7 +74,7 @@ const char *turnarg;
 	spl = sp_lookup(spk, spt_mesh[L_HOST]);
 	if (spl == NULL || spl->data == NULL) {
 	  /* Not found, nothing to do.. */
-	  return;
+	  return 0;
 	}
 	wp = (struct web *)spl->data;
 	
@@ -97,9 +86,10 @@ const char *turnarg;
 
 	  if (wp == ncuritem->whost && ncuritem->proc == NULL) {
 	    ncuritem->wakeup = 0; /* Force its starttime! */
-	    thread_start(ncuritem, 1);
+	    rc += thread_start(ncuritem, 1);
 	    /* We MAY get multiple matches, though it is unlikely.. */
 	  }
 	  ncuritem = nncuritem;
 	}
+	return rc;
 }   
