@@ -113,7 +113,7 @@ main(argc, argv)
 	progname = argv[0];
 	verbose = debug = errflg = status = user = onlyuser = summary = 0;
 	while (1) {
-	  c = getopt(argc, argv, "dip:r:stu:vVSQ");
+	  c = getopt(argc, argv, "dip:r:stu:U:vVSQ");
 	  if (c == EOF)
 	    break;
 	  switch (c) {
@@ -165,6 +165,22 @@ main(argc, argv)
 	  case 'Q':
 	    ++schedq;
 	    break;
+	  case 'U':
+	    v2username = optarg;
+	    v2password = strchr(v2username,'/');
+	    if (v2password) *v2password++ = 0;
+	    else {
+	      v2password = strchr(v2username,':');
+	      if (v2password) *v2password++ = 0;
+	      else {
+		v2password = strchr(v2username,',');
+		if (v2password) *v2password++ = 0;
+		else {
+		  v2password = "nobody";
+		}
+	      }
+	    }
+	    break;
 	  case 'V':
 	    prversion("mailq");
 	    exit(0);
@@ -419,6 +435,87 @@ main(argc, argv)
 	return 0;
 }
 
+
+/* Lifted from BIND res/res_debug.c */
+/*
+ * Return a mnemonic for a time to live
+ */
+char *
+saytime(value, buf, shortform)
+	long value;
+	char *buf;
+	int shortform;
+{
+	int secs, mins, hours, fields = 0;
+	register char *p;
+
+	p = buf;
+
+	while (*p) ++p;
+	if (value < 0) {
+	  *p++ = '-'; *p = 0;
+	  value = -value;
+	}
+
+	if (value == 0) {
+	  if (shortform)
+	    strcpy(p,"0s");
+	  else
+	    strcpy(p,"0 sec");
+	  return buf;
+	}
+
+	secs = value % 60;
+	value /= 60;
+	mins = value % 60;
+	value /= 60;
+	hours = value % 24;
+	value /= 24;
+
+#define	PLURALIZE(x)	x, (x == 1) ? "" : "s"
+	if (value) {
+	  if (shortform)
+	    sprintf(p, "%ldd", value);
+	  else
+	    sprintf(p, "%ld day%s", PLURALIZE(value));
+	  ++fields;
+	  while (*++p);
+	}
+	if (hours) {
+	  if (shortform)
+	    sprintf(p, "%dh", hours);
+	  else {
+	    if (value && p != buf)
+	      *p++ = ' ';
+	    sprintf(p, "%d hour%s", PLURALIZE(hours));
+	  }
+	  ++fields;
+	  while (*++p);
+	}
+	if (mins && fields < 2) {
+	  if (shortform)
+	    sprintf(p, "%dm", mins);
+	  else {
+	    if ((hours || value) && p != buf)
+	      *p++ = ' ';
+	    sprintf(p, "%d min%s", PLURALIZE(mins));
+	  }
+	  while (*++p);
+	}
+	if (secs && fields < 2) {
+	  if (shortform)
+	    sprintf(p, "%ds", secs);
+	  else {
+	    if ((mins || hours || value) && p != buf)
+	      *p++ = ' ';
+	    sprintf(p, "%d sec%s", PLURALIZE(secs));
+	  }
+	  while (*++p);
+	}
+	*p = '\0';
+	return buf;
+}
+
 void
 docat(file, fd)
 	const char *file;
@@ -448,6 +545,59 @@ docat(file, fd)
 	if (fpo) fclose(fpo);
 }
 
+int countfiles __((const char *));
+int countfiles(dirpath)
+const char *dirpath;
+{
+	char dpath[512];
+
+	struct dirent *dp;
+	DIR *dirp;
+	int n = 0;
+
+	dirp = opendir(dirpath);
+	if (dirp == NULL) {
+	  fprintf(stderr, "%s: opendir(%s): %s\n",
+		  progname, dpath, strerror(errno));
+	  return -1;
+	}
+	for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
+	  if (dp->d_name[0] == '.' &&
+	      (dp->d_name[1] == 0 || (dp->d_name[1] == '.' &&
+				      dp->d_name[2] == 0)))
+	    continue; /* . and .. */
+	  if (isascii(dp->d_name[0]) && isdigit(dp->d_name[0]))
+	    ++n;
+	  else if (strcmp("core",dp->d_name)==0)
+	    sawcore = 1, ++othern;
+	  else {
+	    if (dp->d_name[0] >= 'A' && dp->d_name[0] <= 'Z' &&
+		dp->d_name[1] == 0) {
+	      struct stat stbuf;
+	      sprintf(dpath, "%s/%s", dirpath, dp->d_name);
+	      if (lstat(dpath,&stbuf) != 0 ||
+		  !S_ISDIR(stbuf.st_mode)) {
+		++othern;
+	      } else {
+		n += countfiles(dpath);
+	      }
+	    } else {
+	      ++othern;
+	    }
+	  }
+	}
+#ifdef	BUGGY_CLOSEDIR
+	/*
+	 * Major serious bug time here;  some closedir()'s
+	 * free dirp before referring to dirp->dd_fd. GRRR.
+	 * XX: remove this when bug is eradicated from System V's.
+	 */
+	close(dirp->dd_fd);
+#endif
+	closedir(dirp);
+	return n;
+}
+
 /*
  * Determine if the Router is alive, how many entries are in the queue,
  * and whether the router dumped core last time it died.
@@ -465,25 +615,9 @@ checkrouter()
 	if (postoffice == NULL)
 	  return;
 	sprintf(path, "%s/%s", postoffice, ROUTERDIR);
-	dirp = opendir(path);
-	if (dirp == NULL) {
-	  fprintf(stderr, "%s: opendir(%s): %s\n",
-		  progname, path, strerror(errno));
-	  return;
-	}
-	for (dp = readdir(dirp), n = 0; dp != NULL; dp = readdir(dirp)) {
-	  if (isascii(dp->d_name[0]) && isdigit(dp->d_name[0]))
-	    ++n;
-	}
-#ifdef	BUGGY_CLOSEDIR
-	/*
-	 * Major serious bug time here;  some closedir()'s
-	 * free dirp before referring to dirp->dd_fd. GRRR.
-	 * XX: remove this when bug is eradicated from System V's.
-	 */
-	close(dirp->dd_fd);
-#endif
-	closedir(dirp);
+
+	n = countfiles(path);
+
 	fprintf(stdout,"%d entr%s in router queue: ", n, n != 1 ? "ies" : "y");
 
 	if (nonlocal)
@@ -542,59 +676,6 @@ checkrouter()
 	  fprintf(stdout,"%d message%s deferred\n", n, n != 1 ? "s" : "");
 }
 
-int countfiles __((const char *));
-int countfiles(dirpath)
-const char *dirpath;
-{
-	char dpath[512];
-
-	struct dirent *dp;
-	DIR *dirp;
-	int n = 0;
-
-	dirp = opendir(dirpath);
-	if (dirp == NULL) {
-	  fprintf(stderr, "%s: opendir(%s): %s\n",
-		  progname, dpath, strerror(errno));
-	  return -1;
-	}
-	for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
-	  if (dp->d_name[0] == '.' &&
-	      (dp->d_name[1] == 0 || (dp->d_name[1] == '.' &&
-				      dp->d_name[2] == 0)))
-	    continue; /* . and .. */
-	  if (isascii(dp->d_name[0]) && isdigit(dp->d_name[0]))
-	    ++n;
-	  else if (strcmp("core",dp->d_name)==0)
-	    sawcore = 1, ++othern;
-	  else {
-	    if (dp->d_name[0] >= 'A' && dp->d_name[0] <= 'Z' &&
-		dp->d_name[1] == 0) {
-	      struct stat stbuf;
-	      sprintf(dpath, "%s/%s", dirpath, dp->d_name);
-	      if (lstat(dpath,&stbuf) != 0 ||
-		  !S_ISDIR(stbuf.st_mode)) {
-		++othern;
-	      } else {
-		n += countfiles(dpath);
-	      }
-	    } else {
-	      ++othern;
-	    }
-	  }
-	}
-#ifdef	BUGGY_CLOSEDIR
-	/*
-	 * Major serious bug time here;  some closedir()'s
-	 * free dirp before referring to dirp->dd_fd. GRRR.
-	 * XX: remove this when bug is eradicated from System V's.
-	 */
-	close(dirp->dd_fd);
-#endif
-	closedir(dirp);
-	return n;
-}
-
 
 void
 checkscheduler()
@@ -1080,7 +1161,7 @@ void query2(fpi, fpo)
 	MD5_CTX CTX;
 	unsigned char digbuf[16];
 
-	/* XX: Authenticate the query - get challenge */
+	/* Authenticate the query - get challenge */
 	bufsize = 0;
 	if (GETLINE(challenge, bufsize, bufspace, fpi))
 	  return;
@@ -1146,7 +1227,142 @@ void query2(fpi, fpo)
 
 	  }
 	} else {
-	  fprintf(stdout,"Sorry, scheduler with protocol version 2 can't give results without -Q option\n");
+
+	  /* Non -Q* -mode processing */
+
+	  int linespace = 256;
+	  int linecnt = 0;
+	  char **lines = (char **) malloc(sizeof(char *) * linespace);
+
+	  fprintf(fpo, "SHOW QUEUE THREADS2\n");
+	  fflush(fpo);
+
+	  bufsize = 0;
+	  if (GETLINE(buf, bufsize, bufspace, fpi))
+	    return;
+
+	  if (*buf != '+') {
+	    fprintf(stdout,"Scheduler response: '%s'\n",buf);
+	    return;
+	  }
+
+	  for (;;) {
+	    char *b;
+	    bufsize = 0;
+	    if (GETLINE(buf, bufsize, bufspace, fpi))
+	      break;
+	    if (buf[0] == '.' && buf[1] == 0)
+	      break;
+
+	    if (linecnt+1 >= linespace) {
+	      linespace *= 2;
+	      lines = (char **)realloc((void**)lines,
+				       sizeof(char *) * linespace);
+	    }
+
+	    /* Do leading dot duplication suppression */
+	    b = buf;
+	    if (*b == '.') {
+	      --bufsize;
+	      ++b;
+	    }
+
+	    lines[linecnt] = malloc(bufsize+2);
+	    memcpy(lines[linecnt], b, bufsize+1);
+	    lines[++linecnt] = NULL;
+
+	    /* fprintf(stdout,"%s\n", b); */
+	  }
+
+	  for (i = 0; lines[i] != NULL; ++i) {
+	    char *channel = lines[i];
+	    char *host    = strchr(channel, '\t');
+	    char *rest    = "";
+	    char *b;
+
+	    if (host) {
+	      *host++ = 0;
+	      rest = strchr(host,'\t');
+	      if (rest) *rest++ = 0;
+	    } else host = "";
+
+	    fprintf(fpo, "SHOW THREAD %s %s\n",channel,host);
+	    fflush(fpo);
+
+	    bufsize = 0;
+	    if (GETLINE(buf, bufsize, bufspace, fpi))
+	      return;
+
+	    if (*buf != '+') {
+	      fprintf(stdout,"Scheduler response: '%s'\n",buf);
+	      return;
+	    }
+
+
+	    printf("%s/%s:\n",channel, host);
+
+	    for (;;) {
+	      int j;
+	      char *split[10];
+	      char timebuf[30];
+	      bufsize = 0;
+	      if (GETLINE(buf, bufsize, bufspace, fpi))
+		break;
+	      if (buf[0] == '.' && buf[1] == 0)
+		break;
+
+	      /* Do leading dot duplication suppression */
+	      b = buf;
+	      if (*b == '.') {
+		--bufsize;
+		++b;
+	      }
+
+	      /* Array elts:
+		 0) filepath under $POSTOFFICE/transport/
+		 1) error address in brackets
+		 2) recipient line offset within the control file
+		 3) message expiry time (time_t)
+		 4) next wakeup time (time_t)
+		 5) last feed time (time_t)
+		 6) count of attempts at the delivery
+		 7) "retry in NNN" or a pending on "channel"/"thread"
+		 8) possible diagnostic message from previous delivery attempt
+	      */
+
+	      for (j = 0; b && j < 9; ++j) {
+		split[j] = b;
+		b = strchr(b ? b : "", '\t');
+		if (b) *b++ = 0;
+	      }
+	      
+	      if (j != 9) {
+		fprintf(stderr,"Communication error! Malformed data entry!\n");
+		continue;
+	      }
+
+	      printf("\t%s: (", split[0]);
+
+	      *timebuf = 0;
+	      saytime((long)(atol(split[3]) - now), timebuf, 1);
+
+	      printf("%s tries, expires in %s", split[6], timebuf);
+
+	      printf(") %s\n", split[8]);
+
+	      if (verbose) {
+		/* XXX: Pick message file control data to present old-style
+		   verbose report */
+		/* XXX: printf("\t  id\t%s", ... );
+		   XXX: printf("...\n"); */
+		printf("\t  from\t%s\n", *split[1] ? split[1] : "<>");
+		/* XXX: TO line ... */
+	      }
+
+	    }
+
+	  }
+
 	}
 }
 
