@@ -4,7 +4,9 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1999
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2000
+ *
+ *	We produce RFC 1894 format reports with everything it contains.
  */
 
 #include "hostenv.h"
@@ -28,7 +30,7 @@ extern char *strnsave();
 #endif
 extern int errno;
 
-extern int never_full_content; /* at conf.c */
+extern int default_full_content; /* at conf.c */
 
 /*
  * There has to be some way of collecting the error messages produced during
@@ -41,6 +43,10 @@ extern int never_full_content; /* at conf.c */
  * respect (knock wood).
  */
 
+typedef enum { ACTSET_DELIVERED = 0, ACTSET_FAILED = 1,
+	       ACTSET_RELAYED   = 2, ACTSET_DELAYED = 3,
+	       ACTSET_NONE = 4 } ACTSETENUM;
+
 struct not {
 	char	    *not;
 	const char  *message;
@@ -49,6 +55,7 @@ struct not {
 	int          notifyflgs;
 	char	    *rcpntp;
 	time_t       tstamp;
+	ACTSETENUM   thisaction;
 };
 
 static void decodeXtext __((Sfio_t *, const char *));
@@ -79,6 +86,7 @@ scnotaryreport(errfp,notary,haserrsp,notifyrespectflg)
 	int notifyrespectflg;
 {
 	char *rcpt, *action, *status, *diagstr, *wtt;
+	char *taid;
 	char *cp;
 	const char *typetag;
 
@@ -115,9 +123,15 @@ scnotaryreport(errfp,notary,haserrsp,notifyrespectflg)
 	  *cp++  = 0;
 	  diagstr = cp;
 	  cp    = strchr(diagstr,'\001');
-	  if (cp) /* Uhh... Mal-formed input from the transporters ? */
-	    *cp++    = 0;
+	  if (cp) *cp++    = 0;
 	  wtt = cp;
+	  if (cp)  cp   = strchr(cp,'\001');
+	  if (cp) *cp++ = 0;
+	}
+	taid = cp;
+	if (taid) {
+	  cp = strchr(cp+1, '\001'); /* Chop it.. */
+	  if (cp) *cp++ = 0;
 	}
 
 	if (*rcpt == '\003' /* XX: MAGIC! */) {
@@ -162,13 +176,15 @@ scnotaryreport(errfp,notary,haserrsp,notifyrespectflg)
 	if (notary->tstamp != 0)
 	  sfprintf(errfp, "Last-Attempt-Date: %s",
 		   rfc822date(&notary->tstamp));
+	if (taid)
+	  sfprintf(errfp, "X-ZTAID: %s\n", taid);
 	sfprintf(errfp, "\n");
+
 	action[-1] = '\001';
-	if (status)
-	  status[-1] = '\001';
+	if (status) status[-1] = '\001';
 	diagstr[-1] = '\001';
-	if (wtt)
-	  wtt[-1]    = '\001';
+	if (wtt)  wtt[-1]  = '\001';
+	if (taid) taid[-1] = '\001';
 }
 
 
@@ -267,15 +283,16 @@ static void pick_env_addr(buf,mfp)
 }
 
 
-static void writeheader __((Sfio_t *, const char *, int *, const char *, const char *));
+static void writeheader __((Sfio_t *, const char *, int *, const char *, const char *, int *actionset));
 
 static void
-writeheader(errfp, eaddr, no_error_reportp, deliveryform, boundary)
+writeheader(errfp, eaddr, no_error_reportp, deliveryform, boundary, actionset)
      Sfio_t *errfp;
      const char *eaddr;
      int *no_error_reportp;
      const char *deliveryform;
      const char *boundary;
+     int *actionset;
 {
 	Sfio_t *fp;
 	char path[MAXPATHLEN];
@@ -321,26 +338,53 @@ writeheader(errfp, eaddr, no_error_reportp, deliveryform, boundary)
 	    } else if (strncmp(buf,"ADR",3)==0) {
 	      sfprintf(errfp, "%s", buf+4);
 	    } else if (strncmp(buf,"SUB",3)==0) {
+	      char *s = strchr(buf+4, '\n');
+	      if (s) *s = 0;
+	      s = buf+4;
 	      hadsubj = 1;
 	      if (*no_error_reportp < 0) {
 		/* We modify the subject string... */
-		char *s = buf+4;
 		while (*s && *s != ':') ++s;
 		if (*s == ':') {
 		  ++s;
 		  if (*s) *s++ = 0;
 		  sfprintf(errfp,"%s Double-fault: %s", buf+4, s);
 		} else {
-		  sfprintf(errfp,"Subject: Double-fault: unknown delivery error\n");
+		  sfprintf(errfp,"Subject: Double-fault: unknown delivery error");
 		}
 	      } else
 		sfprintf(errfp, "%s", buf+4);
+	      s = " [";
+	      if (actionset[ACTSET_FAILED]) {
+		sfprintf(errfp,"%sFAILED(%d)",s,
+			 actionset[ACTSET_FAILED]);
+		s = ",";
+	      }
+	      if (actionset[ACTSET_DELAYED]) {
+		sfprintf(errfp,"%sDELAYED(%d)",s,
+			 actionset[ACTSET_DELAYED]);
+		s = ",";
+	      }
+	      if (actionset[ACTSET_RELAYED]) {
+		sfprintf(errfp,"%sRELAYED-OK(%d)",s,
+			 actionset[ACTSET_RELAYED]);
+		s = ",";
+	      }
+	      if (actionset[ACTSET_DELIVERED]) {
+		sfprintf(errfp,"%sDELIVERED-OK(%d)",s,
+			 actionset[ACTSET_DELIVERED]);
+		s = ",";
+	      }
+	      s = (*s == ',') ? "]" : "";
+	      sfprintf(errfp,"%s\n", s);
 	    } else {
 	      if (inhdr) {
 		inhdr = 0;
 		sfprintf(errfp,"MIME-Version: 1.0\n");
 		sfprintf(errfp,"Content-Type: multipart/report; report-type=delivery-status;\n");
 		sfprintf(errfp,"\tboundary=\"%s\"\n\n",boundary);
+		sfprintf(errfp, "This is MULTIPART/REPORT structured message as defined at RFC 1894.\n\n");
+		sfprintf(errfp, "As your email client software vendor, when will they support this\nreport format by showing its formal part in your preferred language.\n\n");
 		sfprintf(errfp, "--%s\n", boundary);
 		sfprintf(errfp, "Content-Type: text/plain\n");
 	      }
@@ -376,13 +420,14 @@ writeheader(errfp, eaddr, no_error_reportp, deliveryform, boundary)
 /* called to process errors at sensible intervals */
 
 void
-reporterrs(cfpi)
+reporterrs(cfpi, delayreports)
 	struct ctlfile *cfpi;
+	const int delayreports;
 {
 	int i, n, wroteheader, byteidx, headeridx = -1, drptidx, fd;
 	int *lp;
 	time_t tstamp;
-	char *midbuf, *cp, *eaddr;
+	char *midbuf, *cp, *cp2, *action, *eaddr;
 	char *deliveryform;
 	Sfio_t *errfp, *fp;
 	char *notary;
@@ -399,6 +444,9 @@ reporterrs(cfpi)
 	int mypid;
 	long format;
 	char boundarystr[400];
+	char spoolid[30];
+	int actionsets[4]; /* 0:DELIVERED, 1:FAILED, 2:RELAYED, 3:DELAYED */
+	ACTSETENUM thisaction;
 
 	if (cfpi->haderror == 0)
 		return;
@@ -406,6 +454,8 @@ reporterrs(cfpi)
 		return;
 	no_error_report = cfpi->iserrmesg;
 
+	taspoolid(spoolid, cfpi->mtime, cfpi->id);
+
 	eaddr        = cfpi->erroraddr;
 	deliveryform = cfpi->deliveryform;
 	envid        = cfpi->envid;
@@ -435,6 +485,7 @@ reporterrs(cfpi)
 	wroteheader = 0;
 	lastoffset = cfp->offset[cfp->nlines-1];
 	format = 0L;
+	actionsets[0] = actionsets[1] = actionsets[2] = actionsets[3] = 0;
 	for (i = 0; i < cfp->nlines; ++i, ++lp) {
 	  cp = cfp->contents + *lp;
 
@@ -444,6 +495,9 @@ reporterrs(cfpi)
 	    continue;
 	  }
 
+	  /* FIXME: FIXME: DELAYED reporting must check entries:
+	     _CF_RECIPIENT && _CFTAG_NORMAL */
+
 	  if (!(*cp == _CF_DIAGNOSTIC && *++cp == _CFTAG_NORMAL))
 	    continue;
 
@@ -492,12 +546,33 @@ reporterrs(cfpi)
 	  /* If ever need to add more integer offsets, add them
 	     here with a colon prefix... */
 	  notary = NULL;
+	  action = NULL;
+	  /* If this is nothing, this is about FAILURE */
+	  thisaction = ACTSET_FAILED;
 	  if (*cp == '\t') {
 	    notary = ++cp;
 	    while (*cp && *cp != '\t') ++cp;
 	    notary = strnsave(notary, cp-notary); /* Make a copy of it */
 	    ++cp;
+	    action = strchr(notary,'\001');
+
+	    if (action) {
+	      ++action;
+	      cp2 = strchr(action,'\001');
+
+	      if (memcmp       (action,"delayed",  7)==0) {
+		thisaction = ACTSET_DELAYED;
+	      } else if (memcmp(action,"delivered",9)==0) {
+		thisaction = ACTSET_DELIVERED;
+	      } else if (memcmp(action,"relayed",  7)==0) {
+		thisaction = ACTSET_RELAYED;
+	      } else if (memcmp(action,"failed",   6)==0) {
+		thisaction = ACTSET_FAILED;
+	      } else
+		thisaction = ACTSET_NONE;
+	    }
 	  }
+
 	  rcpntpointer = cfp->contents + byteidx + 2 + _CFTAG_RCPTPIDSIZE;
 	  if (format & _CF_FORMAT_DELAY1)
 	    rcpntpointer += _CFTAG_RCPTDELAYSIZE;
@@ -515,13 +590,14 @@ reporterrs(cfpi)
 	      notaries = (void*)erealloc((void*)notaries,
 					 sizeof(struct not)*(notaryspc+1));
 	    }
-	    notaries[notarycnt].not     = notary;
-	    notaries[notarycnt].orcpt   = NULL;
-	    notaries[notarycnt].notify  = NULL;
-	    notaries[notarycnt].message = NULL;
-	    notaries[notarycnt].notifyflgs = 0;
-	    notaries[notarycnt].rcpntp  = rcpntpointer;
-	    notaries[notarycnt].tstamp  = tstamp;
+	    notaries[notarycnt].not        = notary;
+	    notaries[notarycnt].orcpt      = NULL;
+	    notaries[notarycnt].notify     = NULL;
+	    notaries[notarycnt].message    = NULL;
+	    /* If NOTIFY= is not defined, default is: NOTIFY=FAILURE */
+	    notaries[notarycnt].notifyflgs = NOT_FAILURE;
+	    notaries[notarycnt].rcpntp     = rcpntpointer;
+	    notaries[notarycnt].tstamp     = tstamp;
 	    if (drptidx > 0) {
 	      char *d = cfp->contents + drptidx;
 	      while (*d) {
@@ -536,6 +612,7 @@ reporterrs(cfpi)
 		if (CISTREQN(d,"NOTIFY=",7)) {
 		  char *p;
 		  notaries[notarycnt].notify = d+7;
+		  notaries[notarycnt].notifyflgs = 0;
 		  d += 7;
 		  p = d;
 		  while (*d != 0 && (*d != ' ' && *d != '\t')) ++d;
@@ -566,11 +643,35 @@ reporterrs(cfpi)
 	      }
 	      /* Report conditional testing in the latter phases */
 	    }
-	    while (isascii(*cp) && isspace(*cp))
-	      ++cp;
+	    while (*cp == ' ' || *cp == '\t') ++cp;
 	    notaries[notarycnt].message = cp;
+
+	    switch (thisaction) {
+	    case ACTSET_DELIVERED:
+	    case ACTSET_RELAYED:
+	      if (notaries[notarycnt].notifyflgs & NOT_SUCCESS)
+		actionsets[thisaction] += 1;
+	      else thisaction = ACTSET_NONE;
+	      break;
+	    case ACTSET_DELAYED:
+	      if (notaries[notarycnt].notifyflgs & NOT_DELAY)
+		actionsets[thisaction] += 1;
+	      else thisaction = ACTSET_NONE;
+	      break;
+	    case ACTSET_FAILED:
+	      if (notaries[notarycnt].notifyflgs & NOT_FAILURE)
+		actionsets[thisaction] += 1;
+	      else thisaction = ACTSET_NONE;
+	      break;
+	    default:
+	      thisaction = ACTSET_NONE;
+	      break;
+	    }
+	    notaries[notarycnt].thisaction = thisaction;
+
 	    ++notarycnt;
 	    notaries[notarycnt].not = NULL;
+
 	  } else {
 	    /* No notaries ?!  What ?  Store the CP pointer anyway.. */
 	    /* End of failure list processing */
@@ -588,6 +689,24 @@ reporterrs(cfpi)
 	  return;
 	}
 
+	if (!(actionsets[ACTSET_FAILED]   |
+	      actionsets[ACTSET_RELAYED]  |
+	      actionsets[ACTSET_DELAYED]  |
+	      actionsets[ACTSET_DELIVERED] )) {
+
+	  /* No reports what so ever ? */
+
+	  /* Release 'notaries' datasets */
+	  for (i = 0; i < notarycnt; ++i) {
+	    if (notaries[i].not != NULL)
+	      free(notaries[i].not);
+	  }
+	  free((void*)notaries);
+	  close(cfp->fd);
+	  free_cfp_memory(cfp);
+	  return;
+	}
+
 	errfp = sfmail_open(MSG_RFC822);
 	if (errfp == NULL) {
 	  sfprintf(sfstderr,
@@ -616,24 +735,45 @@ reporterrs(cfpi)
 	  strcat(boundarystr, dom);
 	}
 
-	writeheader(errfp, eaddr, &no_error_report, deliveryform, boundarystr);
-
-
+	writeheader(errfp, eaddr, &no_error_report, deliveryform, boundarystr,
+		    actionsets);
 
+	n = 0;
 	for (i = 0; i < notarycnt; ++i) {
 	  /* Scan to the start of the message text */
 	  const char *ccp, *s;
 	  if ((notaries[i].notifyflgs & NOT_NEVER) && (no_error_report >= 0))
 	    continue;
 	  /* Report is not outright rejected, or this is double fault */
-	  sfprintf(errfp, "<%s>: ", notaries[i].rcpntp);
+	  ++n;
+	  switch (notaries[i].thisaction) {
+	  case ACTSET_DELAYED:
+	    sfprintf(errfp,"DELAYED (still in queue):\n");
+	    break;
+	  case ACTSET_RELAYED:
+	    sfprintf(errfp,"RELAYED (into system not supporting DSN facility):\n");
+	    break;
+	  case ACTSET_FAILED:
+	    sfprintf(errfp,"FAILED:\n");
+	    break;
+	  case ACTSET_DELIVERED:
+	    sfprintf(errfp,"DELIVERED (successfully):\n");
+	    break;
+	  case ACTSET_NONE:
+	    sfprintf(errfp,"BUG (unknown ACTSET value: %d):\n",
+		     notaries[i].thisaction);
+	    --n;
+	    break;
+	  }
+	  sfprintf(errfp, "  <%s>: ", notaries[i].rcpntp);
 	  ccp = notaries[i].message;
 	  if (strchr(ccp, '\r')) {
-	    sfprintf(errfp, "...\\\n\t");
-	    for (s = ccp; *s != '\0'; ++s) {
+	    sfprintf(errfp, "...\\\n        ");
+	    s = ccp;
+	    if (*s == '\r') ++s; /* Skip possible first '\r' */
+	    for (; *s != '\0'; ++s) {
 	      if (*s == '\r') {
-		sfputc(errfp, '\n');
-		sfputc(errfp, '\t');
+		sfprintf(errfp,"\n        ");
 	      } else {
 		sfputc(errfp, *s);
 	      }
@@ -644,7 +784,6 @@ reporterrs(cfpi)
 	    sfprintf(errfp, "%s\n", ccp);
 	}
 
-
 	sfprintf(errfp,"\n\
 Following is a copy of MESSAGE/DELIVERY-STATUS format section below.\n\
 It is copied here in case your email client is unable to show it to you.\n\
@@ -653,8 +792,8 @@ assist automatic, and accurate presentation and usage of said information.\n\
 In case you need human assistance from the Postmaster(s) of the system which\n\
 sent you this report, please include this information in your question!\n\
 \n\
-\tVirtually Yours,\n\
-\t\tAutomatic Email delivery Software\n\
+    Virtually Yours,\n\
+        Automatic Email Delivery Software\n\
 \n");
 
 
@@ -670,12 +809,12 @@ sent you this report, please include this information in your question!\n\
 	}
 	/* rfc822date() returns a string with trailing newline! */
 	sfprintf(errfp, "Arrival-Date: %s", rfc822date(&cfpi->mtime));
+	sfprintf(errfp, "Local-Spool-ID: %s\n", spoolid);
 	sfprintf(errfp, "\n");
 
 	/* Now scan 'em all again for IETF-NOTARY */
-	for (i = 0; i < notarycnt; ++i) {
+	for (i = 0; i < notarycnt; ++i)
 	  scnotaryreport(errfp, &notaries[i],&has_errors,no_error_report);
-	}
 
 	sfprintf(errfp,"\n\
 Following is copy of the message headers. Original message content may\n\
@@ -738,6 +877,7 @@ be in subsequent parts of this MESSAGE/DELIVERY-STATUS structure.\n\n");
 	}
 	/* rfc822date() returns a string with trailing newline! */
 	sfprintf(errfp, "Arrival-Date: %s", rfc822date(&cfpi->mtime));
+	sfprintf(errfp, "Local-Spool-ID: %s\n", spoolid);
 	sfprintf(errfp, "\n");
 
 	/* Now scan 'em all again for IETF-NOTARY */
@@ -785,8 +925,8 @@ be in subsequent parts of this MESSAGE/DELIVERY-STATUS structure.\n\n");
 	  }
 	  if (has_errors &&
 	      ((no_error_report < 0) ||
-	       (!never_full_content &&
-		(!cfp->dsnretmode || CISTREQN(cfp->dsnretmode,"FULL",4))))) {
+	       (default_full_content && !cfp->dsnretmode) ||
+	       (cfp->dsnretmode && CISTREQN(cfp->dsnretmode,"FULL",4)))) {
 	    /* Copy out the rest (=body) with somewhat more efficient method */
 
 	    while ((n = sfread(fp, buf, sizeof buf)) > 0)
@@ -808,6 +948,185 @@ be in subsequent parts of this MESSAGE/DELIVERY-STATUS structure.\n\n");
 	free_cfp_memory(cfp);
 }
 
+
+/* ---------------- DELAYED reporter --------------------------*/
+
+void
+msgdelayed(vp, offset, notary, message)
+	struct vertex *vp;
+	long offset;
+	char *notary;
+	const char *message;
+{
+	Sfio_t *fp;
+	char path[410];
+	char *not[20], *s;
+	int i;
+
+	/* Split the notary string into components. */
+	for (i = 0; i < 20; ++i) not[i] = NULL;
+	s = notary;
+	i = 0;
+	while (i < 20 && s) {
+	  not[i++] = s;
+	  s = strchr(s, '\001');
+	  if (s) *s++ = 0;
+	}
+
+	if (vp->cfp->dirind > 0) {
+	  sprintf(path, "%.300s/%.100s",
+		  cfpdirname(vp->cfp->dirind), vp->cfp->mid);
+	} else {
+	  strncpy(path,vp->cfp->mid,400);
+	  path[400] = 0;
+	}
+
+	/* exclusive access required, but we're the only scheduler... */
+	fp = sfopen(NULL, path, "a");
+	if (fp == NULL) {
+	  sfprintf(sfstderr,
+		   "Cannot open control file %s to deposit", vp->cfp->mid);
+	  sfprintf(sfstderr,
+		   " error message for offset %ld:\n", offset);
+	  sfprintf(sfstderr, "\t%s\n", message);
+	  return;
+	}
+	vp->cfp->haderror = 1;
+	sfprintf(fp, "%c%c%ld:%ld:%ld::%ld\t",
+		 _CF_DIAGNOSTIC, _CFTAG_NORMAL, offset,
+		 (long)vp->headeroffset, (long)vp->drptoffset,
+		 time(NULL));
+	s = "";
+
+	/* Mark this DELAYED */
+	if (not[1]) {
+	  strcpy(path,"_delayed");
+	  not[1] = path+1;
+	}
+
+	for (i = 0; i < 20 && not[i]; ++i) {
+	  sfprintf(fp, "%s%s", s, not[i]);
+	  s = "\001";
+	  if (i > 0) not[i][-1] = '\001';
+	}
+	sfprintf(fp, "%s", notary);
+	sfprintf(fp, "\t%s\n", message);
+	sfsync(fp);
+#ifdef HAVE_FSYNC
+	fsync(sffileno(fp));
+#endif
+	sfclose(fp);
+}
+
+
+
+static void delayaux(vp, index, notary, buf)
+	struct vertex *vp;
+	int index;
+	char *notary;
+	const char *buf;
+{
+	int i;
+
+	/* Mark the delay info in.. */
+	for (i = 0 ; i < vp->ngroup; ++i)
+	  if (vp->index[i] == index) {
+	    msgdelayed(vp, vp->cfp->offset[index], notary, buf);
+	    break;
+	  }
+}
+
+/* FIXME: FIXME:  delayreport() calling, report times
+   control and tracking, etc... */
+
+void
+delayreport(vp)
+	struct vertex *vp;
+{
+	int i;
+	char *emsg;
+	char buf[BUFSIZ];
+	char notbuf[BUFSIZ];
+	const char *fmt = "\r%s, problem was:\r%s";
+	char *notary = vp->notary;
+
+	if (vp->nextdlyrprttime > now) return; /* Nothing yet! */
+
+	if (notary == NULL) {
+	  /* addres / action / status / diagnostic / wtt */
+	  sprintf(notbuf, "%s\001%s\001%s\001%s",
+		  "\003", /* XX: recipient address! XX: MAGIC INFO! */
+		  "delayed",
+		  "4.4.7 (no attempt done yet; long queue?)",
+		  "smtp; 400 (no delivery attempt done in ");
+	  saytime((u_long)(vp->nextdlyrprttime - vp->cfp->mtime), notbuf, 0);
+	  strcat(notbuf,")\001");
+	  notary = notbuf;
+	}
+
+	strcpy(buf, "delayed; no successfull delivery in ");
+	saytime((u_long)(vp->nextdlyrprttime - vp->cfp->mtime), buf, 0);
+
+	if (vp->message != NULL && *(vp->message) != '\0') {
+	  emsg = emalloc(strlen(buf) + strlen(vp->message) + strlen(fmt));
+	  sprintf(emsg, fmt, buf, vp->message);
+	} else
+	  emsg = buf;
+
+	/* Report all vertices having had a delay.. */
+	for (i = vp->ngroup -1; i >= 0; --i)
+	  delayaux(vp, vp->index[i], notary, emsg);
+
+	if (emsg != buf)
+	  free(emsg);
+}
+
+
+/* FIXME: reporting subsystem needs tuning! */
+/*        ... implementing/tuning DELAYED reports! */
+
+static int ctl_report_1 __((struct spblk *spl));
+static int ctl_report_1(spl)
+struct spblk *spl;
+{
+	struct vertex *vp, *nvp;
+	struct ctlfile * cfp = (struct ctlfile *)spl->data;
+	int doreport = 0;
+
+	queryipccheck(); /* Sets 'now', among other things */
+
+	for (vp = cfp->head; vp != NULL; vp = nvp) {
+	  nvp = vp->next[L_CTLFILE];
+
+	  /* We essentially report accumulated reports every
+	     ``global_report_interval'' seconds (2h by default) */
+
+	  if (vp->nextrprttime <= now)
+	    doreport = 1;
+	}
+
+	if (doreport) {
+	  reporterrs(cfp, 0);
+
+	  for (vp = cfp->head; vp != NULL; vp = nvp) {
+
+	    nvp = vp->next[L_CTLFILE];
+	    vp->nextrprttime = now + global_report_interval;
+
+	  }
+	}
+
+	return 0;
+}
+
+
+void
+interim_report_run __((void))
+{
+	sp_scan(ctl_report_1, NULL, spt_mesh[L_CTLFILE]);
+}
+
+
 /*
 char *
 mail_alloc(n)
