@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1999
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2000
  */
 
 #include "hostenv.h"
@@ -24,7 +24,6 @@
 
 static void celink __((struct config_entry *, struct config_entry **, struct config_entry **));
 static int readtoken __((Sfio_t *fp, char *buf, int buflen, int *linenump));
-static u_int parse_intvl __((char *string));
 static int paramparse __((char *line));
 
 #define RCKEYARGS __((char *key, char *arg, struct config_entry *ce))
@@ -40,11 +39,13 @@ static int rc_maxta		RCKEYARGS;
 static int rc_maxthr		RCKEYARGS;
 static int rc_idlemax		RCKEYARGS;
 static int rc_retries		RCKEYARGS;
+static int rc_reporttimes	RCKEYARGS;
 static int rc_user		RCKEYARGS;
 static int rc_skew		RCKEYARGS;
 static int rc_bychannel		RCKEYARGS;
 static int rc_ageorder		RCKEYARGS;
 static int rc_queueonly		RCKEYARGS;
+static int rc_wakeuprestartonly	RCKEYARGS;
 static int rc_deliveryform	RCKEYARGS;
 static int rc_overfeed		RCKEYARGS;
 static int rc_priority		RCKEYARGS;
@@ -77,6 +78,7 @@ static struct rckeyword {
 {	"expiryform",		rc_expform	},	/* string */
 {	"deliveryform",		rc_deliveryform	},	/* string */
 {	"retries",		rc_retries	},	/* array of numbers */
+{	"reporttimes",		rc_reporttimes	},	/* array of numbers */
 {	"maxring",		rc_maxring	},	/* number */
 {	"maxrings",		rc_maxring	},	/* number */
 {	"maxta",		rc_maxta	},	/* number */
@@ -89,6 +91,7 @@ static struct rckeyword {
 {	"bychannel",		rc_bychannel	},	/* boolean */
 {	"ageorder",		rc_ageorder	},	/* boolean */
 {	"queueonly",		rc_queueonly	},	/* boolean */
+{	"wakeuprestartonly",	rc_wakeuprestartonly },	/* boolean */
 {	"overfeed",		rc_overfeed	},	/* number */
 {	"priority",		rc_priority	},	/* number */
 {	"nice",			rc_nice		},	/* number */
@@ -125,6 +128,10 @@ defaultconfigentry(ce,defaults)
 	  ce->argv		= defaults->argv;
 	  ce->nretries		= defaults->nretries;
 	  ce->retries		= defaults->retries;
+	  ce->reporttimes[0]	= defaults->reporttimes[0];
+	  ce->reporttimes[1]	= defaults->reporttimes[1];
+	  ce->reporttimes[2]	= defaults->reporttimes[2];
+	  ce->reporttimes[3]	= defaults->reporttimes[3];
 	  ce->skew		= defaults->skew;
 	  ce->deliveryform	= defaults->deliveryform;
 	  ce->overfeed		= defaults->overfeed;
@@ -149,6 +156,10 @@ defaultconfigentry(ce,defaults)
 	  ce->argv	= NULL;
 	  ce->nretries	= 0;
 	  ce->retries	= NULL;
+	  ce->reporttimes[0] = 0;
+	  ce->reporttimes[1] = 0;
+	  ce->reporttimes[2] = 0;
+	  ce->reporttimes[3] = 0;
 	  ce->skew	= 5;
 	  ce->deliveryform = NULL;
 	  ce->overfeed	= 0;
@@ -181,9 +192,11 @@ vtxprint(vp)
 	if (ce->flags == 0)
 	  sfprintf(sfstdout," (none)");
 	else {
-	  if (ce->flags & CFG_BYCHANNEL) sfprintf(sfstdout," BYCHANNEL");
 	  if (ce->flags & CFG_WITHHOST)  sfprintf(sfstdout," WITHHOST");
 	  if (ce->flags & CFG_AGEORDER)  sfprintf(sfstdout," AGEORDER");
+	  if (ce->flags & CFG_QUEUEONLY) sfprintf(sfstdout," QUEUEONLY");
+	  if (ce->flags & CFG_WAKEUPRESTARTONLY)sfprintf(sfstdout,
+							 " WAKEUPRESTARTONLY");
 	}
 	sfprintf(sfstdout,"\n");
 	sfprintf(sfstdout,"\tmaxkids %d\n",	ce->maxkids);
@@ -201,6 +214,7 @@ vtxprint(vp)
 	  for (i = 0; ce->argv[i] != NULL; ++i)
 	    sfprintf(sfstdout,"\targv[%d] = %s\n", i, ce->argv[i]);
 	}
+
 	sfprintf(sfstdout,"\tnretries %d\n", ce->nretries);
 	if (ce->nretries > 0) {
 	  sfprintf(sfstdout,"\tretries = (");
@@ -212,6 +226,15 @@ vtxprint(vp)
 	  }
 	  sfprintf(sfstdout,")\n");
 	}
+
+	sfprintf(sfstdout,"\treporttimes = (");
+	for (i = 0; i < 4 ; ++i ) {
+	  sfprintf(sfstdout,"%d", ce->reporttimes[i]);
+	  if (i < 3)
+	    sfprintf(sfstdout," ");
+	}
+	sfprintf(sfstdout,")\n");
+
 	sfprintf(sfstdout,"\tskew %d\n", ce->skew);
 }
 
@@ -389,7 +412,7 @@ redo_readtoken:
 	  lp = line;
 	}
 	/* Skip initial white-space */
-	while (*lp != '\0' && isspace(*lp))
+	while (*lp != '\0' && isspace(0xFF & *lp))
 	  ++lp;
 	/* Now it is one of: a token, a comment start, or end of line */
 	if (*lp == '\0' || *lp == '#') {
@@ -399,12 +422,12 @@ redo_readtoken:
 	}
 	/* Now we scan for the token + possible value */
 	elp = lp;
-	while (*elp && !isspace(*elp) && *elp != '=' && *elp != '#')
+	while (*elp && !isspace(0xFF & *elp) && *elp != '=' && *elp != '#')
 	  ++elp;
-	if (isspace(*elp)) {
+	if (isspace(0xFF & *elp)) {
 	  /* Allow spaces after the token and before '=' */
 	  char *p = elp;
-	  while (*p && isspace(*p)) ++p;
+	  while (*p && isspace(0xFF & *p)) ++p;
 	  if (*p == '=')
 	    elp = p;
 	}
@@ -412,7 +435,7 @@ redo_readtoken:
 	if (*elp == '=') {
 	  /* Allow spaces between '=', and value */
 	  ++elp;
-	  while (*elp && isspace(*elp)) ++elp;
+	  while (*elp && isspace(0xFF & *elp)) ++elp;
 	  if (*elp == '"') {
 	    ++elp;
 	    while (*elp != '"' && *elp != '\0') {
@@ -434,7 +457,7 @@ redo_readtoken:
 	    }
 	    ++elp;
 	  } else {
-	    while (*elp && !isspace(*elp) && *elp != '"')
+	    while (*elp && !isspace(0xFF & *elp) && *elp != '"')
 	      ++elp;
 	  }
 	}
@@ -448,37 +471,6 @@ redo_readtoken:
 	return rv;
 }
 
-static u_int
-parse_intvl(string)
-	char *string;
-{
-	u_int	intvl;
-	int	val;
-
-	for (intvl = 0; *string != '\0'; ++string) {
-	  while (isascii(*string) && !isdigit(*string))
-	    ++string;
-	  val = atoi(string);
-	  while (isascii(*string) && isdigit(*string))
-	    ++string;
-	  switch (*string) {
-	    case 'd':		/* days */
-	      val *= 24*60*60;
-	      break;
-	    case 'h':		/* hours */
-	      val *= 3600;
-	      break;
-	    case 'm':		/* minutes */
-	      val *= 60;
-	      break;
-	    case 's':		/* seconds */
-	      /* val *= 1; */
-	      break;
-	  }
-	  intvl += val;
-	}
-	return intvl;
-}
 
 struct config_entry *
 rereadconfig(head, file)
@@ -550,12 +542,12 @@ static int rc_command(key, arg, ce)
 	  argv[j++] = cp;
 	  if (j >= (sizeof argv)/(sizeof argv[0]))
 	    break;
-	  while (*cp != '\0' && isascii(*cp) && !isspace(*cp))
+	  while (*cp != '\0' && !isspace(0xFF & *cp))
 	    ++cp;
 	  if (*cp == '\0')
 	    break;
 	  *cp++ = '\0';
-	  while (*cp != '\0' && isascii(*cp) && isspace(*cp))
+	  while (*cp != '\0' && isspace(0xFF & *cp))
 	    ++cp;
 	}
 	argv[j++] = NULL;
@@ -570,17 +562,10 @@ static int rc_command(key, arg, ce)
 	      break;
 	    }
 	}
-	if (!(ce->flags & CFG_BYCHANNEL)) {
-	  for (av = &ce->argv[0]; *av != NULL; ++av)
-	    if (strcmp(*av, replchannel) == 0) {
-	      ce->flags |= CFG_BYCHANNEL;
-	      if (ce->channel && strtok(ce->channel,"[{*?") != NULL)
-		ce->flags |= CFG_WITHHOST; /* Well, sort of..
-					      Channel-part is wild-card, thus
-					      it is also very restrictive.. */
-	      break;
-	    }
-	}
+
+	/* ``replchannel'' need not matched, idle processing
+	   takes it properly into account. */
+
 	return 0;
 }
 
@@ -596,7 +581,7 @@ static int rc_expiry(key, arg, ce)
 	char *key, *arg;
 	struct config_entry *ce;
 {
-	ce->expiry = parse_intvl(arg);
+	ce->expiry = parse_interval(arg,NULL);
 	return 0;
 }
 
@@ -620,7 +605,7 @@ static int rc_interval(key, arg, ce)
 	char *key, *arg;
 	struct config_entry *ce;
 {
-	ce->interval = parse_intvl(arg);
+	ce->interval = parse_interval(arg,NULL);
 	return 0;
 }
 
@@ -628,7 +613,7 @@ static int rc_idlemax(key, arg, ce)
 	char *key, *arg;
 	struct config_entry *ce;
 {
-	ce->idlemax = parse_intvl(arg);
+	ce->idlemax = parse_interval(arg,NULL);
 	return 0;
 }
 
@@ -704,7 +689,7 @@ static int rc_maxchannel(key, arg, ce)
 {
 	ce->maxkidChannel = atoi(arg);
 	if (ce->maxkidChannel <= 0)
-	  ce->maxkidChannel = 1000;
+	  ce->maxkidChannel = 10000;
 	return 0;
 }
 
@@ -714,7 +699,7 @@ static int rc_maxring(key, arg, ce)
 {
 	ce->maxkidThreads = atoi(arg);
 	if (ce->maxkidThreads <= 0)
-	  ce->maxkidThreads = 1000;
+	  ce->maxkidThreads = 10000;
 	return 0;
 }
 
@@ -724,7 +709,7 @@ static int rc_maxta(key, arg, ce)
 {
 	ce->maxkids = atoi(arg);
 	if (ce->maxkids <= 0)
-	  ce->maxkids = 1000;
+	  ce->maxkids = 10000;
 	if (ce->maxkids > global_maxkids)
 	  ce->maxkids = global_maxkids;
 	return 0;
@@ -749,12 +734,12 @@ static int rc_retries(key, arg, ce)
 
 	j = 0;
 	for (cp = arg; *cp != '\0'; ++cp) {
-	  while (*cp != '\0' && isspace(*cp))
+	  while (*cp != '\0' && isspace(0xFF & *cp))
 	    ++cp;
 	  if (*cp == '\0')
 	    break;
 	  d = cp++;
-	  while (*cp != '\0' && !isspace(*cp))
+	  while (*cp != '\0' && !isspace(0xFF & *cp))
 	    ++cp;
 	  c = *cp;
 	  *cp = '\0';
@@ -784,6 +769,48 @@ static int rc_retries(key, arg, ce)
 	return 0;
 }
 
+static int rc_reporttimes(key, arg, ce)
+	char *key, *arg;
+	struct config_entry *ce;
+{
+	int i, j, arr[_CFTAG_RCPTDELAYSIZE];
+	char c, *cp, *d;
+
+	for (j = 0; j < sizeof(arr)/(sizeof(arr[0])); ++j)  arr[j] = 0;
+
+	j = 0;
+	for (cp = arg; *cp != '\0'; ++cp) {
+	  while (*cp != '\0' && isspace(0xFF & *cp))
+	    ++cp;
+	  if (*cp == '\0')
+	    break;
+	  d = cp++;
+	  while (*cp != '\0' && !isspace(0xFF & *cp))
+	    ++cp;
+	  c = *cp;
+	  *cp = '\0';
+	  i = parse_interval(d, NULL);
+	  if (i > 0)
+	    arr[j++] = i;
+	  else {
+	    sfprintf(sfstderr,
+		    "%s: not a numeric factor: %s\n",
+		    progname, d);
+	    return 1;
+	  }
+	  if (j >= (sizeof arr)/(sizeof arr[0]))
+	    break;
+	  *cp = c;
+	  if (*cp == '\0')
+	    break;
+	}
+
+	for (j = 0; j < sizeof(arr)/(sizeof(arr[0])); ++j)
+	  ce->reporttimes[j] = arr[j];
+
+	return 0;
+}
+
 static int rc_user(key, arg, ce)
 	char *key, *arg;
 	struct config_entry *ce;
@@ -818,7 +845,6 @@ static int rc_bychannel(key, arg, ce)
 	char *key, *arg;
 	struct config_entry *ce;
 {
-	ce->flags |= CFG_BYCHANNEL;
 	return 0;
 }
 
@@ -846,6 +872,14 @@ static int rc_queueonly(key, arg, ce)
 	return 0;
 }
 
+static int rc_wakeuprestartonly(key, arg, ce)
+	char *key, *arg;
+	struct config_entry *ce;
+{
+	ce->flags |= CFG_WAKEUPRESTARTONLY;
+	return 0;
+}
+
 extern int mailqmode;
 
 static int paramparse(line)
@@ -890,5 +924,11 @@ static int paramparse(line)
 	  mailqsock = strsave(a);
 	  return 0;
 	}
+
+	if (cistrcmp(line,"global-report-interval")==0 && a) {
+	  global_report_interval = parse_interval(a, NULL);
+	  return 0;
+	}
+
 	return 1;
 }
