@@ -4,7 +4,7 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1999
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2000
  */
 
 /*
@@ -17,7 +17,6 @@
 #include "hostenv.h"
 #include <ctype.h>
 #include <errno.h>
-#include "scheduler.h"
 #include <sys/stat.h>
 #include <fcntl.h>
 #include "mail.h"
@@ -49,6 +48,7 @@
 #include <sys/mman.h>
 #endif
 
+#include "scheduler.h"
 #include "prototypes.h"
 #include "zsyslog.h"
 #include "libz.h"
@@ -129,9 +129,10 @@ static int vtxprep_skip_any  = 0;
 static int vtxprep_skip_lock = 0;
 static time_t next_dirscan     = 0;
 static time_t next_idlecleanup = 0;
+static time_t next_interim_report_time = 0;
 static struct sptree *dirscan_mesh = NULL;
 static int newents_limit = 400;
-extern int never_full_content; /* on conf.c */
+extern int default_full_content; /* on conf.c */
 
 #include "memtypes.h"
 extern memtypes stickymem;
@@ -230,7 +231,7 @@ struct dirqueue {
 };
 
 static int dirqueuescan __((const char *dir,struct dirqueue *dq, int subdirs));
-static int syncweb __((struct dirqueue *dq));
+int syncweb __((struct dirqueue *dq));
 
 int global_maxkids = 1000;
 time_t now;
@@ -290,7 +291,7 @@ struct ctlfile *cfp;
 	for (vp = cfp->head; vp != NULL; vp = nvp) {
 	  /* Stash the next pointer, because the CFP content
 	     may become trashed at the end of the run. */
-	  nvp = cfp->head->next[L_CTLFILE];
+	  nvp = vp->next[L_CTLFILE];
 
 	  MIBMtaEntry->mtaStoredRecipients     -= vp->ngroup;
 	  MIBMtaEntry->mtaReceivedRecipientsSc -= vp->ngroup;
@@ -340,6 +341,9 @@ struct ctlfile *cfp;
 	if (cfp->erroraddr)	free(cfp->erroraddr);
 	if (cfp->logident)	free(cfp->logident);
 	if (cfp->envid)		free(cfp->envid);
+
+memset(cfp, 0x55, sizeof(*cfp));
+
 	free((char *)cfp);
 }
 
@@ -515,10 +519,31 @@ main(argc, argv)
 
 	resources_maximize_nofiles();
 
-	/* The theory is, that scheduler needs circa 10 fd's for its own uses,
+	/* The theory is, that scheduler needs circa 30 fd's for its own uses,
 	   and it will use all others on child-process communication fifos. */
+	/* Of these, 20 are for MQ2 sockets */
 
-	global_maxkids = resources_query_nofiles()-10;
+	global_maxkids = resources_query_nofiles()-30;
+
+	/* Probe how many FDs each TA needs! */
+
+	{
+	  int to[2], from[2];
+	  pipes_create(to,from);
+
+	  close(to[0]); close(to[1]);
+
+	  if (to[0] != from[1]) {
+
+	    /* Two pipes! */
+	    close(from[0]); close(from[1]);
+
+	    global_maxkids >>= 1; /* Half the MAXKIDS count */
+
+	  } /* else
+	       Socketpair or equivalent bidirectional one!
+	       Only one FD per child is used!               */
+	}
 
 	postoffice = rendezvous = logfn = statusfn = config = NULL;
 	daemonflg = 1;
@@ -570,7 +595,7 @@ main(argc, argv)
 			}
 			break;
 		case 'n':
-			never_full_content = !never_full_content;
+			default_full_content = !default_full_content;
 			break;
 		case 'N':
 			if ((transportmaxnofiles = atoi(optarg)) < 10)
@@ -634,6 +659,15 @@ main(argc, argv)
 		   progname);
 	  exit(128+errflg);
 	}
+	cp = getzenv("SCHEDULERDIRHASH");
+	if (cp){
+	  if ((cp[0] == '1' || cp[0] == '2') &&
+	      cp[1] == 0)
+	    hashlevels = cp[0] - '0';
+	  else
+	    hashlevels = 0;
+	}
+
 	mailshare = getzenv("MAILSHARE");
 	if (mailshare == NULL)
 	  mailshare = MAILSHARE;
@@ -756,7 +790,7 @@ main(argc, argv)
 	SIGNAL_HANDLE(SIGUSR2, sig_iot);	/* dump queue info */
 
 	/* call it to create the timeserver -- if possible */
-	init_timeserver();
+	init_timeserver(); /* Will take around 3-4 secs.. */
 
 	queryipcinit();
 
@@ -805,7 +839,12 @@ main(argc, argv)
 	       so that if we forget some jobs, they will
 	       become relearned soon enough.          */
 	    if (dirqueuescan(".", dirq, 
-			     (now >= next_idlecleanup)) < 150) {
+#if 1 /* Do recursively every time */
+			     1
+#else
+			     (now >= next_idlecleanup)
+#endif
+			     ) < 150) {
 	      /* If we have more things to scan, don't quit yet! */
 	      next_dirscan = now + sweepinterval;  /* 10 seconds interval */
 	      if (dirq->wrksum > 100)
@@ -819,19 +858,10 @@ main(argc, argv)
 	    idle_cleanup();
 	  }
 
-	  /* Submit one item from pre-scheduler-queue into the scheduler */
-	  if (dirq->wrksum > 1 ) {
-	    /* If more things in queue, submit them up to 200 pieces, or
-	       until spent two seconds at it! */
-	    time_t now0 = now + 2;
-	    for (i=0; i < 200 && now > now0; ++i) {
-	      if (syncweb(dirq) < 0)
-		break; /* Out of queue! */
-	      mytime(&now);
-	    }
+	  if (now >= next_interim_report_time) {
+	    next_interim_report_time = now + 120; /* every 2 minutes */
+	    interim_report_run();
 	  }
-	  if (dirq->wrksum > 0)
-	    syncweb(dirq);
 
 	  /* See when to timeout from mux() */
 	  timeout = next_dirscan;
@@ -860,6 +890,21 @@ main(argc, argv)
 	      rereadcf = 0;
 	    }
 	    mytime(&now);
+
+	    /* Submit one item from pre-scheduler-queue into the scheduler */
+	    if (dirq->wrksum > 1 ) {
+	      /* If more things in queue, submit them up to 200 pieces, or
+		 until spent two seconds at it! */
+	      time_t now0 = now + 2;
+	      for (i=0; i < 200 && now > now0; ++i) {
+		if (syncweb(dirq) < 0)
+		  break; /* Out of queue! */
+		mytime(&now);
+	      }
+	    }
+	    if (dirq->wrksum > 0)
+	      syncweb(dirq);
+
 	    if (slow_shutdown) {
 	      timeout = now+2;
 	      shutdown_kids();  /* If there are any to shut down.. */
@@ -911,12 +956,6 @@ int sig;
 	freeze = 1;
 }
 
-static RETSIGTYPE sig_alarm(sig)
-int sig;
-{
-	gotalarm = 1;
-	SIGNAL_HANDLE(SIGALRM, sig_alarm);
-}
 
 #ifdef SIGUSR2
 static RETSIGTYPE sig_iot(sig)
@@ -1093,6 +1132,10 @@ static int dirqueuescan(dir, dq, subdirs)
 	for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
 	  /* Scan filenames into memory */
 
+	  /* Here we are presuming the time retrieval is essentially
+	     _free_ in form of shared memory segment.. */
+	  queryipccheck();
+
 	  if (!syncstart && newents > newents_limit)
 	    break; /* At most NNN per one go */
 
@@ -1166,13 +1209,15 @@ static int dirqueuescan(dir, dq, subdirs)
 	return newents;
 }
 
-static int syncweb(dq)
+int syncweb(dq)
 	struct dirqueue *dq;
 {
 	struct stat *stbuf;
 	char *file;
+	struct dirstatname *dqstats;
 	struct spblk *spl;
 	int wrkcnt = 0;
+	int wrkidx = dq->wrkcount -1;
 	long ino;
 
 	/* Any work to do ? */
@@ -1207,19 +1252,36 @@ static int syncweb(dq)
 
 	mytime(&now);
 
-	if (dq->stats[dq->wrkcount-1]->not_before > now)
-	  return -1; /* WAIT! */
+	for (; wrkidx >= 0; --wrkidx) {
+
+	  /* Process only one -- from the tail -- but leave
+	     'not_before' things behind... */
+	  if (dq->stats[wrkidx]->not_before > now)
+	    continue; /* WAIT! */
+
+	  break;
+	}
 
 	/* Ok some, decrement the count to change it to index */
 	dq->wrkcount -= 1;
 	dq->wrksum   -= 1;
-	file  =   dq->stats[dq->wrkcount]->name;
-	stbuf = &(dq->stats[dq->wrkcount]->st);
-	ino   =   dq->stats[dq->wrkcount]->ino;
+
+	dqstats = dq->stats[wrkidx];
+	file  =   dqstats->name;
+	stbuf = &(dqstats->st);
+	ino   =   dqstats->ino;
+
+	if (dq->wrkcount > wrkidx) {
+	  /* Skipped some ! Compact the array ! */
+	  memcpy( &dq->stats[wrkidx], &dq->stats[wrkidx+1],
+		  sizeof(dq->stats[0]) * (dq->wrkcount - wrkidx));
+	}
+	dq->stats[dq->wrkcount] = NULL;
+
 	/* Now we have pointers */
 
 	/* Deletion from the  dirscan_mesh  should ALWAYS succeed.. */
-	spl = sp_lookup((u_long)ino,dirscan_mesh);
+	spl = sp_lookup((u_long)ino, dirscan_mesh);
 	if (spl != NULL)
 	  sp_delete(spl,dirscan_mesh);
 
@@ -1246,8 +1308,7 @@ static int syncweb(dq)
 	}
 
 	/* Free the pre-schedule queue entry */
-	free(dq->stats[dq->wrkcount]);
-	dq->stats[dq->wrkcount] = NULL;
+	free(dqstats);
 
 	return wrkcnt;
 }
@@ -1502,7 +1563,7 @@ void resync_file(proc, file)
 	  if (oldcfp->head == NULL) {
 	    cfp_free(oldcfp,spl);
 	    if (verbose)
-	      sfprintf(sfstdout," .. LOST in resync ??!\n");
+	      sfprintf(sfstdout," .. LOST in resync ?!\n");
 	  } else
 	    if (verbose)
 	      sfprintf(sfstdout," .. resynced!\n");
@@ -1548,24 +1609,24 @@ static struct ctlfile *schedule(fd, file, ino, reread)
 	  vtxprep_skip_any += vtxprep_skip;
 	  return NULL;
 	}
+
+	++MIBMtaEntry->mtaStoredMessages;
+	++global_wrkcnt;
+
 	if (cfp->head == NULL) {
-	  ++global_wrkcnt;
-	  ++MIBMtaEntry->mtaStoredMessages;
 	  unctlfile(cfp, 0); /* Delete the file.
 				(decrements those counters above!) */
 	  return NULL;
 	}
 
-	if (!reread)
+	if (!reread) {
 	  sp_install(cfp->id, (void *)cfp, 0, spt_mesh[L_CTLFILE]);
-	++MIBMtaEntry->mtaStoredMessages;
-	++global_wrkcnt;
-	
-	if (!reread)
+
 	  for (vp = cfp->head; vp != NULL; vp = vp->next[L_CTLFILE]) {
 	    /* Put into the schedules */
 	    vtxdo(vp, cehead, file);
 	  }
+	}
 
 	/* Now we have no more need for the contents in core */
 	if (cfp->contents != NULL) {
@@ -1646,7 +1707,8 @@ slurp(fd, ino)
 		offset = (int*)erealloc(offset,sizeof(int)*offsetspace);
 	      }
 	      offset[i++] = s - contents;
-	      if (*s == _CF_MSGHEADERS) {
+	      if (*s == _CF_MSGHEADERS ||
+		  *s == _CF_MIMESTRUCT) {
 		/* find a \n\n combination */
 		while (!(*s == '\n' && *(s+1) == '\n'))
 		  if (s-contents < stbuf.st_size)
@@ -1768,6 +1830,7 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	register int i, opcnt;
 	register int *lp;
 	int svn;
+	int c_echannel, c_ehost, c_l_echannel, c_l_ehost;
 	char *cp, *channel, *host, *l_channel, *l_host;
 	char *echannel, *ehost, *l_echannel, *l_ehost, mfpath[100], flagstr[2];
 	char *latest_sender = NULL;
@@ -1859,15 +1922,19 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	  if (cp[-1] == _CF_RECIPIENT) {
 	    ++cfp->rcpnts_total;
 	    if (*cp == _CFTAG_NOTOK) {
+#if 0
 	      ++cfp->rcpnts_failed;
 	      ++MIBMtaEntry->mtaReceivedRecipientsSc,
 	      ++MIBMtaEntry->mtaTransmittedRecipientsSc;
+#endif
 	      prevrcpt = -1;
 	    } else if (*cp != _CFTAG_OK) {
 	      ++cfp->rcpnts_work;
 	    } else { /* _CFTAG_OK */
+#if 0
 	      ++MIBMtaEntry->mtaReceivedRecipientsSc,
 	      ++MIBMtaEntry->mtaTransmittedRecipientsSc;
+#endif
 	    }
 	  }
 	  if (*cp == _CFTAG_NORMAL ||
@@ -1903,9 +1970,10 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	      *cp = 0;
 	      if (cfp->erroraddr)  free(cfp->erroraddr);
 	      cfp->erroraddr = strsave(latest_sender);
-	      if (strcmp(senderchannel,"error")==0)
+	      if (strcmp(senderchannel,"error")==0) {
 		cfp->iserrmesg = 1;
-	      break;
+		cfp->erroraddr[0] = 0; /* Make it '<>' */
+	      } break;
 	    case _CF_RECIPIENT:
 	      if (opcnt >= offspc-1) {
 		offspc *= 2;
@@ -2129,6 +2197,7 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	 */
 	strcpy(flagstr," ");
 	l_channel = l_echannel = l_host = l_ehost = flagstr;
+	c_echannel = c_ehost = c_l_echannel = c_l_ehost = 0;
 	svn = 0;
 	pvp = NULL;
 	head = NULL;
@@ -2136,12 +2205,9 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	for (i = 0; i < opcnt; ++i) {
 	  channel = bcp + offarr[i].offset /* - 2 */;
 	  while (*channel == ' ') ++channel; /* Skip possible space(s) */
-	  echannel = strchr(channel, ' ');
-	  if (echannel == NULL) /* error! */
-	    continue;
-	  *echannel = '\0';
-	  host = echannel + 1;
-	  while (*host == ' ') ++host; /* Skip space */
+	  host = echannel = skip821address(channel);
+	  c_l_echannel = *host; if (c_l_echannel) *host++ = '\0';
+	  while (*host == ' ') ++host; /* Skip space(s) */
 #if 1
 	  /* [mea]   channel ((mx.target)(mx.target mx.target)) rest.. */
 	  cp = host;
@@ -2179,7 +2245,7 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 		continue;
 	    }
 
-	  *ehost = '\0';
+	  c_ehost = *ehost; *ehost = '\0';
 	  /* compare with the last ones */
 	  if (strcmp(channel, l_channel) || strcmp(host, l_host)) {
 	    /* wrap and tie the old vertex node */
@@ -2223,11 +2289,14 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	  }
 	  /* stick the current 'r'ecipient  line into the current vertex */
 	  /* restore the characters */
-	  *l_echannel = *l_ehost = ' ';
-	  l_echannel  = echannel;
-	  l_ehost     = ehost;
-	  l_channel   = channel;
-	  l_host      = host;
+	  *l_echannel  = c_l_echannel;
+	  *l_ehost     = c_l_ehost;
+	  l_echannel   = echannel;
+	  l_ehost      = ehost;
+	  c_l_echannel = c_echannel;
+	  c_l_ehost    = c_ehost;
+	  l_channel    = channel;
+	  l_host       = host;
 	} /* for( .. i < opcnt .. ) */
 
 	/* wrap and tie the old vertex node (this is a copy of code above) */
@@ -2265,7 +2334,9 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	  link_in(L_CHANNEL, vp, channel);
 	}
 
-	*l_echannel = *l_ehost = ' ';
+	*l_echannel = c_l_echannel;
+	*l_ehost    = c_l_ehost;
+
 	/*
 	   for (vp = head; vp != NULL; vp = vp->next[L_CTLFILE]) {
 	     sfprintf(sfstdout,"--\n");
@@ -2273,6 +2344,7 @@ static struct ctlfile *vtxprep(cfp, file, rereading)
 	       sfprintf(sfstdout,"\t%s\n", cfp->contents+cfp->offset[vp->index[i]]);
 	   }
 	*/
+
 	cfp->head = head;
 	free(offarr);
 	if (verbose) {
@@ -2418,6 +2490,11 @@ static void vtxdo(vp, cehdr, path)
 	else
 	  vp->ce_expiry = 0;
 
+	if (tp->reporttimes[0])
+	  vp->nextrprttime = now + tp->reporttimes[0];
+	else
+	  vp->nextrprttime = now + global_report_interval;
+
 	thread_linkin(vp,tp,cnt,ce_fillin);
 
 	if (verbose>1)
@@ -2599,6 +2676,18 @@ timestring()
 	return timebuf;
 }
 
+static int running_on_alarm;
+static time_t alarm_time;
+
+static RETSIGTYPE sig_alarm(sig)
+int sig;
+{
+	time(&alarm_time);
+	SIGNAL_HANDLE(SIGALRM, sig_alarm);
+	alarm(1);
+}
+
+
 #if defined(HAVE_MMAP)
 struct timeserver {
 	int	pid;
@@ -2660,8 +2749,27 @@ static void init_timeserver()
 
 	ppid = fork();
 	if (ppid > 0) {
+	  time_t now2;
 	  timeserver_pid = ppid;
 	  timeserver_segment->pid = ppid;
+
+	  time(&now);
+	  sleep(3);
+
+#ifdef HAVE_SELECT
+	  now2 = timeserver_segment->tv.tv_sec;
+#else
+	  now2 = timeserver_segment->time_sec;
+#endif
+	  if (now2 == 0 || now2 == now) {
+	    /* HUH ?! No incrementation in 3 seconds ?? */
+	    /* Start ticking in alarm() mode! */
+	    kill (ppid, SIGTERM);
+	    timeserver_pid = 0;
+	    timeserver_segment = NULL;
+	    alarm(1);
+	  }
+
 	  return;
 	}
 	if (ppid < 0) return; /* Error ?? brr.. */
@@ -2671,6 +2779,12 @@ static void init_timeserver()
 
 	ppid = getppid(); /* who is our parent ? */
 
+#ifdef HAVE_SELECT
+	gettimeofday(&timeserver_segment->tv, NULL);
+#else
+	time(&timeserver_segment->time_sec);
+#endif
+
 	for(;;) {
 #ifdef HAVE_SELECT
 	  struct timeval tv;
@@ -2716,6 +2830,11 @@ time_t *timep;
 	  return t;
 	}
 #endif
+	if (running_on_alarm) {
+	  if (timep)
+	    *timep = alarm_time;
+	  return alarm_time;
+	}
 #ifdef HAVE_GETTIMEOFDAY
 	{
 	  struct timeval tv;
