@@ -14,7 +14,6 @@
 #include "zmalloc.h"
 #include "splay.h"
 #include "shmmib.h"
-#include "servauth.h"
 
 /* Some "forward declarations" */
 struct config_entry;
@@ -36,10 +35,13 @@ struct config_entry {
 	int	gid;		/* what gid to run the transport agent under */
 	char	*command;	/* the command to run			     */
 	int	flags;		/* miscellaneous flags			     */
-#define CFG_BYCHANNEL 0x0001
-#define CFG_WITHHOST  0x0002
-#define CFG_AGEORDER  0x0004	/* In order of the ctlfile->ctime -fields    */
-#define CFG_QUEUEONLY 0x0008
+
+/* #define CFG_BYCHANNEL	0x0001 */ /* obsolete thing.. */
+#define CFG_WITHHOST		0x0002
+#define CFG_AGEORDER		0x0004	/* by ctlfile->ctime -value    */
+#define CFG_QUEUEONLY		0x0008
+#define CFG_WAKEUPRESTARTONLY	0x0010
+
 #if 0
 	int	bychannel;	/* indicates $channel occurs in command      */
 	int	withhost;	/* indicated $host occurs in command	     */
@@ -53,11 +55,13 @@ struct config_entry {
 	int	skew;		/* retry skew parameter			     */
 	int	mark;		/* non-0 if we started a TA the last time    */
 	int	priority;	/* Scheduling priority			     */
-	int	overfeed;	/* How much overfeeding instead of sync processing ? */
+	int	overfeed;	/* How much overfeeding instead of
+				   sync processing ? */
 	char	**argv;		/* execv parameters for the command	     */
 	int	nretries;	/* number of retry factors known	     */
 	int	*retries;	/* list of nretries retry factors	     */
 	char	*deliveryform;	/* name of delivery error form		     */
+	int	reporttimes[4]; /* Shall match  _CFTAG_RCPTDELAYSIZE         */
 };
 
 #define	L_CTLFILE	0
@@ -138,6 +142,7 @@ struct thread {
 	struct procinfo	*proc;		/* NULL or ptr to xport proc	    */
 	int		thrkids;	/* Number of procs at this thread   */
 	int		jobs;		/* How many items in this thread    */
+	int		unfed;		/* How many not yet fed to TAs	    */
 	struct vertex   *nextfeed;	/* vertex within that thread	    */
 					/* feed_child() forwards nextfeed   */
 };
@@ -227,12 +232,38 @@ struct vertex {
 	int		retryindex;	/* cur index into ce->retries array  */
 	time_t		wakeup;		/* time to wake up and run this      */
 	time_t		lastfeed;	/* When the last feed was ?	     */
+	time_t		nextrprttime;	/* next time after which collected
+					   reports of this message will be
+					   produced.                         */
+	time_t		nextdlyrprttime;
 	char		*sender;	/* Message Sender/error recipient    */
 	int		ngroup;		/* number of addresses in group      */
 	int		index[1];	/* index of cfp->offset for group    */
 };
 
 
+
+#ifdef HAVE_SELECT
+#if	defined(BSD4_3) || defined(sun)
+#include <sys/file.h>
+#endif
+#include <errno.h>
+#include <netdb.h>
+#include <netinet/in.h>
+#include <sys/socket.h>
+#include <sys/time.h>
+
+#ifndef __Usockaddr__  /* Match the same one in  smtpserver.h */
+typedef union {
+    struct sockaddr_in v4;
+#ifdef INET6
+    struct sockaddr_in6 v6;
+#endif
+} Usockaddr;
+#define __Usockaddr__
+#endif
+#endif /* HAVE_SOCKET */
+
 /* mailq iterator state -- non-forking reporter mode */
 
 struct mailq; /* forward declarator */
@@ -241,6 +272,11 @@ struct mailq {
 	struct mailq	*nextmailq;
 	int		auth;		/* what can do */
 	int		fd;		/* FD for I/O (nonblocking-IO) */
+#ifdef HAVE_SELECT
+	Usockaddr	qaddr;
+#endif /* HAVE_SOCKET */
+
+	time_t		apopteosis;
 
 	char		*challenge;
 
@@ -263,5 +299,6 @@ struct mailq {
 #define MQ2MODE_SNMP	0x0001
 #define MQ2MODE_QQ	0x0002
 #define MQ2MODE_FULL	0x0004
-#define MQ2MODE_ETRN	0x0008
-#define MQ2MODE_KILL	0x0010
+#define MQ2MODE_FULL2	0x0008
+#define MQ2MODE_ETRN	0x0010
+#define MQ2MODE_KILL	0x0020
