@@ -1,7 +1,7 @@
 /*
  *	ZMailer 2.99.16+ Scheduler "threads" routines
  *
- *	Copyright Matti Aarnio <mea@nic.funet.fi> 1995-1999
+ *	Copyright Matti Aarnio <mea@nic.funet.fi> 1995-2000
  *
  *
  *	These "threads" are for book-keeping of information
@@ -48,6 +48,7 @@ static long threadid = 0;
 
 static void  thread_vertex_shuffle __((struct thread *thr));
 static struct threadgroup *create_threadgroup __((struct config_entry *cep, struct web *wc, struct web *wh, int withhost, void (*ce_fillin)__((struct threadgroup *, struct config_entry *)) ));
+static int   thread_start_ __((struct thread *thr));
 
 
 static struct threadgroup *
@@ -106,7 +107,7 @@ struct threadgroup *thgp;
 	   still non-zero... */
 	if (thgp->transporters || thgp->threads) return;
 
-if (verbose) sfprintf(sfstdout,"delete_threadgroup(%s/%d/%s)\n",
+if (verbose) sfprintf(sfstderr,"delete_threadgroup(%s/%d/%s)\n",
 		      thgp->wchan->name,thgp->withhost,thgp->whost->name);
 
 	if (thgp->idleproc != NULL || thgp->thread != NULL)
@@ -135,6 +136,9 @@ if (verbose) sfprintf(sfstdout,"delete_threadgroup(%s/%d/%s)\n",
 	  /* We were the only one! */
 	  thrg_root = NULL;
 	}
+
+memset(thgp, 0x55, sizeof(*thgp));
+
 	free(thgp);
 }
 
@@ -274,7 +278,7 @@ struct config_entry *cep;
 	thr->channel    = strsave(vtx->orig[L_CHANNEL]->name);
 	thr->host       = strsave(vtx->orig[L_HOST   ]->name);
 
-	if (verbose) sfprintf(sfstdout,"create_thread(%s/%d/%s) -> %p\n",
+	if (verbose) sfprintf(sfstderr,"create_thread(%s/%d/%s) -> %p\n",
 			      thr->channel,thgrp->withhost,thr->host,thr);
 
 	_thread_timechain_append(thr);
@@ -285,9 +289,10 @@ struct config_entry *cep;
 
 /*
  * Pick next thread from the group which this process serves.
+ * At call the proc->pthread->proc does not contain us!
  *
- * Result is  proc->pthread and proc->pthread->nextfeed being updated to
- * new thread, and function returns 1.
+ * Result is  proc->pthread  and  proc->pthread->nextfeed  being
+ * updated to new thread, and function returns 1.
  * If no new thread can be picked (all are active, and whatnot),
  * return 0.
  */
@@ -301,11 +306,6 @@ pick_next_thread(proc)
 	struct threadgroup  *thg  = proc->thg;
 	int once;
 
-	if (thr0 && thr0->proc) /* Anybody ? */
-	  thr0->proc = NULL; /* Remove it */
-	if (thr0 && thr0->nextfeed)
-	  thr0->nextfeed = NULL;
-
 	proc->pthread = NULL;
 
 	if (thg->cep->flags & CFG_QUEUEONLY)
@@ -335,6 +335,11 @@ pick_next_thread(proc)
 	    if (thr->proc && thr->nextfeed == NULL)
 	      continue; /* Nothing more to feed! See other threads! */
 
+	    if (proc->ch != ch) /* MUST have same CH data - in case we ever
+				   run with a clause where channel side is
+				   partially wild-carded.. */
+	      continue;
+
 	    proc->pthread = thr;
 	    thr->thrkids += 1;
 
@@ -365,8 +370,7 @@ pick_next_thread(proc)
 	      proc->ho->kids += 1;
 	    }
 
-	    /* In theory the CHANNEL could be different, in practice -- NOT! */
-	    proc->ch      = ch;
+	    /* The channel is be the same at old and new threads */
 
 	    /* Move the pickup pointer forward.. */
 	    thg->thread = thg->thread->nextthg;
@@ -384,56 +388,35 @@ pick_next_thread(proc)
 }
 
 
-static void delete_thread __((struct thread *, int));
-static void
-delete_thread(thr, ok)
+int
+delete_thread(thr)
      struct thread *thr;
-     int ok;
 {
 	/* Unlink this thread from thread-chain, and thread
 	   group.  Decrement thread-group count		    */
 
 	struct threadgroup *thg = thr->thgrp;
 
+	if (thr->thrkids || thr->jobs) return 0;
+
 	if (verbose)
-	  sfprintf(sfstdout,"delete_thread(%p:%s/%s) (thg=%p) jobs=%d\n",
+	  sfprintf(sfstderr,"delete_thread(%p:%s/%s) (thg=%p) jobs=%d\n",
 		   thr,thr->channel,thr->host,thg, thr->jobs);
 
 	free(thr->channel);
 	free(thr->host);
 
-	if (thr->jobs != 0) {
-	  sfprintf(sfstdout," DELETE_THREAD() WITH JOBS=%d\n",thr->jobs);
-	  abort(); /* Delete only when no vertices */
-	}
-
 	/* Unlink us from the thread time-chain */
 	/* ... and thread-group-ring */
 	_thread_timechain_unlink(thr);
 
-	if (thr->proc) {
-	  /* If this thread has a process(es), we detach it(them)! */
-	  struct procinfo *proc = thr->proc;
-
-	  for (;proc; proc = proc->pnext)
-	    proc->pthread = NULL;
-
-	  /* These are handled by  transport.c:reclaim():
-	     thr->proc->ho = NULL;
-	     thr->proc->ch = NULL;
-	  */
-	}
-
-	/* If threads count goes zero.. */
-
-	if (verbose)
-	  sfprintf(sfstderr,"delete_thread() thr->proc=%p pid=%d OF=%d\n",
-		   thr->proc, thr->proc ? (int)thr->proc->pid : 0,
-		   thr->proc ? thr->proc->overfed : 0);
+memset(thr, 0x55, sizeof(*thr));
 
 	free(thr);
+	return 1;
 }
 
+#if 0 /* Dead code.. */
 static void _thread_linkfront __((struct thread *, struct vertex *, struct vertex *));
 static void _thread_linkfront(thr,ap,vp)
 struct thread *thr;
@@ -449,7 +432,7 @@ struct vertex *ap, *vp;
 	  thr->thvertices = vp;
 	vp->thread = thr;
 }
-
+#endif
 
 /* the  _thread_linktail()  links a vertex into thread */
 static void _thread_linktail __((struct thread *, struct vertex *));
@@ -488,7 +471,7 @@ void (*ce_fillin) __((struct threadgroup*, struct config_entry *));
 	mytime(&now);
 
 	if (verbose)
-	  sfprintf(sfstdout,"thread_linkin([%s/%s],%s/%d/%s,%d)\n",
+	  sfprintf(sfstderr,"thread_linkin([%s/%s],%s/%d/%s,%d)\n",
 		   wc->name, wh->name, cep->channel,
 		   cep->flags & CFG_WITHHOST, cep->host, cfgid);
 
@@ -541,16 +524,26 @@ void (*ce_fillin) __((struct threadgroup*, struct config_entry *));
 	    /* Link the vertex into this thread! */
 
 	    if (verbose)
-	      sfprintf(sfstdout,"thread_linkin() to thg=%p[%s/%d/%s]; added into existing thread [%s/%s] thr->jobs=%d\n",
+	      sfprintf(sfstderr,"thread_linkin() to thg=%p[%s/%d/%s]; added into existing thread [%s/%s] thr->jobs=%d\n",
 		       thg,cep->channel,thg->withhost,cep->host,
 		       wc->name,wh->name,thr->jobs+1);
 
 	    _thread_linktail(thr,vp);
-	    vp->thgrp = thg;
-	    thr->jobs += 1;
+	    vp->thgrp   = thg;
+	    thr->jobs  += 1;
+
+	    if (thr->proc)     /* Caring about the UF count while running */
+	      thr->unfed += 1;
+
+	    if (thr->proc && (thr->nextfeed == NULL)) {
+	      /* It is running, but no nextfeed is set (anymore),
+		 tackle this vertex into the tail */
+
+	      thr->nextfeed = vp;
+	    }
 
 	    /* Hookay..  Try to start it too... */
-	    thread_start(thr, 0);
+	    thread_start_(thr);
 
 	    return;
 	  }
@@ -562,12 +555,12 @@ void (*ce_fillin) __((struct threadgroup*, struct config_entry *));
 	  vp->thgrp = thg;
 
 	  if (verbose)
-	    sfprintf(sfstdout,"thread_linkin() to thg=%p[%s/%d/%s]; created a new thread %p [%s/%s]\n",
+	    sfprintf(sfstderr,"thread_linkin() to thg=%p[%s/%d/%s]; created a new thread %p [%s/%s]\n",
 		     thg,cep->channel,thg->withhost,cep->host,
 		     thr,wc->name,wh->name);
 
 	  /* Try to start it too */
-	  thread_start(thr, 0);
+	  thread_start_(thr);
 
 	  return;
 	}
@@ -577,12 +570,8 @@ void (*ce_fillin) __((struct threadgroup*, struct config_entry *));
 	thr = create_thread(thg,vp,cep);
 	vp->thgrp = thg;
 
-	if (verbose)
-	  sfprintf(sfstdout,"thread_linkin() to thg=%p[%s/%d/%s]; created a new thread group, and thread [%s/%s]\n",
-		 thg,cep->channel,thg->withhost,cep->host,
-		 wc->name,wh->name);
 	/* Try to start it too */
-	thread_start(thr, 0);
+	thread_start_(thr);
 }
 
 struct web *
@@ -644,6 +633,9 @@ if (verbose)
 	  sp_delete(spl, spt_mesh[flag]);
 	symbol_free_db((u_char *)wp->name, spt_mesh[flag]->symbols);
 	free(wp->name);
+
+memset(wp, 0x55, sizeof(*wp));
+
 	free((char *)wp);
 }
 
@@ -665,6 +657,8 @@ struct vertex *vtx;
 	if (thr) {
 	  thr->jobs -= 1;
 
+	  if (thr->nextfeed     == vtx)
+	    thr->nextfeed       = thr->nextfeed->nextitem;
 	  if (thr->thvertices   == vtx)
 	    thr->thvertices     = vtx->nextitem;
 	  if (thr->lastthvertex == vtx)
@@ -690,14 +684,15 @@ web_detangle(vp, ok)
 
 	struct thread *thr = vp->thread;
 
-	if (thr && thr->nextfeed == vp) abort();
+	if (thr && thr->nextfeed == vp)
+	  thr->nextfeed = vp->nextitem; /* Oh, next one then.. */
 
 	unthread(vp);
 
 	/* The thread can now be EMPTY! */
 
 	if (thr && (thr->thvertices == NULL))
-	  delete_thread(thr, ok);
+	  delete_thread(thr);
 }
 
 static int vtx_mtime_cmp __((const void *, const void *));
@@ -778,8 +773,23 @@ struct thread *thr;
 	thr->thvertices   = ur_arr[  0];
 	thr->nextfeed     = ur_arr[  0];
 	thr->lastthvertex = ur_arr[n-1];
+	thr->unfed = n;
 }
 
+static int
+thread_start_(thr)
+     struct thread *thr;
+{
+	struct config_entry *ce = &(thr->thgrp->ce);
+
+	if (thr->proc != NULL) {
+	  /* There is *somebody* active!  Shall we start, or not ? */
+	  if (ce->flags & CFG_WAKEUPRESTARTONLY)
+	    return 0;
+	}
+
+	return thread_start(thr, 0);
+}
 
 
 /*
@@ -792,21 +802,31 @@ struct thread *thr;
  */
 
 int
-thread_start(thr, queue_only_too)
+thread_start(thr, queueonly_too)
      struct thread *thr;
-     int queue_only_too;
+     int queueonly_too;
 {
 	int rc;
 	struct vertex      *vp  = thr->thvertices;
 	struct threadgroup *thg = thr->thgrp;
 	struct config_entry *ce = &(thr->thgrp->ce);
-	struct web         *ho  = vp->orig[L_HOST];
-	struct web         *ch  = vp->orig[L_CHANNEL];
+	struct web         *ho;
+	struct web         *ch;
 
+	queryipccheck();
 
-	if (!queue_only_too && (thg->cep->flags & CFG_QUEUEONLY)) return 0;
+	if (!thr->thrkids && !thr->jobs) {
+	  /* Cleanup when no processes, nor vertices */
+	  delete_thread(thr);
+	  return 0;
+	}
 
 	if (syncstart || (freeze && !slow_shutdown)) return 0;
+	if (!queueonly_too && (ce->flags & CFG_QUEUEONLY)) return 0;
+
+	ho = vp->orig[L_HOST];
+	ch = vp->orig[L_CHANNEL];
+
 	if (procselect) {
 	  thr->pending = "procsel-mismatch";
 	  if (*procselect != '*' &&
@@ -825,7 +845,7 @@ thread_start(thr, queue_only_too)
 
 	if ((thr->thrkids >= ce->maxkidThread) ||
 	    /* FIXME: real unfed count ? */
-	    (thr->thrkids >= thr->jobs)) {
+	    (thr->proc && (thr->thrkids >= thr->unfed))) {
 	  if (verbose) {
 	    struct procinfo * proc = thr->proc;
 	    sfprintf(sfstderr," -- already running; thrkids=%d jobs=%d procs={ %p",
@@ -843,18 +863,23 @@ thread_start(thr, queue_only_too)
 	if (thg->idleproc) {
 	  struct procinfo *proc;
 
-	  /* Idle processor(s) exists, try to optimize by finding
-	     an idle proc with matching channel & host from previous
-	     activity. If not found, pick any. */
-	  
 	  /* There is at least one.. */
 	  proc = thg->idleproc;
 
+	  /* Idle processor(s) exists, try to optimize by finding
+	     an idle proc with matching channel & host from previous
+	     activity. If not found, pick any with matching CHANNEL,
+	     unless must have also matching HOST... */
+
 	  for (; proc && (proc->ho != ho || proc->ch != ch); proc = proc->pnext) ;
-	  if (!proc) {
-	    /* None of the previous ones matched, pick the first anyway */
+	  if (!proc && !thg->withhost) {
+	    /* None of the previous ones matched, pick with matching CHANNEL,
+	       HOST is allowed to wild-card.. */
 	    proc = thg->idleproc;
+	    for (; proc && (proc->ch != ch); proc = proc->pnext) ;
 	  }
+	  if (!proc)
+	    goto create_new;
 
 	  /* Selected one of them.. */
 
@@ -907,8 +932,6 @@ thread_start(thr, queue_only_too)
 	    thread_vertex_shuffle(thr);
 	  }
 
-	  thr->attempts += 1;
-
 	  /* Its idle process, feed it! */
 
 	  proc->state   = CFSTATE_LARVA;
@@ -932,6 +955,8 @@ thread_start(thr, queue_only_too)
 	  return 1;
 	}
 
+ create_new:
+
 	/* Check resource limits - MaxTa, MaxChan, MaxThrGrpTa */
 	/* If resources are exceeded, reschedule.. */
 
@@ -956,7 +981,7 @@ thread_start(thr, queue_only_too)
 
 	if (vp->ce_pending) {
 	  if (verbose)
-	    sfprintf(sfstdout,"%s: (%d %dC %dT %dTh) >= (%d %dC %dT %dTh)\n",
+	    sfprintf(sfstderr,"%s: (%d %dC %dT %dTh) >= (%d %dC %dT %dTh)\n",
 		     ce->command,
 		     numkids,
 		     vp->orig[L_CHANNEL]->kids,
@@ -994,10 +1019,6 @@ thread_start(thr, queue_only_too)
 			 thr->thvertices->orig[L_CHANNEL],
 			 thr->thvertices->orig[L_HOST]);
 
-	if (rc) { /* non-zero when child has started */
-	  thr->attempts += 1;
-	}
-
 	if (thr->proc && verbose)
 	  sfprintf(sfstderr,"%% thread_start(thr=%s/%d/%s) (proc=%p dt=%d thr=%p jobs=%d)\n",
 		   ch->name, thg->withhost, ho->name, thr->proc,
@@ -1012,7 +1033,7 @@ thread_start(thr, queue_only_too)
  * pick_next_vertex() -- pick next free to process vertex in this thread
  *
  * This is called *only* by  feed_child(), and  proc->vertex  directs
- * then caller of feed_child() to tune the process state.
+ * then the caller of feed_child() to tune the process state.
  * (From STUFFING to FINISHING and possibly to IDLE.)
  * 
  * - if (proc->pthread->nextfeed != NULL) ...nextfeed = ...nextfeed->nextitem;
@@ -1032,12 +1053,12 @@ pick_next_vertex(proc)
 	if (thr) vtx = thr->nextfeed;
 
 	if (verbose)
-	  sfprintf(sfstdout,"pick_next_vertex(proc=%p) proc->tofd=%d, thr=%p, pvtx=%p, jobs=%d OF=%d S=%s\n",
+	  sfprintf(sfstderr,"pick_next_vertex(proc=%p) proc->tofd=%d, thr=%p, pvtx=%p, jobs=%d OF=%d S=%s\n",
 		   proc, proc->tofd, thr, vtx, thr ? thr->jobs : 0,
 		   proc->overfed, proc_state_names[proc->state]);
 
 	if (proc->pid < 0 || proc->tofd < 0) {	/* "He is dead, Jim!"	*/
-	  if (verbose) sfprintf(sfstdout," ... NONE, 'He is dead, Jim!'\n");
+	  if (verbose) sfprintf(sfstderr," ... NONE, 'He is dead, Jim!'\n");
 	  return 0;
 	}
 
@@ -1050,9 +1071,10 @@ pick_next_vertex(proc)
 /*
  * The  thread_reschedule()  updates threads time-chain to match the
  * new value of wakeup for the  doagenda()  to latter use.
+ * Return 0 for DESTROYED thread, 1 for EXISTING thread.
  */
 
-void
+int
 thread_reschedule(thr, retrytime, index)
 struct thread *thr;
 int index;
@@ -1063,11 +1085,17 @@ time_t retrytime;
 	int skew;
 
 	if (verbose)
-	  sfprintf(sfstdout,"thread_reschedule() ch=%s ho=%s jobs=%d thr=%p proc=%p\n",
+	  sfprintf(sfstderr,"thread_reschedule() ch=%s ho=%s jobs=%d thr=%p proc=%p\n",
 		   thr->channel,thr->host,thr->jobs,thr,thr->proc);
 
+
+	if (!thr->thrkids && !thr->jobs) {
+	  delete_thread(thr);
+	  return 0;
+	}
+
 	/* If there are multiple kids working still, DON'T reschedule! */
-	if (thr->thrkids > 0) return;
+	if (thr->thrkids > 0 || !vtx) return 1;
 
 	/* find out when to retry */
 	mytime(&now);
@@ -1076,14 +1104,14 @@ time_t retrytime;
 	if (vtx->wakeup > now) {
 	  thr->wakeup = vtx->wakeup;
 	  if (verbose)
-	    sfprintf(sfstdout,"...prescheduled\n");
+	    sfprintf(sfstderr,"...prescheduled\n");
 	  goto timechain_handling;
 	} else if (vtx->wakeup < now-7200 /* more than 2h in history .. */ )
 	  vtx->wakeup = now;
 
 	if (vtx->thgrp->ce.nretries <= 0) {
 	  if (verbose)
-	    sfprintf(sfstdout,"...ce->retries = %d\n", vtx->thgrp->ce.nretries);
+	    sfprintf(sfstderr,"...ce->retries = %d\n", vtx->thgrp->ce.nretries);
 	  goto timechain_handling;
 	}
 
@@ -1172,6 +1200,8 @@ time_t retrytime;
 	  _thread_timechain_unlink(thr);
 	  _thread_timechain_append(thr);
 	}
+
+	return (thr != NULL);
 }
 
 
@@ -1180,8 +1210,6 @@ time_t retrytime;
  *
  */
 
-#define SALARM(N)  (verbose ? sfprintf(sfstdout,"alarm(%d) = %d\n", (N),alarm(N)):alarm(N))
-
 void
 reschedule(vp, factor, index)
 	struct vertex *vp;
@@ -1199,7 +1227,7 @@ reschedule(vp, factor, index)
 	mytime(&now);
 
 	if (verbose)
-	  sfprintf(sfstdout,"reschedule %p now %d expiry in %d attempts %d factor %d inum %d (%s/%s: %s)\n",
+	  sfprintf(sfstderr,"reschedule %p now %d expiry in %d attempts %d factor %d inum %d (%s/%s: %s)\n",
 		   vp, (int)now,
 		   (int)((vp->ce_expiry > 0) ? (vp->ce_expiry - now) : -999),
 		   vp->attempts,
@@ -1210,14 +1238,14 @@ reschedule(vp, factor, index)
 	/* if we are already scheduled for the future, don't reschedule */
 	if (vp->wakeup > now) {
 	  if (verbose)
-	    sfprintf(sfstdout,"prescheduled\n");
+	    sfprintf(sfstderr,"prescheduled\n");
 	  return;
 	} else if (vp->wakeup < now-7200 /* more than 2h .. */ )
 	  vp->wakeup = now;
 
 	if (ce->nretries <= 0) {
 	  if (verbose)
-	    sfprintf(sfstdout,"ce->retries = %d\n", ce->nretries);
+	    sfprintf(sfstderr,"ce->retries = %d\n", ce->nretries);
 	  return;
 	}
 	if (factor == -1 && vp->attempts) {
@@ -1247,19 +1275,24 @@ reschedule(vp, factor, index)
 	} else
 	  vp->wakeup += factor * ce->interval;
 
+	/* I THINK there could be an assert that if this happens,
+	   something is WRONG.. */
 	if (vp->attempts == 0)
 	  vp->wakeup = now;
 
 	/* XX: change this to a mod expression */
 	if (vp->wakeup < now)
 	  vp->wakeup = ((((now - vp->wakeup) / ce->interval)+1)
-			* ce->interval);
+			* ce->interval) + 10 + 2*thr->jobs;
+
+	/* Makes sure that next future event is at +10+2*jobcount seconds
+	   in the future..  A kludge approach, but still.. */
 
 	if (vp->ce_expiry > 0
 	    && vp->ce_expiry <= vp->wakeup
 	    && vp->attempts > 0) {
 	  if (verbose)
-	    sfprintf(sfstdout,"ce_expiry = %d, %d attempts\n",
+	    sfprintf(sfstderr,"ce_expiry = %d, %d attempts\n",
 		     (int)(vp->ce_expiry), vp->attempts);
 
 	  /* expire() will delete this vertex in due time */
@@ -1287,7 +1320,7 @@ idle_cleanup()
 
 	mytime(&now);
 
-	if (verbose) sfprintf(sfstdout,"idle_cleanup()\n");
+	if (verbose) sfprintf(sfstderr,"idle_cleanup()\n");
 
 	if (!thrg_root) return 0; /* No thread group! */
 
@@ -1321,7 +1354,7 @@ idle_cleanup()
 		/* Close the command channel, let it die itself.
 		   Rest of the cleanup happens via mux() service. */
 		if (verbose)
-		  sfprintf(sfstdout,"idle_cleanup() killing TA on tofd=%d pid=%d\n",
+		  sfprintf(sfstderr,"idle_cleanup() killing TA on tofd=%d pid=%d\n",
 			 p->tofd, (int)p->pid);
 
 		thr->wakeup = now-1; /* reschedule immediately! */
@@ -1356,7 +1389,7 @@ idle_cleanup()
 		/* Close the command channel, let it die itself.
 		   Rest of the cleanup happens via mux() service. */
 		if (verbose)
-		  sfprintf(sfstdout,"idle_cleanup() killing TA on tofd=%d pid=%d\n",
+		  sfprintf(sfstderr,"idle_cleanup() killing TA on tofd=%d pid=%d\n",
 			   p->tofd, (int)p->pid);
 		write(p->tofd,"\n",1);
 		pipes_shutdown_child(p->tofd);
@@ -1406,6 +1439,8 @@ void thread_report(fp,mqmode)
 	struct procinfo *p;
 	struct thread *thr;
 
+	int spc = (mqmode & MQ2MODE_FULL) ? ' ' : '\t';
+
 	mytime(&now);
 
 #if 0
@@ -1461,7 +1496,13 @@ void thread_report(fp,mqmode)
 	      }
 	    }
 
-	    if (mqmode & MQ2MODE_FULL) {
+	    if (mqmode & MQ2MODE_FULL2) {
+	      width = sfprintf(fp,"%s\t%s\t",
+			       /* thr->thvertices->orig[L_CHANNEL]->name */
+			       thr->channel,
+			       /* thr->thvertices->orig[L_HOST]->name */
+			       thr->host);
+	    } else if (mqmode & MQ2MODE_FULL) {
 	      width = sfprintf(fp,"    %s/%s/%d",
 			       /* thr->thvertices->orig[L_CHANNEL]->name */
 			       thr->channel,
@@ -1486,29 +1527,29 @@ void thread_report(fp,mqmode)
 
 	    if (mqmode & MQ2MODE_FULL)
 	      sfprintf(fp,"R=%-3d A=%-2d", thr->jobs, thr->attempts);
+	    if (mqmode & MQ2MODE_FULL2)
+	      sfprintf(fp,"R=%d\tA=%d", thr->jobs, thr->attempts);
 
 	    ++cnt;
 	    if (thr->proc != NULL &&
 		thr->proc->pthread == thr) {
 
 	      int thrprocs = 0;
-	      struct procinfo *proc = thr->proc;
+	      struct procinfo *proc;
 
-	      proc = thr->proc;
-	      while (proc) {
+	      for (proc = thr->proc; proc; proc = proc->pnext) {
 		++procs;
 		++thrprocs;
 		++thrkidsum;
-		proc = proc->pnext;
 	      }
 
-	      if (mqmode & MQ2MODE_FULL) {
+	      if (mqmode & (MQ2MODE_FULL|MQ2MODE_FULL2)) {
 		proc = thr->proc;
 
 		if (thr->thrkids != thrprocs)
-		  sfprintf(fp, " Kids=%d/%d", thr->thrkids, thrprocs);
+		  sfprintf(fp, "%cKids=%d/%d", spc, thr->thrkids, thrprocs);
 
-		sfprintf(fp, " P={");
+		sfprintf(fp, "%cP={", spc);
 		while (proc) {
 		  sfprintf(fp, "%d", (int)proc->pid);
 		  if (proc->pnext) sfprintf(fp, ",");
@@ -1517,7 +1558,7 @@ void thread_report(fp,mqmode)
 		sfprintf(fp, "}");
 
 		proc = thr->proc;
-		sfprintf(fp, " HA={");
+		sfprintf(fp, "%cHA={", spc);
 		while (proc) {
 		  sfprintf(fp, "%d", (int)(now - proc->hungertime));
 		  if (proc->pnext) sfprintf(fp, ",");
@@ -1526,7 +1567,7 @@ void thread_report(fp,mqmode)
 		sfprintf(fp, "}s");
 
 		proc = thr->proc;
-		sfprintf(fp, " FA={");
+		sfprintf(fp, "%cFA={", spc);
 		while (proc) {
 		  if (proc->feedtime == 0)
 		    sfprintf(fp, "never");
@@ -1538,7 +1579,7 @@ void thread_report(fp,mqmode)
 		sfprintf(fp, "}s");
 
 		proc = thr->proc;
-		sfprintf(fp, " OF={");
+		sfprintf(fp, "%cOF={", spc);
 		while (proc) {
 		  sfprintf(fp, "%d", proc->overfed);
 		  if (proc->pnext) sfprintf(fp, ",");
@@ -1547,32 +1588,34 @@ void thread_report(fp,mqmode)
 		sfprintf(fp, "}");
 
 		proc = thr->proc;
-		sfprintf(fp, " S={");
+		sfprintf(fp, "%cS={", spc);
 		while (proc) {
 		  sfprintf(fp, "%s", proc_state_names[proc->state]);
 		  if (proc->pnext) sfprintf(fp, ",");
 		  proc = proc->pnext;
 		}
 		sfprintf(fp, "}");
+
+		sfprintf(fp, "%cUF=%d", spc, thr->unfed);
 	      }
 
 	    } else if (thr->wakeup > now) {
-	      if (mqmode & MQ2MODE_FULL) {
-		sfprintf(fp," W=%ds",(int)(thr->wakeup - now));
+	      if (mqmode & (MQ2MODE_FULL|MQ2MODE_FULL2)) {
+		sfprintf(fp,"%cW=%ds", spc, (int)(thr->wakeup - now));
 	      }
 	    } else if (thr->pending) {
-	      if (mqmode & MQ2MODE_FULL) {
-		sfprintf(fp," pend=%s", thr->pending);
+	      if (mqmode & (MQ2MODE_FULL|MQ2MODE_FULL2)) {
+		sfprintf(fp,"%cpend=%s", spc, thr->pending);
 	      }
 	    }
 
-	    if (mqmode & MQ2MODE_FULL) {
+	    if (mqmode & (MQ2MODE_FULL|MQ2MODE_FULL2)) {
 	      *timebuf = 0;
 	      saytime((long)oldest_age_on_thread(thr), timebuf, 1);
-	      sfprintf(fp, " QA=%s", timebuf);
+	      sfprintf(fp, "%cQA=%s", spc, timebuf);
 
 	      if (thr->thvertices && thr->thvertices->ce_pending)
-		if (thr->thvertices->ce_pending != SIZE_L)
+		if (thr->thvertices->ce_pending != SIZE_L && spc == ' ')
 		  sfprintf(fp, "%s",
 			   (thr->thvertices->ce_pending ==
 			    L_CHANNEL ? " channelwait" : " threadwait"));
@@ -1642,6 +1685,94 @@ void thread_report(fp,mqmode)
 }
 
 
+void thread_detail_report(fp,mqmode,channel,host)
+     Sfio_t *fp;
+     int mqmode;
+     char *channel, *host;
+{
+	struct thread *th;
+	spkey_t spk;
+	struct spblk *spl;
+	struct web *wh, *wc;
+	struct vertex *vp;
+	int i;
+	char buf[100];
+
+	mytime(&now);
+
+	spk = symbol_lookup_db((void*)channel, spt_mesh[L_CHANNEL]->symbols);
+	spl = sp_lookup(spk, spt_mesh[L_CHANNEL]);
+	if (spl == NULL || spl->data == NULL) {
+	  /* Not found, nothing to do.. */
+	  return;
+	}
+	wc = (struct web *)spl->data;
+
+	spk = symbol_lookup_db((void*)host, spt_mesh[L_HOST]->symbols);
+	spl = sp_lookup(spk, spt_mesh[L_HOST]);
+	if (spl == NULL || spl->data == NULL) {
+	  /* Not found, nothing to do.. */
+	  return;
+	}
+	wh = (struct web *)spl->data;
+
+	for (th = thread_head; th; th = th->nexttr) {
+	  if (wh == th->whost && wc == th->wchan) {
+	    break;
+	  }
+	}
+	if (th) {
+	  /* Found it! */
+	  for (vp = th->thvertices; vp; vp = vp->nextitem) {
+	    struct ctlfile *cfp = vp->cfp;
+	    for (i = 0; i < vp->ngroup; ++i) {
+	      /* Spoolfile */
+	      if (vp->cfp->dirind >= 0)
+		sfprintf(fp, "%s/%s", cfpdirname(cfp->dirind), cfp->mid);
+	      else
+		sfprintf(fp, "%s", cfp->mid);
+	      /* Sender index -- or sender address */
+	      sfprintf(fp, "\t%s", cfp->erroraddr);
+	      /* Recipient offset */
+	      sfprintf(fp,"\t%d", cfp->offset[vp->index[i]]);
+	      /* Expiry stamp */
+	      sfprintf(fp,"\t%ld", (long)vp->ce_expiry);
+	      /* next wakeup */
+	      sfprintf(fp,"\t%ld", (long)vp->wakeup);
+	      /* last feed time */
+	      sfprintf(fp,"\t%ld", (long)vp->lastfeed);
+	      /* attempts */
+	      sfprintf(fp,"\t%d\t", vp->attempts);
+	      /* ce_pending */
+	      if (vp->wakeup > now) {
+		*buf = 0;
+		saytime((long)(vp->wakeup - now), buf, 1);
+		sfprintf(fp,"retry in %s", buf);
+	      } else {
+		switch(vp->ce_pending) {
+		case SIZE_L: /* BAD! */
+		  break;
+		case L_CHANNEL:
+		  sfprintf(fp,"channel");
+		  break;
+		default:
+		  sfprintf(fp,"thread");
+		  break;
+		}
+	      }
+	      /* message - if any */
+	      sfprintf(fp, "\t");
+	      if (vp->message)
+		sfprintf(fp,"%s", vp->message);
+	      sfprintf(fp, "\n");
+	    }
+	  }
+	}
+
+	sfsync(fp);
+}
+
+
 int thread_count_recipients()
 {
 	struct threadgroup *thg;
