@@ -4,14 +4,13 @@
  */
 /*
  *	Lots of modifications (new guts, more or less..) by
- *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-1999
+ *	Matti Aarnio <mea@nic.funet.fi>  (copyright) 1992-2000
  */
 
 
 #include "hostenv.h"
 #include <sfio.h>
 #include <sys/param.h>
-#include "scheduler.h"
 #include "mail.h"
 #ifdef HAVE_FCNTL_H
 #include <fcntl.h>
@@ -23,7 +22,6 @@
 /* #include <stdlib.h> */
 #include <unistd.h>
 
-#include "prototypes.h"
 #include "zsyslog.h"
 #include <sysexits.h>
 #ifdef HAVE_SYS_RESOURCE_H
@@ -32,6 +30,10 @@
 #ifdef HAVE_SYS_UN_H
 #include <sys/un.h>
 #endif
+
+#include "scheduler.h"
+#include "prototypes.h"
+
 #include "ta.h"
 #include "libz.h"
 
@@ -46,6 +48,14 @@ static void reclaim      __((int, int));
 static void waitandclose __((int));
 static void readfrom     __((int));
 
+/* These two are *not* exactly proper ones, but at least at most
+   systems it links properly without wider exportation of related
+   data structure types.  */
+
+extern int syncweb __((void *));
+extern void *dirq;
+
+
 extern FILE *vfp_open __((struct ctlfile *));
 
 #ifdef  HAVE_WAITPID
@@ -256,6 +266,7 @@ feed_child(proc)
 	    sprintf(cmdbuf, "%s/%s\t%s\n", d, vtx->cfp->mid, proc->ho->name);
 	  }
 	} else {
+
 	  if (proc->thg->withhost) { /* cmd-line was with host */
 	    cmdlen = 1 + strlen(vtx->cfp->mid);
 	    cmdbufalloc(cmdlen, & cmdbuf, & cmdbufspc);
@@ -297,11 +308,14 @@ feed_child(proc)
 	}
 
 	vtx->ce_pending  = 0; /* and clear the pending..       */
-
-	vtx->attempts   += 1;
+	/* DON'T double-count attempts on vertices!
+	   This gets counted when some diagnostics is received for
+	   this vertex. */
+	/* vtx->attempts   += 1; */
 	
 	/* It was fed (to buffer), clear this flag.. */
-	proc->overfed += 1;
+	proc->overfed        += 1;
+	proc->pthread->unfed -= 1;
 
 	if (verbose)
 	  sfprintf(sfstdout,"len=%d buf=%s", cmdlen, cmdbuf);
@@ -345,7 +359,12 @@ ta_hungry(proc)
 
 	  if (proc->overfed == 0 && proc->pthread)
 	    /* If we have a thread here still, reschedule it... */
-	    thread_reschedule(proc->pthread,0,-1);
+	    if (! thread_reschedule(proc->pthread,0,-1))
+	      proc->pthread = NULL;
+
+	  if (proc->pthread)
+	    if (delete_thread(proc->pthread))
+	      proc->pthread = NULL;
 
 	  return;
 	}
@@ -360,76 +379,108 @@ ta_hungry(proc)
 	  /* Thread selected already along with its first vertex,
 	     which is pointer by  proc->pthread->nextfeed  */
 
+	  /* However if we have multiple parallel processes
+	     in single thread, we may end up driving a new
+	     process straight from LARVA to IDLE... */
+
 	  proc->overfed = 0; /* Should not need setting ... */
 
-	  if (feed_child(proc) < 0)
+	  if (proc->pthread->nextfeed != NULL &&
+	      feed_child(proc) < 0)
 	    /* We got some error :-/  D'uh! */
 	    goto feed_error_handler;
 
 	  proc->state = CFSTATE_STUFFING;
-	  return;
+	  /* Not YET drained.. */
+	  if (proc->pthread->nextfeed)
+	    return;
 
 	case CFSTATE_STUFFING: /* "2" */
 
 	  if (proc->overfed > 0) return;
 
-	  while ((proc->state == CFSTATE_STUFFING) &&
-		 proc->pthread && proc->pthread->nextfeed) {
+	  if (proc->pthread && proc->pthread->nextfeed) {
 
-	    /* As long as:
-	       - we have next vertex to feed
-	       - there is no command buffer backlog
-	       - state stays in STUFFING
-	    */
+	    while ((proc->state == CFSTATE_STUFFING) &&
+		   proc->pthread && proc->pthread->nextfeed) {
 
-	    i = feed_child(proc);
-	    if (i < 0)
-	      goto feed_error_handler; /* Outch! */
+	      /* As long as:
+		 - we have next vertex to feed
+		 - there is no command buffer backlog
+		 - state stays in STUFFING
+	      */
 
-	    if (proc->tofd >= 0 && proc->cmdlen != 0)
-	      return; /* Incomplete feed -- stop feeding here */
+	      i = feed_child(proc);
+	      if (i < 0)
+		goto feed_error_handler; /* Outch! */
 
-	    if (proc->overfed > proc->thg->ce.overfeed)
-	      return; /* Or over limit ...       */
+	      if (proc->tofd >= 0 && proc->cmdlen != 0)
+		break; /* Incomplete feed -- stop feeding here */
 
+	      if (proc->overfed > proc->thg->ce.overfeed)
+		break; /* Or over limit ...       */
+
+	    }
+	    /* As long as we have things to feed (or have fed anything!),
+	       we stay in the STUFFING state.  This was if we latter get
+	       new workitems,  thread_linkin() can simply put them into
+	       the 'nextfeed' pointer, and we feed them right away. */
+	    return;
 	  }
 
+	  /* Didn't have anything to feed :-( */
+
 	  proc->state = CFSTATE_FINISHING;
 	  /* FALL THRU! */
 
 	case CFSTATE_FINISHING: /* "3" */
 
+	  /* "retryat" may have kicked us into this state also.. */
+
 	  if (proc->overfed > 0) return;
 
+	  /* Well, this thread was done, so long!
+	     This TA may have other things to poke at! */
+
 	  thr0 = proc->pthread;
-	  if (proc->pthread) {
-	    if (proc->pthread->proc == proc)
-	      proc->pthread->proc = proc->pnext;
-
-	    proc->pthread->thrkids -= 1;
-
-	    /* Possibly also reschedule the thread (if last thrkid!) */
-	    thread_reschedule(proc->pthread,0,-1);
-	  }
-	  proc->pthread = NULL;
+	  /* proc->pthread = NULL; */ /* NO KILL YET! */
 
 	  /* Disconnect the previous thread from the proc. */
+
+	  thr0->thrkids -= 1;
+
+	  if (thr0->proc == proc) /* Thread Process Chain Leader */
+	      thr0->proc = proc->pnext;
+
+	  /* Disconnect from process chain */
 	  if (proc->pnext) proc->pnext->pprev = proc->pprev;
 	  if (proc->pprev) proc->pprev->pnext = proc->pnext;
 	  proc->pnext = proc->pprev = NULL;
 
 	  /* Next: either the thread changes, or
-	     the process moves into IDLE state */
+	     the process moves into IDLE state. */
 
 	  if (pick_next_thread(proc)) {
-	    struct thread *thr = proc->pthread;
+
 	    /* We have WORK !  We are reconnected to the new thread! */
+	    /* Picked a new thread, which isn't same as THR0! */
+
+	    /* Picked something, reschedule the old thread,
+	       and if it got destroyed (by expiry) loose it... */
+	    if (thr0 && !thread_reschedule(thr0, 0, -1))
+	      thr0 = NULL;
+	    if (thr0) delete_thread(thr0); /* possibly kill it if no more
+					      things in it active! */
+	    /* thr0 = NULL; -- this is dead variable.. */
 
 	    if (verbose)
 	      sfprintf(sfstdout, "%% pick_next_thread(proc=%p) gave thread %p\n",
 		       proc, proc->pthread);
 
-	    thr->attempts += 1;
+	    /* DON'T double-count attempts on threads!
+	       This gets counted when some diagnostic is received
+	       for this thing.. */
+	    /* proc->pthread->attempts += 1; */
 
 	    if (feed_child(proc))
 	      /* non-zero return means things went wrong somehow.. */
@@ -439,6 +490,8 @@ ta_hungry(proc)
 	    return;
 	  }
 
+	  /* proc->pthread == NULL  now */
+	  /* thr0->thrkids has been decremented */
 	  /* No work in sight, queue up '#idle\n' string. */
 
 	  if (thr0) {
@@ -447,10 +500,10 @@ ta_hungry(proc)
 	     now join back so that IDLE will disconnect us... */
 
 	    proc->pthread = thr0;
-	    proc->pthread->thrkids += 1;
-	    proc->pnext = proc->pthread->proc;
+	    thr0->thrkids += 1;
+	    proc->pnext   = thr0->proc;
 	    if (proc->pnext) proc->pnext->pprev = proc;
-	    proc->pthread->proc = proc;
+	    thr0->proc = proc;
 	  }
 
 	  if ((proc->cmdlen + 7) >= proc->cmdspc) {
@@ -471,25 +524,29 @@ ta_hungry(proc)
 	case CFSTATE_IDLE: /* "4" */
 	  /* The process has arrived into IDLE pool! */
 
-	  if (proc->overfed != 0) abort();
+	  if (proc->overfed > 0) abort();
+	  /* We have seen at least once that this value went NEGATIVE! */
 
 	  if (verbose)
 	    sfprintf(sfstdout, " ... IDLE THE PROCESS %p (of=%d).\n",
 		     proc, proc->overfed);
 
+	  thr0 = proc->pthread;
+	  if (thr0) {
+
+	    proc->pthread = NULL;
+
+	    if (thr0->proc == proc) /* Was chain head */
+	      thr0->proc = proc->pnext;
+
+	    thr0->thrkids -= 1;
+
+	  }
+
 	  /* Unlink me from the active chain */
 	  if (proc->pnext) proc->pnext->pprev = proc->pprev;
 	  if (proc->pprev) proc->pprev->pnext = proc->pnext;
 
-	  if (proc->pthread) {
-
-	    if (proc->pthread->proc == proc) /* Was chain head */
-	      proc->pthread->proc = proc->pnext;
-
-	    proc->pthread->thrkids -= 1;
-	    proc->pthread = NULL;
-	  }
-
 	  proc->pnext = proc->thg->idleproc;
 	  if (proc->pnext) proc->pnext->pprev = proc;
 	  proc->pprev = NULL;
@@ -498,6 +555,9 @@ ta_hungry(proc)
 	  proc->thg->idlecnt += 1;
 	  ++idleprocs;
 	  
+	  /* Failed to pick anything, reschedule the old thread.
+	     Possibly killed the thread.. */
+	  if (thr0) thread_reschedule(thr0, 0, -1);
 	  return;
 
 	default: /* CFSTATE_ERROR: "0" */
@@ -912,12 +972,12 @@ if (verbose)
 	  if (proc->pthread->proc == proc)
 	    proc->pthread->proc = proc->pnext;
 
-	  /* Conditionally reschedule the thread */
-	  thread_reschedule(proc->pthread,0,-1);
-
 	  /* Disjoin the thread from the proc */
 	  proc->pthread->thrkids -= 1;
 
+	  /* Conditionally reschedule the thread */
+	  thread_reschedule(proc->pthread,0,-1);
+
 	  proc->pthread = NULL;
 
 	} else {
@@ -969,16 +1029,6 @@ static void waitandclose(fd)
 # include <sys/select.h>
 #endif
 
-
-#if	defined(BSD4_3) || defined(sun)
-#include <sys/file.h>
-#endif
-#include <errno.h>
-#include <netdb.h>
-#include <netinet/in.h>
-#include <sys/socket.h>
-#include <sys/time.h>
-
 #ifndef	NFDBITS
 /*
  * This stuff taken from the 4.3bsd /usr/include/sys/types.h, but on the
@@ -1024,7 +1074,7 @@ int
 mux(timeout)
 time_t timeout;
 {
-	int	i, n, rc, maxf;
+	int	i, n, maxf;
 	fd_set	rdmask;
 	fd_set	wrmask;
 	struct timeval tv;
@@ -1035,7 +1085,7 @@ time_t timeout;
 	  return 0;
 	}
 
-	mytime(&now);
+	queryipccheck();
 
 	tv.tv_sec = timeout - now; /* Timeout in seconds */
 	if (timeout < now)
@@ -1070,6 +1120,10 @@ time_t timeout;
 	    maxf = querysocket;
 	}
 
+	/* Although we don't react on the results of these MQ2 fd's,
+	   getting them to break timeouts is important for MAILQv2
+	   responsiveness. ! */
+
 	if (mailqmode == 2)
 	  maxf = mq2add_to_mask(&rdmask, &wrmask, maxf);
 
@@ -1111,42 +1165,10 @@ time_t timeout;
 	  return 1;
 	} else {
 	  /*sfprintf(sfstderr, "got %d ready (%x)\n", n, rdmask.fds_bits[0]);*/
-	  if (querysocket >= 0 && _Z_FD_ISSET(querysocket, rdmask)) {
-	    struct sockaddr_in raddr;
-	    int	raddrlen;
 
-	    _Z_FD_CLR(querysocket, rdmask);
-	    raddrlen = sizeof raddr;
-	    i = accept(querysocket, (struct sockaddr *)&raddr, &raddrlen);
-	    if (i < 0) {
-	      perror("accept");
-	    } else if (mailqmode == 1) {
-	      rc = fork();
-	      if (rc == 0) { /* Child! */
-		close(querysocket);
-#ifdef HAVE_TCPD_H /* TCP-Wrapper code */
-		if (raddr.sin_family == AF_INET &&
-		    wantconn(i, "mailq") == 0) {
-		  char *msg = "refusing 'mailq' query from your whereabouts\r\n";
-		  int   len = strlen(msg);
-		  write(i,msg,len);
-		  _exit(0);
-		}
-#endif
-		qprint(i);
-		/* Silence memory debuggers about this child's
-		   activities by doing exec() on the process.. */
-		/* execl("/bin/false","false",NULL); */
-		_exit(0); /* _exit() should be silent too.. */
-	      }
-	      /* if (rc > 0)
-		 ++numkids; */
-	      close(i);
-	    } else {
-	      /* mailqmode == 2 */
-	      mq2_register(i);
-	    }
-	  }
+	  /* In case we really should react.. */
+	  queryipccheck();
+
 	  if (cpids != NULL) {
 
 	    for (proc = cpids, i = 0; i < maxf; ++i, ++proc) {
@@ -1165,9 +1187,11 @@ time_t timeout;
 
 	      /* Because this loop might take a while ... */
 	      queryipccheck();
+
+	      syncweb(dirq);
+
 	    }
 	  }
-	  mq2_areinsets(&rdmask, &wrmask);
 
 	  in_select = 0;
 	}
@@ -1175,15 +1199,35 @@ time_t timeout;
 	return 0;
 }
 
+/* Call it how often you wish, but act it only once a second, or so.. */
+static time_t lastqueryipccheck;
+static time_t qipcretry;
+
 void
 queryipccheck()
 {
+	int	n;
+	fd_set	rdmask;
+	fd_set	wrmask;
+	struct timeval tv;
+	int maxfd;
+
+	mytime(&now);
+	if (!mq2_active() && (lastqueryipccheck == now)) return;
+	lastqueryipccheck = now;
+
+	if (qipcretry > 0 && qipcretry <= now) {
+	  qipcretry = 0;
+	  queryipcinit();
+	  /*
+	   * If qipcretry is set here, the value will be ignored, but
+	   * that's ok since sweepretry is active by now
+	   */
+	}
+
 	if (querysocket >= 0) {
-	  int	n;
-	  fd_set	rdmask;
-	  fd_set	wrmask;
-	  struct timeval tv;
-	  int maxfd = querysocket;
+
+	  maxfd = querysocket;
 
 	  tv.tv_sec = 0;
 	  tv.tv_usec = 0;
@@ -1192,17 +1236,18 @@ queryipccheck()
 	  _Z_FD_ZERO(wrmask);
 	  _Z_FD_SET(querysocket, rdmask);
 
-	  if (mailqmode == 2)
+	  if (mailqmode == 2) {
 	    maxfd = mq2add_to_mask(&rdmask, &wrmask, maxfd);
+	    n = select(maxfd+1, &rdmask, &wrmask, NULL, &tv);
+	    if (n > 0)
+	      mq2_areinsets(&rdmask, &wrmask);
+	  } else 
+	    n = select(maxfd+1, &rdmask, &wrmask, NULL, &tv);
 
-	  n = select(maxfd+1, &rdmask, &wrmask, NULL, &tv);
-
-	  if (n > 0)
-	    mq2_areinsets(&rdmask, &wrmask);
 
 	  if (n > 0 &&
 	      _Z_FD_ISSET(querysocket, rdmask)) {
-	    struct sockaddr_in raddr;
+	    Usockaddr raddr;
 	    int raddrlen = sizeof(raddr);
 
 	    n = accept(querysocket, (struct sockaddr *)&raddr, &raddrlen);
@@ -1225,12 +1270,12 @@ queryipccheck()
 		  /* Silence memory debuggers about this child's
 		     activities by doing exec() on the process.. */
 		  /* execl("/bin/false","false",NULL); */
-		  _exit(0); /* _exit() should be silent too.. */
+		  _exit(0); /* _exit() should be silent, too.. */
 		}
 		close(n);
 	      } else {
 		/* mailqmode == 2 */
-		mq2_register(n);
+		mq2_register(n, &raddr);
 	      }
 	    }
 	  }
@@ -1308,7 +1353,7 @@ queryipcinit()
 #ifdef	AF_INET
 	if (modecode == 1) {
 	  struct servent *serv;
-	  struct sockaddr_in sad;
+	  Usockaddr ua;
 	  int on = 1;
 	  int port = 174; /* MAGIC knowledge */
 
@@ -1320,21 +1365,45 @@ queryipcinit()
 	      port = ntohs(serv->s_port);
 	  }
 	  qipcretry = now + 5;
-	  sad.sin_port        = htons(port);
-	  sad.sin_family      = AF_INET;
-	  sad.sin_addr.s_addr = htonl(INADDR_ANY);
-	  if ((querysocket = socket(PF_INET, SOCK_STREAM, 0)) < 0) {
+	  memset(&ua, 0, sizeof(ua));
+	  ua.v4.sin_port        = htons(port);
+	  ua.v4.sin_family      = AF_INET;
+	  ua.v4.sin_addr.s_addr = htonl(INADDR_ANY);
+	  querysocket = -1;
+#ifdef INET6_not_now
+	  querysocket = socket(PF_INET6, SOCK_STREAM, 0);
+#endif
+	  if (querysocket < 0)
+	    querysocket = socket(PF_INET, SOCK_STREAM, 0);
+#ifdef INET6
+	  else {
+	    memset(&ua, 0, sizeof(ua));
+	    ua.v6.sin6_port   = htons(port);
+	    ua.v6.sin6_family = AF_INET6;
+	  }
+#endif
+	  if (querysocket < 0) {
 	    perror("querysocket: socket(PF_INET)");
 	    return;
 	  }
 	  setsockopt(querysocket, SOL_SOCKET, SO_REUSEADDR, (void*)&on, sizeof(on));
 
-	  if (bind(querysocket, (struct sockaddr *)&sad, sizeof sad) < 0) {
-	    perror("bind:mailq socket");
-	    close(querysocket);
-	    querysocket = -1;
-	    return;
-	  }
+#ifdef INET6
+	  if (ua.v6.sin6_family == AF_INET6) {
+	    if (bind(querysocket, (struct sockaddr *)&ua, sizeof ua.v6) < 0) {
+	      perror("bind:mailq ipv6 socket");
+	      close(querysocket);
+	      querysocket = -1;
+	      return;
+	    }
+	  } else
+#endif
+	    if (bind(querysocket, (struct sockaddr *)&ua, sizeof ua.v4) < 0) {
+	      perror("bind:mailq ipv4 socket");
+	      close(querysocket);
+	      querysocket = -1;
+	      return;
+	    }
 
 #if defined(F_SETFD)
 	  fcntl(querysocket, F_SETFD, 1); /* close-on-exec */
