@@ -26,12 +26,6 @@ static int ctlowner __((struct ctlfile *));
 static void vtxupdate __((struct vertex *, int, int));
 static void expaux __((struct vertex *, int, const char *));
 
-#if 0
-extern int ermdir();
-extern int rmdir();
-extern int eunlink();
-extern struct spblk *sp_fhead(), *sp_fnext();
-#endif
 extern time_t now;
 extern int global_wrkcnt;
 extern char *procselect;
@@ -85,17 +79,20 @@ update(fd, diagnostic)
 	const char	*type, *message, *notary;
 	long	offset;
 	long	inum;
-	int	index;
+	int	index, c;
 	struct vertex *vp;
 	struct diagcodes *dcp;
 	struct procinfo *proc = &cpids[fd];
 
 	if (*diagnostic == 0) {
+#if 0
 	  /* Lone newline.. old-style indications from the transporter */
 	  ta_hungry(proc);
 	  return;
+#else
+	diagnostic = "#";
+#endif
 	}
-
 	if (*diagnostic == '#') {
 	  /* Now (*diagnostic == '#') */
 
@@ -144,7 +141,7 @@ update(fd, diagnostic)
 	}
 	*cp = 0; /* Trailing TAB after notary string, NULL it */
 	type = ++cp;
-	while (*cp != '\0' && isascii(*cp) && !isspace(*cp))
+	while ( (c = (0xFF & *cp)) && isascii(c) && !isspace(c))
 	  ++cp;
 	if (*cp == '\0') {
 	  message = NULL;
@@ -159,6 +156,10 @@ update(fd, diagnostic)
 	if ((vp = findvertex(inum, offset, &index)) == NULL)
 	  return;
 
+	++vp->attempts; /* Did attempt this vertex! */
+	if (vp->thread)
+	  vp->thread->attempts += 1; /* and on this thread too! */
+
 	if (vp->notary != NULL)
 	  free(vp->notary);
 	vp->notary = NULL;
@@ -200,7 +201,7 @@ unctlfile(cfp, no_unlink)
 	  else
 	    strcpy(path, cfp->mid);
 
-	  reporterrs(cfp);
+	  reporterrs(cfp, 0);
 
 	  if (do_syslog) {
 	    char taspid[30];
@@ -271,8 +272,8 @@ void unvertex(vp, justfree, ok)
 		   "unvertex(vtx=%p (thr=%p, ng=%d) ,%d,%d)\n",
 		   vp, vp->thread, vp->ngroup, justfree, ok);
 
-	if (vp->thread && vp->thread->proc) {
-	  /* Somebody here, move it elsewere! */
+	if (vp->thread) {
+	  /* Somebody possibly here, move it elsewere! */
 	  /* This MAY happen -- expire() hits when some
 	     thread is coming into processing...            */
 	  if (vp->thread->nextfeed == vp)
@@ -309,6 +310,9 @@ void unvertex(vp, justfree, ok)
 	if (vp->message != NULL) free(vp->message);
 	if (vp->notary  != NULL) free(vp->notary);
 	/* if (vp->sender != NULL) free(vp->sender); */ /* XX: cache !! ?? */
+
+memset(vp, 0x55, sizeof(*vp));
+
 	free((char *)vp);
 
 	return;
@@ -581,11 +585,12 @@ static void expaux(vp, index, buf)
 	    msgerror(vp, vp->cfp->offset[index], buf);
 	    break;
 	  }
-
+#if 0
 	/* Log something into the scheduler log */
 	sfprintf(sfstderr, "%s %s: %s/%s from %s %s\n", timestring(), progname,
 		 vp->orig[L_CHANNEL]->name, vp->orig[L_HOST]->name,
 		 vp->cfp->erroraddr == NULL ? "?" : vp->cfp->erroraddr, buf);
+#endif
 
 	logstat(vp,"expire");
 
@@ -785,10 +790,11 @@ static int u_error(proc, vp, index, inum, offset, notary, message)
 {
 	if (message == NULL)
 	  message = "(unknown)";
+#if 0
 	sfprintf(sfstderr,
 		 "%s %s: %ld/%ld/%s/error %s\n", timestring(),
 		 vp->thgrp->ce.command, inum, offset, notary, message);
-
+#endif
 	if (!procselect && vp->notaryflg & NOT_FAILURE)
 	  msgerror(vp, offset, message);
 #if 0
@@ -821,10 +827,11 @@ static int u_error2(proc, vp, index, inum, offset, notary, message)
 {
 	if (message == NULL)
 	  message = "(unknown)";
+#if 0
 	sfprintf(sfstderr,
 		"%s %s: %ld/%ld/%s/error2 %s\n", timestring(),
 		vp->thgrp->ce.command, inum, offset, notary, message);
-
+#endif
 	/* We don't need to log it! */
 	vp->cfp->haderror = 1; /* Mark it into the incore dataset */
 #if 0
@@ -861,6 +868,7 @@ static int u_retryat(proc, vp, index, inum, offset, notary, message)
 	time_t	retrytime;
 	long    dtvalue;
 	const char * cp;
+	int c;
 
 	/* If a message gets a "retryat" signal, kick this thread at
 	   next  "#hungry"  into FINISHING state */
@@ -870,14 +878,6 @@ static int u_retryat(proc, vp, index, inum, offset, notary, message)
 		   "RETRYAT: proc=%p (S=%d OF=%d tofd=%d) vp=%p[%d] message='%s'\n",
 		   proc, (int)proc->state, proc->overfed, proc->tofd, vp, index, message);
 
-#if 0
-	if ((proc->state   == CFSTATE_LARVA) &&
-	    (proc->overfed == 1) &&
-	    (proc->tofd    >= 0))
-	  proc->state = CFSTATE_FINISHING;
-	/* That previous state change would mean a fork()ed sub-program
-	   has spontaneously decided to produce  retryat  state... */
-#endif
 
 	if ((proc->state   == CFSTATE_STUFFING) &&
 	    (proc->tofd    >= 0)) {
@@ -886,17 +886,26 @@ static int u_retryat(proc, vp, index, inum, offset, notary, message)
 	     feeding and move to the finishing state before
 	     continuing anywhere... */
 
-	  /* sfprintf(sfstderr,
-	     "%% u_retryat(proc=%p) -> CFSTATE_FINISHING\n", proc); */
+	  if (proc->pthread && proc->pthread->nextfeed &&
+	      proc->pthread->nextfeed->ce_expiry > 0 &&
+	      proc->pthread->nextfeed->ce_expiry < now) {
 
-	  proc->state = CFSTATE_FINISHING;
+	    /* Whops! next one is an *old* message, keep it in
+	       the STUFFING state, and let other messages of
+	       the queue possibly get the same quick diagnostics
+	       after at least one activation... */
+	  } else {
+	    /* sfprintf(sfstderr,
+	       "%% u_retryat(proc=%p) -> CFSTATE_FINISHING\n", proc); */
+	    proc->state = CFSTATE_FINISHING;
+	  }
 	}
 
 	if (*message == '+')
 	  ++message;
 	dtvalue = 0;
 	sscanf(message, "%ld", &dtvalue);
-	for (cp = message; *cp != '\0' && isdigit(*cp); ++cp)
+	for (cp = message; (c = 0xFF & *cp) && isdigit(c); ++cp)
 	  continue;
 	if (*cp != '\0')
 	  ++cp;
