@@ -20,7 +20,7 @@ static void cfparam __((char *));
 static void cfparam(str)
      char *str;
 {
-    char *name, *param1, *param2;
+    char *name, *param1, *param2, *param3;
 
     name = strchr(str, '\n');	/* The trailing newline chopper ... */
     if (name)
@@ -55,13 +55,18 @@ static void cfparam(str)
     }
 
     SKIPWHILE(isspace, str);
-    param1 = str;
+    param1 = *str ? str : NULL;
 
     SKIPWHILE(!isspace, str);
     if (*str != 0)
 	*str++ = 0;
     SKIPWHILE(isspace, str);
-    param2 = str;
+    param2 = *str ? str : NULL;
+    SKIPWHILE(!isspace, str);
+    if (*str != 0)
+	*str++ = 0;
+    SKIPWHILE(isspace, str);
+    param3 = *str ? str : NULL;
     SKIPWHILE(!isspace, str);
     if (*str != 0)
 	*str++ = 0;
@@ -123,6 +128,8 @@ static void cfparam(str)
 	  }
 	  bindaddr.v4.sin_family = AF_INET;
 	} else {
+	  /* XXX: TODO: Try to see if this is an interface name, and pick
+	     IPv4 and/or IPv6 addresses for that interface. */
 	  bindaddr_set = 0;
 	}
     }
@@ -140,6 +147,9 @@ static void cfparam(str)
     } else if (cistrcmp(name, "RcptLimitCnt") == 0 && param1) {
 	sscanf(param1, "%d", &rcptlimitcnt);
 	if (rcptlimitcnt < 100) rcptlimitcnt = 100;
+    } else if (cistrcmp(name, "RcptLimitCount") == 0 && param1) {
+	sscanf(param1, "%d", &rcptlimitcnt);
+	if (rcptlimitcnt < 100) rcptlimitcnt = 100;
     } else if (cistrcmp(name, "Rcpt-Limit-Count") == 0 && param1) {
 	sscanf(param1, "%d", &rcptlimitcnt);
 	if (rcptlimitcnt < 100) rcptlimitcnt = 100;
@@ -153,23 +163,32 @@ static void cfparam(str)
       allow_source_route = 1;
     } else if (cistrcmp(name, "max-error-recipients") == 0 && param1) {
 	sscanf(param1, "%d", &MaxErrorRecipients);
+    } else if (cistrcmp(name, "max-unknown-commands") == 0 && param1) {
+	sscanf(param1, "%d", &unknown_cmd_limit);
+    } else if (cistrcmp(name, "sum-sizeoption-value") == 0) {
+      sum_sizeoption_value = 1;
     }
 
     else if (cistrcmp(name, "use-tcp-wrapper") == 0) {
 	use_tcpwrapper = 1;
     }
 
-    else if (cistrcmp(name, "tarpit") == 0 && param1 && param2) {
+    else if (cistrcmp(name, "tarpit") == 0 && param2 /* 2 params */) {
 	sscanf(param1,"%d",&tarpit_initial);
 	sscanf(param2,"%d",&tarpit_exponent);
     }
 
     /* Two parameter policydb option: DBTYPE and DBPATH */
 
-    else if (cistrcmp(name, "policydb") == 0 && param1 && param2) {
+    else if (cistrcmp(name, "policydb") == 0 && param2 /* 2 params */) {
 	policydefine(&policydb, param1, param2);
     }
 
+    else if (cistrcmp(name, "contentfilter") == 0 && param1) {
+      if (access(param1, X_OK) == 0)
+	contentfilter = strdup(param1);
+    }
+
     /* A few facility enablers: (default: off) */
 
     else if (cistrcmp(name, "debugcmd") == 0) {
@@ -275,8 +294,19 @@ static void cfparam(str)
 #ifdef HAVE_OPENSSL
       sscanf(param1,"%d", & tls_scache_timeout);
 #endif /* - HAVE_OPENSSL */
-
     }
+
+    /* Cluster-wide ETRN support for load-balanced smtp relay use */
+    else if (cistrcmp(name, "etrn-cluster") == 0 && param3 /* 3 params */) {
+      static int idx = 0;
+      if (idx < MAX_ETRN_CLUSTER_IDX) {
+	etrn_cluster[idx].nodename = strdup(param1);
+	etrn_cluster[idx].username = strdup(param2);
+	etrn_cluster[idx].password = strdup(param3);
+	++idx;
+      }
+    }
+
     else {
       /* XX: report error for unrecognized PARAM keyword ?? */
     }
