@@ -4,7 +4,7 @@
  */
 /*
  *    Several extensive changes by Matti Aarnio <mea@nic.funet.fi>
- *      Copyright 1991-1999.
+ *      Copyright 1991-2000.
  */
 /*
  * Zmailer SMTP-server divided into bits
@@ -14,7 +14,6 @@
  *  - HELO/EHLO
  *  - MAIL 
  *  - RCPT
- *  - ETRN/TURNME
  *  - VRFY
  *  - EXPN
  *
@@ -169,11 +168,18 @@ const char *buf, *cp;
     if (logfp) {
       char *s = policymsg(policydb, &SS->policystate);
       if (SS->policyresult != 0 || s != NULL) {
-	fprintf(logfp, "%s\t-- policy result=%d, msg: %s\n", logtag,
+	fprintf(logfp, "%s#\t-- policy result=%d, msg: %s\n", logtag,
 		SS->policyresult, (s ? s : "<NONE!>"));
 	fflush(logfp);
       }
     }
+    if (logfp_to_syslog) {
+      char *s = policymsg(policydb, &SS->policystate);
+      if (SS->policyresult != 0 || s != NULL) {
+	zsyslog((LOG_DEBUG, "%s # policy result=%d, msg: %s", logtag,
+		SS->policyresult, (s ? s : "<NONE!>")));
+      }
+    }
 
     /* actually we accept here about anything, and mark
        our decissions up into the 'policystate' variables */
@@ -403,16 +409,16 @@ int insecure;
 	}
     if (*cp == '\0') {
 	smtp_tarpit(SS);
-	type(SS, 501, m517, "where is <...> in that?");
+	type(SS, 501, m517, "where is <...> in this: %s", cp);
 	return;
     } else if (*cp != '<' && !sloppy) {
 	smtp_tarpit(SS);
-	type(SS, 501, m517, "strangeness between : and <");
+	type(SS, 501, m517, "strangeness between ':' and '<': %s", cp);
 	return;
     }
     if (*(cp + 1) == '<') {
 	smtp_tarpit(SS);
-	type(SS, 501, m517, "there are too many <'s in that!");
+	type(SS, 501, m517, "there are too many <'s in this: %s", cp);
 	return;
     }
     /* "<" [ <a-t-l> ":" ] <localpart> "@" <domain> ">" */
@@ -421,12 +427,13 @@ int insecure;
 	s = rfc821_path(cp, strict || strict_protocol);
 	if (s == cp) {
 	  /* Failure.. ? */
+	  type(SS, -501, m517, "For input: %s", cp);
 	  type821err(SS, 501, m517, buf, "Path data: %.200s", rfc821_error);
 	  return;
 	}
 	if (*s == '>') {
 	  smtp_tarpit(SS);
-	  type(SS, 501, m517, "there are too many >'s in that!");
+	  type(SS, 501, m517, "there are too many >'s in this: <%s", cp);
 	  return;
 	}
 	/* Ok, now it is a moment to see, if we have source routes: @a,@b:c@d */
@@ -446,6 +453,7 @@ int insecure;
 	s = rfc821_path2(cp, 0);
 	if (s == cp && *s != '>') {
 	  /* Failure.. ? */
+	  type(SS, -501, m517, "For input: %s", cp);
 	  type821err(SS, 501, m517, buf, "Path data: %.200s", rfc821_error);
 	  return;
 	}
@@ -463,6 +471,7 @@ int insecure;
 	while (*s == ' ' || *s == '\t') ++s;
 	if (*s != '>') {
 	  rfc821_error_ptr = s;
+	  type(SS, -501, m517, "For input: %s", cp);
 	  type821err(SS, 501, m517, buf, "Missing ending '>' bracket");
 	  return;
 	}
@@ -474,6 +483,7 @@ int insecure;
       s = rfc821_path2(cp, strict);
       if (s == cp) {
 	/* Failure.. */
+	type(SS, -501, m517, "For input: %s", cp);
 	type821err(SS, 501, m517, buf, "Path data: %.200s", rfc821_error);
 	return;
       }
@@ -647,72 +657,85 @@ int insecure;
     SS->policyresult = policytest(policydb, &SS->policystate,
 				  POLICY_MAILFROM, cp, addrlen,
 				  SS->authuser);
-    if (logfp) {
+    if (logfp || logfp_to_syslog) {
       char *ss = policymsg(policydb, &SS->policystate);
       if (SS->policyresult != 0 || ss != NULL) {
 	type(NULL,0,NULL,"-- policy result=%d, msg: %s",
 	     SS->policyresult, (ss ? ss : "<NONE!>"));
-	fflush(logfp);
+	if (logfp)
+	  fflush(logfp);
       }
     }
 
     if (SS->policyresult < 0) {
       char *ss = policymsg(policydb, &SS->policystate);
       if (ss != NULL) {
+	int code = 450;
+	const char *mcode = m471;
+	if (SS->policyresult >= -99) {
+	  code = 553;
+	  mcode = m571;
+	}
 	smtp_tarpit(SS);
-	type(SS, 450, m471, "Policy analysis reported: %s", ss);
+	type(SS, code, mcode, "For MAIL FROM address <%.*s> policy analysis reported: %s",
+	     addrlen, cp, ss);
       } else if (SS->policyresult < -99) {
 	smtp_tarpit(SS);
 	if (SS->policyresult < -103) { /* -104 */
 	  if (!multilinereplies) {
-	    type(SS,450,m443, "Policy analysis reports temporary DNS error with your source domain.");
+	    type(SS,450,m443, "For input: <%.*s> policy analysis reports temporary DNS error with your source domain.", addrlen, cp);
 	  } else {
 	    type(SS, -450, m443, "Policy analysis reports temporary DNS error");
 	    type(SS, -450, m443, "with your source domain.  Retrying may help,");
 	    type(SS, -450, m443, "or if the condition persists, you may need");
-	    type(SS,  450, m443, "to get somebody to fix your DNS servers.");
+	    type(SS,  450, m443, "to get somebody to fix your DNS servers: <%.*s>",
+		 addrlen, cp);
 	  }
 	} else if (SS->policyresult < -100) {
 	  if (!multilinereplies) {
-	    type(SS,450,m443, "Policy analysis reports DNS error with your source domain.");
+	    type(SS,450,m443, "For input: <%.*s> policy analysis reports DNS error with your source domain.", addrlen, cp);
 	  } else {
 	    type(SS, -450, m443, "Policy analysis reports DNS error with your");
 	    type(SS, -450, m443, "source domain.   Please correct your source");
-	    type(SS,  450, m443, "address and/or the info at the DNS.");
+	    type(SS,  450, m443, "address and/or the info at the DNS: <%.*s>",
+		 addrlen, cp);
 	  }
 	} else {
 	  if (!multilinereplies) {
-	    type(SS,450,m471, "Access denied by the policy analysis functions.");
+	    type(SS,450,m471, "For address <%.*s> access denied by the policy analysis functions.", addrlen, cp);
 	  } else {
 	    type(SS, -450, m471, "Access denied by the policy analysis functions.");
 	    type(SS, -450, m471, "This may be due to your source IP address,");
 	    type(SS, -450, m471, "the IP reversal domain, the data you gave for");
 	    type(SS, -450, m471, "the HELO/EHLO parameter, or address/domain you");
-	    type(SS,  450, m471, "gave at the MAIL FROM:<...> address.");
+	    type(SS,  450, m471, "gave at the MAIL FROM:<%.*s> address.",
+		 addrlen, cp);
 	  }
 	}
       } else {
 	char *ss = policymsg(policydb, &SS->policystate);
 	smtp_tarpit(SS);
 	if (ss != NULL) {
-	  type(SS, 550, m571, "Policy analysis reported: %s", ss);
+	  type(SS, 553, m571, "For input address <%.*s> Policy analysis reported: %s", addrlen, cp, ss);
 	} else if (SS->policyresult < -1) {
 	  if (!multilinereplies) {
-	    type(SS,550,m543,"Policy analysis reports DNS error with your source domain.");
+	    type(SS,553,m543,"For MAIL FROM address <%.*s> the policy analysis reports DNS error with your source domain.", addrlen, cp);
 	  } else {
-	    type(SS, -550, m543, "Policy analysis reports DNS error with your");
-	    type(SS, -550, m543, "source domain.   Please correct your source");
-	    type(SS,  550, m543, "address and/or the info at the DNS.");
+	    type(SS, -553, m543, "Policy analysis reports DNS error with your");
+	    type(SS, -553, m543, "source domain.   Please correct your source");
+	    type(SS,  553, m543, "address and/or the info at the DNS: <%.*s>",
+		 addrlen, cp);
 	  }
 	} else {
 	  if (!multilinereplies) {
-	    type(SS,550,m571,"Access denied by the policy analysis functions.");
+	    type(SS,553,m571,"For MAIL FROM address <%.*s> access is denied by the policy analysis functions.", addrlen, cp);
 	  } else {
-	    type(SS, -550, m571, "Access denied by the policy analysis functions.");
-	    type(SS, -550, m571, "This may be due to your source IP address,");
-	    type(SS, -550, m571, "the IP reversal domain, the data you gave for");
-	    type(SS, -550, m571, "the HELO/EHLO parameter, or address/domain you");
-	    type(SS,  550, m571, "gave at the MAIL FROM:<...> address.");
+	    type(SS, -553, m571, "Access denied by the policy analysis functions.");
+	    type(SS, -553, m571, "This may be due to your source IP address,");
+	    type(SS, -553, m571, "the IP reversal domain, the data you gave for");
+	    type(SS, -553, m571, "the HELO/EHLO parameter, or address/domain you");
+	    type(SS,  553, m571, "gave at the MAIL FROM:<%.*s> address.",
+		 addrlen, cp);
 	  }
 	}
       }
@@ -924,7 +947,7 @@ int insecure;
       SS->state = Recipient;
 
     SS->rcpt_count = 0;
-    SS->from_box = (*cp == 0);
+    SS->from_box = (addrlen == 0);
 }
 
 
@@ -949,7 +972,7 @@ const char *buf, *cp;
 	 policydb != NULL && SS->policyresult < 0 ) {
       smtp_tarpit(SS);
       if (!multilinereplies)
-	type(SS, 550, m571, "Access denied by the policy analysis functions.");
+	type(SS, 550, m571, "Access denied by the policy analysis functions by earlier rejection");
       else {
 	type(SS, -550, m571, "Access denied by the policy analysis functions.");
 	type(SS, -550, m571, "This may be due to your source IP address,");
@@ -983,7 +1006,7 @@ const char *buf, *cp;
 
     if (!CISTREQN(cp, "To:", 3)) {
 	smtp_tarpit(SS);
-	type(SS, 501, m552, "where is To: in that?");
+	type(SS, 501, m552, "where is To: in this?  %s", cp);
 	return;
     }
     cp += 3;
@@ -992,27 +1015,27 @@ const char *buf, *cp;
 	if (!isspace(*cp)) {
 	  if (!sloppy) {
 	    smtp_tarpit(SS);
-	    type(SS, 501, m513, "where is <...> in that?");
+	    type(SS, 501, m513, "where is <...> in this?  %s", cp);
 	    return;
 	  }
 	  break; /* Sigh, be sloppy.. */
 	}
     if (*cp == '\0') {
 	smtp_tarpit(SS);
-	type(SS, 501, m513, "where is <...> in that?");
+	type(SS, 501, m513, "where is <...> in this?  %s", cp);
 	return;
     } else if (*cp != '<' && !sloppy) {
 	smtp_tarpit(SS);
-	type(SS, 501, m513, "strangeness between : and <");
+	type(SS, 501, m513, "strangeness between ':' and '<': %s", cp);
 	return;
     } else if (*(cp+1) == '>') {
 	smtp_tarpit(SS);
-	type(SS, 501, m513, "Null address valid only as source");
+	type(SS, 501, m513, "Null address valid only as source: %s", cp);
 	return;
     }
     if (*(cp + 1) == '<') {
 	smtp_tarpit(SS);
-	type(SS, 501, m513, "there are too many <'s in that!");
+	type(SS, 501, m513, "there are too many <'s in this: %s", cp);
 	return;
     }
     if (*cp == '<') {
@@ -1032,7 +1055,7 @@ const char *buf, *cp;
 	}
 	if (*s == '>') {
 	  smtp_tarpit(SS);
-	  type(SS, 501, m513, "there are too many >'s in that!");
+	  type(SS, 501, m513, "there are too many >'s in this: %s", cp);
 	  return;
 	}
 	/* Ok, now it is a moment to see, if we have source routes: @a,@b:c@d */
@@ -1070,7 +1093,7 @@ const char *buf, *cp;
 	while (*s == ' ' || *s == '\t') ++s;
 	if (*s != '>') {
 	  rfc821_error_ptr = s;
-	  type821err(SS, 501, m517, buf, "Missing ending '>' bracket");
+	  type821err(SS, 501, m517, buf, "Missing ending '>' bracket: %s", cp);
 	  return;
 	}
 	++s;
@@ -1185,12 +1208,13 @@ const char *buf, *cp;
     SS->policyresult = policytest(policydb, &SS->policystate,
 				  POLICY_RCPTTO, cp, addrlen,
 				  SS->authuser);
-    if (logfp) {
+    if (logfp || logfp_to_syslog) {
       char *ss = policymsg(policydb, &SS->policystate);
       if (SS->policyresult != 0 || ss != NULL) {
 	type(NULL,0,NULL,"-- policy result=%d, msg: %s",
 	     SS->policyresult, (ss ? ss : "<NONE!>"));
-	fflush(logfp);
+	if (logfp)
+	  fflush(logfp);
       }
     }
 
@@ -1207,12 +1231,13 @@ const char *buf, *cp;
 	    if (rc == 0)
 	      SS->policyresult = 0;
 
-	    if (logfp) {
+	    if (logfp || logfp_to_syslog) {
 	      char *ss = policymsg(policydb, &SS->policystate);
 	      if (SS->policyresult != 0 || ss != NULL) {
 		type(NULL,0,NULL,"-- policy result=%d, msg: %s",
 		     SS->policyresult, (ss ? ss : "<NONE!>"));
-		fflush(logfp);
+		if (logfp)
+		  fflush(logfp);
 	      }
 	    }
 	  }
@@ -1230,67 +1255,74 @@ const char *buf, *cp;
 
 	if (SS->policyresult < -99) { /* "soft error, 4XX code */
 	  if (ss != NULL) {
-	    type(SS, 450, m471, "Policy analysis reported: %s", ss);
+	    type(SS, 450, m471, "For recipient address <%.*s> the policy analysis reported: %s", addrlen, cp, ss);
 	  } else if (SS->policyresult < -103) { /* -104 */
 	    if (!multilinereplies)
-	      type(SS, 450, m443, "Policy analysis reports temporary DNS error with the target domain.");
+	      type(SS, 450, m443, "Policy analysis reports temporary DNS error with the target domain: <%.*s>", addrlen, cp);
 	    else {
 	      type(SS, -450, m443, "Policy analysis reports temporary DNS error");
 	      type(SS, -450, m443, "with this target domain. Retrying may help,");
 	      type(SS, -450, m443, "or if the condition persists, some further");
 	      type(SS, -450, m443, "work may be in need with the target domain");
-	      type(SS,  450, m443, "DNS servers.");
+	      type(SS,  450, m443, "DNS servers; RCPT TO:<%.*s>", addrlen, cp);
 	    }
 
 	  } else if (SS->policyresult < -102) {
 	    /* Code: -103 */
 	    if (!multilinereplies) {
-	      type(SS,450, m471, "This target address is not our MX service client.");
+	      type(SS,450, m471, "This target address is not our MX service client: <%.*s>", addrlen, cp);
 	    } else {
 	      type(SS,-450, m471, "This target address is not our MX service");
 	      type(SS,-450, m471, "client, nor you are connecting from address");
 	      type(SS,-450, m471, "that is allowed to openly use us to relay");
 	      type(SS,-450, m471, "to any arbitary address thru us.");
-	      type(SS, 450, m471, "We don't accept this recipient.");
+	      type(SS, 450, m471, "We don't accept this recipient: <%.*s>",
+		   addrlen, cp);
 	    }
 	  } else if (SS->policyresult < -100) {
 	    /* Code: -102 */
 	    if (!multilinereplies)
-	      type(SS, 450, m443, "Policy analysis found DNS error on the target address");
+	      type(SS, 450, m443, "Policy analysis found DNS error on the target address: <%.*s>", addrlen, cp);
 	    else {
 	      type(SS,-450, m443, "Policy analysis found DNS error on");
 	      type(SS,-450, m443, "the target address. This address is");
-	      type(SS, 450, m443, "not currently acceptable.");
+	      type(SS, 450, m443, "not currently acceptable: <%.*s>",
+		   addrlen, cp);
 	    }
 	  } else {
-	    type(SS, 450, m443, "Policy rejection on the target address");
+	    type(SS,450,m443, "Policy rejection on the target address: <%.*s>",
+		 addrlen, cp);
 	  }
 	} else {
 	  if (ss != NULL) {
-	    type(SS, 550, m571, "Policy analysis reported: %s", ss);
+	    type(SS, 550, m571, "Policy analysis reported: %s rcpt=<%.*s>",
+		 ss, addrlen, cp);
 	  } else if (SS->policyresult < -2) {
 	    /* Code: -3 */
 	    if (!multilinereplies)
-	      type(SS,550,m571, "This target address is not our MX service client.");
+	      type(SS,550,m571, "This target address is not our MX service client: <%.*s>", addrlen, cp);
 	    else {
 	      type(SS,-550, m571, "This target address is not our MX service");
 	      type(SS,-550, m571, "client, nor you are connecting from address");
 	      type(SS,-550, m571, "that is allowed to openly use us to relay");
 	      type(SS,-550, m571, "to any arbitary address thru us.");
-	      type(SS, 550, m571, "We don't accept this recipient.");
+	      type(SS, 550, m571, "We don't accept this recipient: <%.*s>",
+		   addrlen, cp);
 	    }
 
 	  } else if (SS->policyresult < -1) {
 	    /* Code: -2 */
 	    if (!multilinereplies) {
-	      type(SS,550,m543, "Policy analysis found DNS error on the target domain.");
+	      type(SS,550,m543, "Policy analysis found DNS error on the target domain: <%.*s>", addrlen, cp);
 	    } else {
 	      type(SS,-550, m543, "Policy analysis found DNS error on");
 	      type(SS,-550, m543, "the target address. This address is");
-	      type(SS, 550, m543, "not currently acceptable.");
+	      type(SS, 550, m543, "not currently acceptable: <%.*s>",
+		   addrlen, cp);
 	    }
 	  } else {
-	    type(SS, 550, m571, "Policy rejection on the target address");
+	    type(SS,550,m571, "Policy rejection on the target address: <%.*s>",
+		 addrlen, cp);
 	  }
 	}
 	if (newcp)
@@ -1356,15 +1388,17 @@ const char *buf, *cp;
     if (SS->sizeoptval < 0)
 	SS->sizeoptval = 0;
 
-    /* SS->sizeoptsum += SS->sizeoptval; */
-    SS->sizeoptsum = SS->sizeoptval;
+    if (sum_sizeoption_value)
+      SS->sizeoptsum += SS->sizeoptval;
+    else
+      SS->sizeoptsum = SS->sizeoptval;
 
     if (ferror(SS->mfp)) {
 	smtp_tarpit(SS);
 	type(SS, 452, m430, (char *) NULL);
-    } else if (maxsize > 0 && SS->sizeoptsum > maxsize) {
+    } else if (maxsize > 0 && SS->sizeoptval > maxsize) {
 	smtp_tarpit(SS);
-	type(SS, 552, m534, "Message size exceeds fixed maximum size of %ld chars for acceptable email", maxsize);
+	type(SS, 552, m534, "Message size exceeds fixed maximum size of %ld chars for acceptable email; rcpt=<%.*s>", maxsize, addrlen, cp);
     } else if (SS->sizeoptsum > availspace) {
 	smtp_tarpit(SS);
 	type(SS, 452, m431, "insufficient storage space, try again later");
@@ -1380,11 +1414,11 @@ const char *buf, *cp;
 	    smtp_tarpit(SS);
 	    type(SS, 552, m571, "SPAM trap -- too many recipients for an empty source address!");
 	} else if (SS->sizeoptval)
-	    type(SS, 250, "2.1.5", "Ok; can accomodate %d byte message%s",
-		 SS->sizeoptval, srcrtestatus);
+	    type(SS, 250, "2.1.5", "Ok; can accomodate %d byte message%s for <%.*s>",
+		 SS->sizeoptval, srcrtestatus, addrlen, cp);
 	else
-	    type(SS, 250, "2.1.5", "Recipient address syntax Ok%s",
-		 srcrtestatus);
+	    type(SS, 250, "2.1.5", "Recipient address syntax Ok%s; rcpt=<%.*s>",
+		 srcrtestatus, addrlen, cp);
 	SS->rcpt_count += 1;
     }
     if (s)
@@ -1392,54 +1426,6 @@ const char *buf, *cp;
     SS->state = RecipientOrData;
 }
 
-void smtp_turnme(SS, name, cp)
-SmtpState *SS;
-const char *name, *cp;
-{
-    FILE *mfp;
-    int rc;
-
-    while (*cp == ' ' || *cp == '\t') ++cp;
-    if (*cp == 0) {
-	type(SS, 552, "5.0.0", "ETRN needs target domain name parameter.");
-	typeflush(SS);
-	return;
-    }
-
-    if (!((*cp >= 'A' && *cp <= 'Z') || (*cp >= 'a' && *cp <= 'z') ||
-	  (*cp >= '0' && *cp <= '9'))) {
-      /* Has some special character beginning it; we don't support
-	 either arbitary subdomains (@foo.dom), nor "channel-based"
-	 starting (#foo) */
-      type(SS, 458, m571, "Sorry, only literal target domains accepted");
-      typeflush(SS);
-      return;
-    }
-
-    mfp = mail_open(MSG_RFC822);
-    if (!mfp) {
-	type(SS, 452, m400, "Failed to initiate ETRN request;  Disk full?");
-	typeflush(SS);
-	return;
-    }
-    fprintf(mfp, "%c%c%s\n", _CF_TURNME, _CFTAG_NORMAL, cp);
-    /* printf("050-My uid=%d/%d\r\n",getuid(),geteuid()); */
-    runasrootuser();
-    rc = mail_close_alternate(mfp, TRANSPORTDIR, "");
-    runastrusteduser();
-    if (rc) {
-	type(SS,452,m400,"Failed to initiate ETRN request; Permission denied?");
-    } else {
-      if (multilinereplies) {
-	type(SS,-250,m200,"An ETRN request is initiated - lets hope the system");
-	type(SS,-250,m200,"has resources to honour it. We call the remote,");
-	type(SS, 250,m200,"if we have anything to send there.");
-      } else {
-	type(SS, 250, m200, "An ETRN request is submitted - something may get sent.");
-      }
-    }
-    typeflush(SS);
-}
 
 void smtp_verify(SS, buf, cp)
 SmtpState *SS;
