@@ -4,22 +4,34 @@
  */
 /*
  *    Several extensive changes by Matti Aarnio <mea@nic.funet.fi>
- *      Copyright 1991-1997.
+ *      Copyright 1991-1999.
  */
 /*
  * Zmailer SMTP-server divided into bits
  *
  * The basic commands:
  *
- *  - DATA
- *  - BDAT
+ *  - DATA (RFC 821)
+ *  - BDAT (RFC 1830)
  *
  */
 
+/* XX: for anti-spam hack */
+/* #define USE_ANTISPAM_HACKS */
+/* #define USE_STRICT_MSGID_FREEZING */
+
+#define FREEZE__X_ADVERTISEMENT_FOUND                   951
+#ifdef  USE_ANTISPAM_HACKS
+#define FREEZE__X_UIDL_FOUND                            952
+#define FREEZE__IMPROBABLE_RECEIVED_HEADER_FOUND        952
+#define FREEZE__MALFORMED_MESSAGE_ID_HEADER             953
+#endif
+
+#include "smtpserver.h"
+
 #ifdef USE_TRANSLATION
 #include <libtrans.h>
 #endif				/* USE_TRANSLATION */
-#include "smtpserver.h"
 
 static int mvdata __((SmtpState *, char *));
 static int mvbdata __((SmtpState *, char *, long));
@@ -34,6 +46,12 @@ const char *buf, *cp;
     int ino;
     char msg[2048];
 
+    while (!strict_protocol && (*cp == ' ' || *cp == '\t')) ++cp;
+    if (strict_protocol && *cp != 0) {
+      type(SS, 501, m554, "Extra junk after 'DATA' verb");
+      return 0;
+    }
+
     if (SS->state != RecipientOrData) {
 	switch (SS->state) {
 	case Hello:
@@ -97,12 +115,24 @@ const char *buf, *cp;
     type(SS, 354, NULL, (char *) NULL);
     typeflush(SS);
     fputs("env-end\n", SS->mfp);
+
+    if (msa_mode && SS->authuser != NULL ) {
+      fprintf(SS->mfp, "X-Comment: RFC 2476 MSA function at %s logged sender identity as: %s\n", SS->myhostname, SS->authuser);
+    }
+
     /* We set alarm()s inside the mvdata() */
     *msg = 0;
     filsiz = mvdata(SS, msg);
     alarm(0);			/* cancel the alarm() */
     tell = ftell(SS->mfp);
 
+    report(SS, "Got '.'; tell=%ld", tell);
+
+    availspace = fd_statfs(FILENO(SS->mfp));
+    if (availspace < 0)
+	availspace = 2000000000;	/* Over 2G ? */
+    availspace -= minimum_availspace;
+
     if (*msg != 0) {
 	mail_abort(SS->mfp);
 	SS->mfp = NULL;
@@ -110,58 +140,89 @@ const char *buf, *cp;
 	typeflush(SS);
     } else if (s_feof(SS)) {
 	/* [mea@utu.fi] says this can happen */
-	mail_abort(SS->mfp);
+	if (STYLE(SS->cfinfo,'D')) {
+	  /* Says: DON'T DISCARD -- aka DEBUG ERRORS! */
+	  mail_close_alternate(SS->mfp,"public",".DATA-EOF");
+	} else
+	  mail_abort(SS->mfp);
 	SS->mfp = NULL;
 	reporterr(SS, tell, "premature EOF on DATA input");
 	typeflush(SS);
 	return -1;
-    } else if (ferror(SS->mfp)) {
-	type(SS, 452, m430, (char *) NULL);
+    } else if (availspace < 0 || ferror(SS->mfp)) {
+	type(SS, 452, m430, NULL); /* insufficient system storage */
 	typeflush(SS);
+	reporterr(SS, tell, ferror(SS->mfp) ? "write to spool file failed" : "system free storage under limit");
 	clearerr(SS->mfp);
 	mail_abort(SS->mfp);
 	SS->mfp = NULL;
-	reporterr(SS, tell, "message file error");
     } else if (maxsize > 0 && filsiz > maxsize) {
 	mail_abort(SS->mfp);
 	SS->mfp = NULL;
-	type(SS, 552, "5.3.4", "Message size exceeded fixed maximum size of %ld chars for acceptable email", maxsize);
+	type(SS, 552, "5.3.4", "Size of this message exceeds the fixed maximum size of  %ld  chars for received email ", maxsize);
 	typeflush(SS);
     } else {
-	if (SS->policyresult > 0) {
-	    char polbuf[20];
-	    struct stat stbuf;
-	    fflush(SS->mfp);
-	    fstat(FILENO(SS->mfp), &stbuf);
-	    runasrootuser();
-	    sprintf(polbuf,"policy-%d",SS->policyresult);
-	    if (mail_close_alternate(SS->mfp, FREEZERDIR, polbuf) != 0) {
-		type(SS, 452, m430, (char *) NULL);
-		typeflush(SS);
-		SS->mfp = NULL;
-		reporterr(SS, tell, "message file close failed");
-	    } else {
-		static int freezecnt = 1;
-		freezecnt <<= 1;
-		sleep(freezecnt);
-		type(SS, 250, "2.6.0", "message accepted; into freezer[%d] area...",SS->policyresult);
-		typeflush(SS);
-		SS->mfp = NULL;
-		zsyslog((LOG_INFO, "accepted id %d (%dc) from %s/%d into freeze[%d]",
-			 (int) stbuf.st_ino, (int) stbuf.st_size,
-			 SS->rhostname, SS->rport, SS->policyresult));
-	    }
-	    runastrusteduser();
-	} else if (_mail_close_(SS->mfp, &ino, &mtime) == EOF) {
+
+	/* Things have been good thus far, now we store
+	   the resulting file into router spool area;
+	   pending a few things we do at first.. */
+
+	/* Lets see what the content-policy will tell now ? */
+	char *fname = mail_fname(SS->mfp);
+	SS->policyresult = contentpolicy(policydb, &SS->policystate, fname);
+
+	if (SS->policyresult < 0) {
+	  char *ss = policymsg(policydb, &SS->policystate);
+	  type(NULL,0,NULL,
+	       "Content-policy analysis ordered message rejection. (code=%d)", SS->policyresult);
+	  type(SS, -552,m571, "Content-Policy analysis rejected this message");
+	  type(SS, 552, m571, "Content-Policy msg: %s", ss ? ss : "rejected");
+	  mail_abort(SS->mfp);
+	  SS->mfp = NULL;
+	} else if (SS->policyresult > 0) {
+	  char polbuf[20];
+	  struct stat stbuf;
+	  char *ss = policymsg(policydb, &SS->policystate);
+
+	  fflush(SS->mfp);
+	  fstat(FILENO(SS->mfp), &stbuf);
+	  runasrootuser();
+	  sprintf(polbuf,"policy-%d",SS->policyresult);
+	  if (mail_close_alternate(SS->mfp, FREEZERDIR, polbuf) != 0) {
+	    type(NULL,0,NULL,
+		 "mail_close_alternate(..'FREEZER','%s') failed, errno=%d (%s)",
+		 polbuf, errno, strerror(errno));
+	    type(SS, 452, m430, "Message file disposition failed");
+	    typeflush(SS);
+	    SS->mfp = NULL;
+	    reporterr(SS, tell, "message file close failed");
+	  } else {
+	    static int freezecnt = 1;
+	    freezecnt <<= 1;
+	    sleep(freezecnt);
+	    type(SS, 250, "2.6.0", "message accepted; into freezer[%d] area; %s", SS->policyresult, ss ? ss : "");
+	    typeflush(SS);
+	    SS->mfp = NULL;
+	    zsyslog((LOG_INFO, "accepted id %d (%dc) from %s/%d into freeze[%d]",
+		     (int) stbuf.st_ino, (int) stbuf.st_size,
+		     SS->rhostname, SS->rport, SS->policyresult));
+	  }
+	  runastrusteduser();
+	} else {
+
+	  /*  Ok, we didn't have smtp-policy defined freezer action,
+	      lets see if we do it some other way.. */
+
+	  if (_mail_close_(SS->mfp, &ino, &mtime) == EOF) {
 	    type(SS, 452, m430, (char *) NULL);
 	    typeflush(SS);
 	    SS->mfp = NULL;
 	    reporterr(SS, tell, "message file close failed");
-	} else {
-	    /* Ok, build responce with proper "spoolid" */
+	  } else {
+	    /* Ok, build response with proper "spoolid" */
 	    char fnam[20], taspid[30];
 	    sprintf(fnam, "%d", ino);
-	    taspoolid(taspid, sizeof(taspid), mtime, fnam);
+	    taspoolid(taspid, mtime, (long)ino);
 
 	    SS->mfp = NULL;
 	    type(SS, 250, "2.6.0", "%s message accepted", taspid);
@@ -170,14 +231,15 @@ const char *buf, *cp;
 	    if (smtp_syslog)
 	      zsyslog((LOG_INFO,
 		       "%s: (%ldc) accepted from %s/%d", taspid, tell,
-                       SS->rhostname, SS->rport));
-
-	    if (logfp != NULL) {
-		fprintf(logfp, "%d#\t%s: %ld bytes\n", pid, taspid, tell);
-		fflush(logfp);
-	    }
+		       SS->rhostname, SS->rport));
+		
+	    type(NULL,0,NULL,"%s: %ld bytes", taspid, tell);
+	    if (logfp)
+	      fflush(logfp);
+	  }
 	}
     }
+
     SS->state = MailOrHello;
     typeflush(SS);
     return 0;
@@ -199,7 +261,7 @@ const char *buf, *cp;
 	SS->mvbstate = -1;
     }
     *msg = 0;
-    rc = sscanf(cp, "%ld %7s %7s", &bdata_chunksize, msg, msg + 8);
+    rc = sscanf(cp, "%ld %7s %7s", &bdata_chunksize, msg, msg + 20);
     SS->bdata_blocknum += 1;
     bdata_last = CISTREQ(msg, "LAST");
     if (!(bdata_chunksize > 0L
@@ -210,6 +272,9 @@ const char *buf, *cp;
     }
     if (SS->bdata_blocknum == 1) {
 	fputs("env-end\n", SS->mfp);
+	if (msa_mode && SS->authuser != NULL ) {
+	  fprintf(SS->mfp, "X-Comment: RFC 2476 MSA function at %s logged sender identity as: %s\n", SS->myhostname, SS->authuser);
+	}
     }
     /* We set alarm()s inside the mvbdata() */
     *msg = 0;
@@ -220,6 +285,8 @@ const char *buf, *cp;
     else
 	tell = 0;
 
+    report(SS, "BDAT; last=%d tell=%ld", bdata_last, tell);
+
     if (SS->state != BData) {
 	switch (SS->state) {
 	case Hello:
@@ -268,6 +335,12 @@ const char *buf, *cp;
 	SS->rcpt_count = 0;	/* now we can zero them.. */
 	SS->sender_ok = 0;
     }
+
+    availspace = fd_statfs(FILENO(SS->mfp));
+    if (availspace < 0)
+	availspace = 2000000000;	/* Over 2G ? */
+    availspace -= minimum_availspace;
+
     /* The common typeflush() is at the end... */
     if (SS->mfp == NULL) {
 	type(SS, 503, m552, "BDAT block discarded due to earlier error");
@@ -277,66 +350,90 @@ const char *buf, *cp;
 	type(SS, 452, "%s", msg);
     } else if (s_feof(SS)) {
 	/* [mea@utu.fi] says this can happen */
-	mail_abort(SS->mfp);
+	if (STYLE(SS->cfinfo,'D')) {
+	  /* Says: DON'T DISCARD -- aka DEBUG ERRORS! */
+	  mail_close_alternate(SS->mfp,"public",".BDAT-EOF");
+	} else
+	  mail_abort(SS->mfp);
 	SS->mfp = NULL;
 	reporterr(SS, tell, "premature EOF on BDAT input");
 	typeflush(SS);
 	return -1;
-    } else if (ferror(SS->mfp)) {
+    } else if (availspace < 0 || ferror(SS->mfp)) {
 	type(SS, 452, m400, (char *) NULL);
+	reporterr(SS, tell, ferror(SS->mfp) ? "write to spool file failed" : "system free storage under limit");
+	clearerr(SS->mfp);
 	mail_abort(SS->mfp);
 	SS->mfp = NULL;
-	reporterr(SS, tell, "write to spool file failed");
     } else if (maxsize > 0 && tell > maxsize) {
 	mail_abort(SS->mfp);
 	SS->mfp = NULL;
-	type(SS, 552, "5.3.4", "Message size exceeded fixed maximum size of %ld chars for acceptable email", maxsize);
+	type(SS, 552, "5.3.4", "Size of this message exceeds the fixed maximum size of  %ld  chars for received email ", maxsize);
     } else if (bdata_last) {
 	time_t mtime;
 	int inum;
 
-	if (SS->policyresult > 0) {
-	    struct stat stbuf;
-	    fflush(SS->mfp);
-	    fstat(FILENO(SS->mfp), &stbuf);
-	    runasrootuser();
-	    if (mail_close_alternate(SS->mfp, FREEZERDIR, "policy") != 0) {
-		type(SS, 452, m430, (char *) NULL);
-		SS->mfp = NULL;
-		reporterr(SS, tell, "message file close failed");
-	    } else {
-		type(SS, 250, "2.6.0", "message accepted; into freezer area...");
-		SS->mfp = NULL;
-		zsyslog((LOG_INFO, "accepted id %d (%dc) from %s/%d into freeze",
-			 (int) stbuf.st_ino, (int) stbuf.st_size,
-			 SS->rhostname, SS->rport));
-	    }
-	    runastrusteduser();
-	} else if (_mail_close_(SS->mfp, &inum, &mtime) == EOF) {
-	    type(SS, 452, m400, (char *) NULL);
+	/* Things have been good thus far, now we store
+	   the resulting file into router spool area;
+	   pending a few things we do at first.. */
+
+	/* Lets see what the content-policy will tell now ? */
+	char *fname = mail_fname(SS->mfp);
+	SS->policyresult = contentpolicy(policydb, &SS->policystate, fname);
+
+	if (SS->policyresult < 0) {
+	  char *ss = policymsg(policydb, &SS->policystate);
+type(NULL,0,NULL,
+  "Content-policy analysis ordered message rejection. (code=%d)", SS->policyresult);
+	  type(SS, -552,m571, "Content-Policy analysis rejected this message");
+	  type(SS, 552, m571, "Content-Policy msg: %s", ss ? ss : "rejected");
+	  mail_abort(SS->mfp);
+	  SS->mfp = NULL;
+	} else if (SS->policyresult > 0) {
+	  struct stat stbuf;
+	  char *ss = policymsg(policydb, &SS->policystate);
+	  fflush(SS->mfp);
+	  fstat(FILENO(SS->mfp), &stbuf);
+	  runasrootuser();
+	  if (mail_close_alternate(SS->mfp, FREEZERDIR, "policy") != 0) {
+	    type(NULL,0,NULL,
+		 "mail_close_alternate(..'FREEZER','%s') failed, errno=%d (%s)",
+		 "policy", errno, strerror(errno));
+	    if (logfp)
+	      fflush(logfp);
+	    type(SS, 452, m430, "Message file disposition failed");
 	    SS->mfp = NULL;
 	    reporterr(SS, tell, "message file close failed");
-	} else {
-	    /* Ok, build responce with proper "spoolid" */
-	    char fnam[20], taspid[30];
-	    sprintf(fnam, "%d", inum);
-	    taspoolid(taspid, sizeof(taspid), mtime, fnam);
-
+	  } else {
+	    type(SS, 250, "2.6.0", "message accepted; into freezer[%d] area; %s", SS->policyresult, ss ? ss : "");
 	    SS->mfp = NULL;
-	    type(SS, 250, "2.6.0", "%s Roger, got %ld bytes in the last chunk, stored %ld bytes into spool",
-		 taspid, bdata_chunksize, (long) tell);
-	    if (logfp)
-		fprintf(logfp, "%d#\t-- pipeline input: %d bytes\n", pid, s_hasinput(SS));
+	    zsyslog((LOG_INFO, "accepted id %d (%dc) from %s/%d into freeze",
+		     (int) stbuf.st_ino, (int) stbuf.st_size,
+		     SS->rhostname, SS->rport));
+	  }
+	  runastrusteduser();
+	} else if (_mail_close_(SS->mfp, &inum, &mtime) == EOF) {
+	  type(SS, 452, m400, (char *) NULL);
+	  SS->mfp = NULL;
+	  reporterr(SS, tell, "message file close failed");
+	} else {
+	  /* Ok, build response with proper "spoolid" */
+	  char taspid[30];
+	  taspoolid(taspid, mtime, inum);
 
-	    if (smtp_syslog)
-	      zsyslog((LOG_INFO,
-		       "%s: (%ldc) accepted from %s/%d", taspid, tell,
-		       SS->rhostname, SS->rport));
+	  SS->mfp = NULL;
+	  type(SS, 250, "2.6.0", "%s Roger, got %ld bytes in the last chunk, stored %ld bytes into spool",
+	       taspid, bdata_chunksize, (long) tell);
+	  type(NULL,0,NULL,"-- pipeline input: %d bytes",s_hasinput(SS));
 
-	    if (logfp != NULL) {
-		fprintf(logfp, "%d#\t%s: %ld bytes\n", pid, taspid, tell);
-		fflush(logfp);
-	    }
+	  if (smtp_syslog)
+	    zsyslog((LOG_INFO,
+		     "%s: (%ldc) accepted from %s/%d", taspid, tell,
+		     SS->rhostname, SS->rport));
+
+	  type(NULL,0,NULL,"%s: %ld bytes", taspid, tell);
+	  if (logfp)
+	    fflush(logfp);
 	}
     } else {			/* Not last chunk! */
 	type(SS, 250, "2.6.0", "Received %ld bytes", bdata_chunksize);
@@ -413,11 +510,119 @@ static char indexnum[256 + 1] =
     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
 };
 
+/*
+ * BASE64 DECODER index table, and ENCODER map array...
+ */
+static int base64decode_index[128] = {
+	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
+	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
+	52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
+	-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14,
+	15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
+	-1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
+	41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1
+};
+#define decodechar64(c)  (((c) < 0 || (c) > 127) ? -1 : base64decode_index[(c)])
+
+static char base64encode_array[64] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
+
+int decodebase64string(instr,inlen,outstr,outspc,inleftover)
+     const char * instr;
+     char * outstr;
+     const char ** inleftover;
+     int inlen, outspc;
+{
+    int b64decoding = 1, outlen = 0;
+    int b64eod = 0, b64i = 0, i, c;
+    char b64c[4];
+
+    while (inlen > 0 && outlen < outspc && !b64eod) {
+      c = *instr++; --inlen;
+
+      if (c == '=' || !b64decoding) {
+	b64eod = 1;
+	continue;
+      }
+      i = decodechar64(c);
+      if (i < 0) continue;
+
+      b64c[b64i++] = i;
+      if (b64i < 2)
+	continue;
+      if (b64i == 2) {
+	c = (b64c[0] << 2) | ((b64c[1] & 0x30) >> 4);
+      } else if (b64i == 3) {
+	c = ((b64c[1] & 0x0f) << 4) | ((b64c[2] & 0x3c) >> 2);
+      } else {
+	c = (b64c[2] << 6) | b64c[3];
+	b64i = 0;
+	if (b64eod)
+	  b64decoding = 0;
+      }
+
+      outstr[outlen] = c;
+      ++outlen;
+    }
+    if (outlen < outspc)
+      outstr[outlen] = 0;
+    while (*instr == '=') ++instr;
+    if (inleftover)
+      *inleftover = instr;
+    return outlen;
+}
+
+int encodebase64string(instr,inlen,outstr,outspc)
+     const char * instr;
+     char *outstr;
+     int inlen, outspc; /* Always positive values .. */
+{
+    /* Build groups of 3 bytes, encode them in 4 chars; if some byte
+       is not filled, mark the incomplete byte with '=' */
+    u_char b64out[4];
+    int b64i, outlen = 0, b64val;
+    while (inlen > 0 && outlen < outspc) {
+      b64i = inlen > 3 ? 3 : inlen;
+
+      b64val    = ((unsigned char) instr[0]) << 16;
+      if (b64i > 1)
+	b64val |= ((unsigned char) instr[1]) << 8;
+      if (b64i > 2)
+	b64val |= ((unsigned char) instr[2]);
+
+      b64out[3] = base64encode_array[ b64val & 63 ]; b64val >>= 6;
+      b64out[2] = base64encode_array[ b64val & 63 ]; b64val >>= 6;
+      b64out[1] = base64encode_array[ b64val & 63 ]; b64val >>= 6;
+      b64out[0] = base64encode_array[ b64val & 63 ];
+
+      switch(b64i) {
+      case 1:
+	b64out[2] = '=';
+      case 2:
+	b64out[3] = '=';
+      }
+
+      instr += b64i;
+      inlen -= b64i;
+
+      b64i = (b64i == 1) ? 3 : 4;
+      if ((outlen + b64i) < outspc) {	/* Can fit in .. */
+	memcpy(outstr+outlen, b64out, b64i);
+	outlen += b64i;
+      } else {				/* Can't fit in :-( */
+	memcpy(outstr+outlen, b64out, outspc-outlen);
+	outlen = outspc;
+      }
+    }
+    return outlen;
+}
+
 /*
  * Copy bytes from stdin to out, obeying sensible SMTP DATA input heuristics.
  *
- * If you can improve on this heavily optimized routine, I'd like to see it.
- * This version goes at better than 100kB/cpu-sec on a Sun 3/180.
+ * Rayan back in 1988:
+ *  "If you can improve on this heavily optimized routine, I'd like to see it.
+ *   This version goes at better than 100kB/cpu-sec on a Sun 3/180."
  */
 
 static int /* count of bytes */ mvdata(SS, msg)
@@ -443,9 +648,8 @@ char *msg;
     int insubject = 0;
     int has8bit = 0;		/* In headers */
     int has8bitsum = 0;
-    int from__err = 0;
+    /* int from__err = 0; */
     int linecnt = 0;
-    int was8bit = 0;
 #ifdef USE_TRANSLATION
     int wi;
     char hdr_cte[4000], hdr_ct[4000];
@@ -459,18 +663,6 @@ char *msg;
     int qp_chars = 0, qp_hex = 0;
     int b64decoding = 1, b64eod = 0, b64i = 0;
     char b64c[4];
-    static char index_64[128] =
-    {
-	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
-	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1,
-	-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, 62, -1, -1, -1, 63,
-	52, 53, 54, 55, 56, 57, 58, 59, 60, 61, -1, -1, -1, -1, -1, -1,
-	-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14,
-	15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, -1, -1, -1, -1, -1,
-	-1, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
-	41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1
-    };
-#define char64(c)  (((c) < 0 || (c) > 127) ? -1 : index_64[(c)])
 #endif				/* USE_TRANSLATION */
 
     typeflush(SS);
@@ -516,7 +708,7 @@ char *msg;
 			else
 			    header_from_mime(linebuf, &col, sizeof(linebuf));
 			for (wi = 0; wi < col; ++wi)
-			    putc(TR_IN(linebuf[wi]), SS->mfp);
+			    fputc(TR_IN(linebuf[wi]), SS->mfp);
 #else				/* USE_TRANSLATION */
 			if (has8bit)
 			    header_to_mime(linebuf, &col, sizeof(linebuf));
@@ -567,12 +759,13 @@ char *msg;
 		do_decode = cte;
 	    if (X_translation && X_8bit && ct_is_text && (X_settrrc == 0))
 		do_translate = 1;
-	    if (logfp)
-		fprintf(logfp, "%d#\t(8bit decode: %s, translate: %s) [%s%s,%s]\n", pid,
-		   do_decode ? "YES" : "NO", do_translate ? "YES" : "NO",
-			X_translation ? "-X " : "",
-			X_8bit ? "-8" : "",
-			ct_is_text ? "text" : "non-text");
+
+	    type(NULL,0,NULL,"(8bit decode: %s, translate: %s) [%s%s,%s]",
+		 do_decode ? "YES" : "NO", do_translate ? "YES" : "NO",
+		 X_translation ? "-X " : "",
+		 X_8bit ? "-8" : "",
+		 ct_is_text ? "text" : "non-text");
+
 	    /* write out content-type and content-transfer-encoding */
 	    if (delay_ct) {
 		if (do_translate) {
@@ -643,7 +836,7 @@ char *msg;
 		header_from_mime(linebuf, &col, sizeof(linebuf));
 	    if (col > 0)
 		for (wi = 0; wi < col; wi++)
-		    putc(TR_IN(linebuf[wi]), SS->mfp);
+		    fputc(TR_IN(linebuf[wi]), SS->mfp);
 #else				/* USE_TRANSLATION */
 	    if (has8bit)
 		header_to_mime(linebuf, &col, sizeof(linebuf));
@@ -671,14 +864,80 @@ char *msg;
 		has8bitsum += has8bit;
 	    }
 	}
-	if (strncmp(linebuf, "X-Advertisement:",16)==0) {
+	/* XX: The anti-spam hacks. To differentiate between the auto-freeze of
+	 * allegedly-spam messages and freezes resulting from user-specified
+	 * smtp-policy.src, we do not use 1 to freeze the messages.
+	 */
+	if (CISTREQN(linebuf, "X-Advertisement:",16)) {
 	  /* Gee... Only SPAMmers (Cyberpromo!) use this .. (I hope..) */
-	  SS->policyresult = 1; /* Mark it for the freezer! */
+	  SS->policyresult = FREEZE__X_ADVERTISEMENT_FOUND;
+	  type(NULL,0,NULL,"Found X-Advertisement header");
 	}
-#if 0
+	if (CISTREQN(linebuf, "X-Advertisment:",15)) {
+	  /* Gee... Only SPAMmers (Cyberpromo!) use this .. (I hope..) */
+	  SS->policyresult = FREEZE__X_ADVERTISEMENT_FOUND;
+	  type(NULL,0,NULL,"Found X-Advertisment header");
+	}
+#ifdef USE_ANTISPAM_HACKS
 	if (strncmp(linebuf, "X-UIDL:", 7)==0) {
 	  /* Sigh... SPAMmers use this .. but it is valid AOL too.. */
-	  SS->policyresult = 1; /* Mark it for the freezer! */
+	  SS->policyresult = FREEZE__X_UIDL_FOUND;
+	}
+#endif
+#ifdef USE_ANTISPAM_HACKS
+	/* This test probably doesn't work */
+	if (CISTREQN(linebuf, "Received:", 9)) {
+	  /* Scan for "(really ". Anything with this string in the Received
+	   * header is highly likely to have a forged Received header
+	   * characteristic of spams. Unfortunately this string may
+	   * also be a result of pure coincidence.
+	   */
+	  s = linebuf + 11;
+	  while (*s) {
+	    if (CISTREQN(s, "(really ", 8)) {
+	      SS->policyresult = FREEZE__IMPROBABLE_RECEIVED_HEADER_FOUND;
+	      type(NULL,0,NULL,"Improbable Received: header");
+	      break;
+	    }
+	    ++s;
+	  }
+	}
+#endif
+#ifdef USE_ANTISPAM_HACKS
+	if (CISTREQN(linebuf, "Message-ID:", 11)) {
+	  /* Freeze any mail with no message id in the message-id header,
+	   * or a message id with obvious syntax errors, or message id
+	   * with junk after it. These are highly likely to be spam, though
+	   * they might only be a result of buggy software. (MS Exchange?)
+	   */
+	  s = linebuf + 11;
+	  while (*s == ' ' || *s == '\t')
+	    ++s;
+	  if (*s != '<') {
+	    SS->policyresult = FREEZE__MALFORMED_MESSAGE_ID_HEADER;
+	    type(NULL,0,NULL,"No <> around Message-Id");
+	  } else if (s[1] == '@') {
+	    SS->policyresult = FREEZE__MALFORMED_MESSAGE_ID_HEADER;
+	    type(NULL,0,NULL,"Source route in Message-Id:");
+	  } else if (s[1] == '>') {
+	    SS->policyresult = FREEZE__MALFORMED_MESSAGE_ID_HEADER;
+	    type(NULL,0,NULL,"Empty Message-Id:");
+	  } else {
+	    const char *t = rfc821_path(s, 1);
+	    if (s == t) { /* error */
+#ifdef USE_STRICT_MSGID_FREEZING
+	      SS->policyresult = FREEZE__MALFORMED_MESSAGE_ID_HEADER;
+	      type(NULL,0,NULL,"Message-Id: syntax error");
+#endif
+	    } else {
+	      while (*t == ' ' || *t == '\t' || *t == '\r' || *t == '\n')
+		++t;
+	      if (*t) {
+		SS->policyresult = FREEZE__MALFORMED_MESSAGE_ID_HEADER;
+		type(NULL,0,NULL,"Spurious junk after Message-Id:");
+	      }
+	    }
+	  }
 	}
 #endif
 	if (CISTREQN(linebuf, "Precedence:", 11)) {
@@ -687,11 +946,23 @@ char *msg;
 		++s;
 	    if ((eol - s) < 4)
 		continue;	/* Hmm.. */
-	    if (CISTREQN("junk", s, 4))
+	    if (CISTREQN("high", s, 4))
+		mail_priority = _MAILPRIO_HIGH;
+	    else if (CISTREQN("junk", s, 4))
 		mail_priority = _MAILPRIO_JUNK;
 	    else if (CISTREQN("bulk", s, 4))
 		mail_priority = _MAILPRIO_BULK;
+	    else if (((eol - s) >= 6) && CISTREQN("normal", s, 6))
+		mail_priority = _MAILPRIO_NORMAL;
 	}
+#if 0 /* Nice in theory - impractical in reality */
+	if (msa_mode && CISTREQN(linebuf, "Sender:", 7)) {
+	    if ( SS->authuser != NULL ) {
+	      fprintf(SS->mfp, "Sender: %s@%s\n", SS->authuser, SS->myhostname);
+	      fprintf(SS->mfp, "Old-");
+	    }
+        }
+#endif
 #ifdef USE_TRANSLATION
 	if (X_translation && (X_settrrc == 0)) {
 	    if (CISTREQN(linebuf, "Content-Transfer-Encoding:", 26)) {
@@ -726,12 +997,15 @@ char *msg;
 			++s;
 		    if ((eol - s) < 10)
 			continue;	/* Hmm.. */
-#ifdef PARANOID_TRANSLATION
+
+		    /* Must ALWAYS check for C-T: TEXT/any ! */
+		    /* #ifdef PARANOID_TRANSLATION */
 		    if (CISTREQN("text", s, 4))
 			ct_is_text = 1;
 		    else
 			ct_is_text = 0;
-#endif				/* PARANOID_TRANSLATION */
+		    /* #endif */ /* PARANOID_TRANSLATION */
+
 		    continue;	/* do not write out this one */
 		}
 	    }
@@ -757,7 +1031,7 @@ char *msg;
 	/* Write the line out */
 	if (col > 0)
 	    for (wi = 0; wi < col; ++wi)
-		putc(TR_IN(linebuf[wi]), SS->mfp);
+		fputc(TR_IN(linebuf[wi]), SS->mfp);
 #else				/* USE_TRANSLATION */
 	if (has8bit)
 	    header_to_mime(linebuf, &col, sizeof(linebuf));
@@ -768,8 +1042,8 @@ char *msg;
 	has8bit = 0;
 	col = 0;
     }
-    if (logfp && verbose)
-	fprintf(logfp, "%d#\t(mail_priority=%d)\n", pid, mail_priority);
+    if (verbose)
+      type(NULL,0,NULL,"(mail_priority=%d)", mail_priority);
 #endif
 
     /* ================ Normal email BODY input.. ================ */
@@ -796,7 +1070,7 @@ char *msg;
 		continue;
 	    }
 	    if (!ferror(SS->mfp))
-		putc((state >> 8), SS->mfp);
+		fputc((state >> 8), SS->mfp);
 	    state = state & 0xFF;
 	}
 	if (!ferror(SS->mfp)) {
@@ -836,7 +1110,7 @@ char *msg;
 		    if ((c == ' ') || (c == '\t') ||
 			(c == '\n') || (c == '\r'))
 			continue;
-		    b64c[b64i++] = char64(c);
+		    b64c[b64i++] = decodechar64(c);
 		    if (c == '=') {
 			b64eod = 1;
 			continue;
@@ -859,10 +1133,10 @@ char *msg;
 		}
 	    }
 	    if (do_translate)
-		putc(TR_IN(c), SS->mfp);
+		fputc(TR_IN(c), SS->mfp);
 	    else
 #endif				/* USE_TRANSLATION */
-		putc(c, SS->mfp);
+		fputc(c, SS->mfp);
 	}
     }
     typeflush(SS);
@@ -903,19 +1177,19 @@ register long incount;
 		/* Suspended lone CR */
 		if (SS->mfp) {	/* We just discard it, if no output stream */
 		    if (!ferror(SS->mfp))
-			putc(SS->mvbstate, SS->mfp);
+			fputc(SS->mvbstate, SS->mfp);
 		    if (!ferror(SS->mfp))
-			putc(c, SS->mfp);
+			fputc(c, SS->mfp);
 		}
 	    } else {
 		/* CR + LF -- forget the CR */
 		if (SS->mfp && !ferror(SS->mfp))
-		    putc(c, SS->mfp);
+		    fputc(c, SS->mfp);
 	    }
 	} else {
 	    /* Anything else, just output it! */
 	    if (SS->mfp && !ferror(SS->mfp))
-		putc(c, SS->mfp);
+		fputc(c, SS->mfp);
 	}
 	SS->mvbstate = c;
     }
