@@ -4,7 +4,7 @@
  */
 /*
  *    Several extensive changes by Matti Aarnio <mea@nic.funet.fi>
- *      Copyright 1991-1999.
+ *      Copyright 1991-2000.
  */
 
 /*
@@ -15,7 +15,7 @@
 
 const char *VerbID = "ZMailer SMTP server %s";
 const char *Copyright = "Copyright 1990 Rayan S. Zachariassen";
-const char *Copyright2 = "Copyright 1991-1999 Matti Aarnio";
+const char *Copyright2 = "Copyright 1991-2000 Matti Aarnio";
 
 /* Timing parameters -- when expired, session is killed ... */
 
@@ -85,6 +85,12 @@ struct command command_list[] =
 #ifdef HAVE_OPENSSL
     {"STARTTLS", StartTLS}, /* RFC 2487 */
 #endif /* - HAVE_OPENSSL */
+
+    {"550", Silent},	/* Some Windows SMTP systems are mixing their
+			   threads - they send smtp server error messages
+			   to stream where they should be sending SMTP
+			   client verbs.. */
+
     {0, Null}
 };
 
@@ -102,8 +108,10 @@ int verbose = 0;
 int daemon_flg = 1;
 int netconnected_flg = 0;
 int pid, routerpid = -1;
+extern int contentpolicypid;
 int router_status = 0;
 FILE *logfp = NULL;
+int   logfp_to_syslog = 0;
 int D_alloc = 0;
 int smtp_syslog = 0;
 #ifdef USE_TRANSLATION
@@ -115,10 +123,13 @@ int strict_protocol = 0;
 int mustexit = 0;
 int configuration_ok = 0;
 int gotalarm;
+int unknown_cmd_limit = 10;
+int sum_sizeoption_value = 0;
+int always_flush_replies = 0;
 
 char logtag[16];
 
-jmp_buf jmpalarm;		/* Return-frame for breaking smtpserver
+sigjmp_buf jmpalarm;		/* Return-frame for breaking smtpserver
 				   when timeout hits.. */
 
 
@@ -175,7 +186,7 @@ int MaxSameIpSource = 100;	/* Max number of smtp connections in progress
 				   the remote SMTP server... */
 int MaxParallelConnections = 800; /* Total number of childs allowed */
 
-int MaxErrorRecipients = 10;	/* Max number of recipients for a message
+int MaxErrorRecipients = 3;	/* Max number of recipients for a message
 				   that has a "box" ( "<>" ) as its source
 				   address. */
 int percent_accept = -1;
@@ -210,6 +221,7 @@ int ehlo_ok = 1;
 int etrn_ok = 1;
 int starttls_ok = 0;
 int msa_mode = 0;
+etrn_cluster_ent etrn_cluster[MAX_ETRN_CLUSTER_IDX] = { {NULL,}, };
 char *tls_cert_file = NULL;
 char *tls_key_file  = NULL;
 char *tls_CAfile    = NULL;
@@ -227,7 +239,9 @@ int log_rcvd_tls_peer = 0;
 int auth_login_without_tls = 0;
 char *smtpauth_via_pipe = NULL;
 Usockaddr bindaddr;
+Usockaddr testaddr;
 int bindaddr_set    = 0;
+int testaddr_set    = 0;
 u_short bindport = 0;
 int bindport_set = 0;
 int use_tcpwrapper = 0;
@@ -259,6 +273,7 @@ extern void type220headers __((SmtpState *SS, const int identflg, const char *xl
 
 
 extern void openlogfp __((SmtpState * SS, int insecure));
+extern char taspid_encodechars[];
 
 void openlogfp(SS, insecure)
 SmtpState *SS;
@@ -266,13 +281,18 @@ int insecure;
 {
     /* opening the logfile should be done before we reset the uid */
     time_t now = time(NULL);
-
+    struct tm *tt;
+    tt = gmtime(&now);
     pid = getpid();
-    sprintf(logtag, "%05d%c%c%c%c", pid,
-	    'A'+(((int)now / (26*26*26)) % 26),
-	    'A'+(((int)now / (26*26   )) % 26),
-	    'A'+(((int)now / (26      )) % 26),
-	    'A'+(((int)now             ) % 26));
+
+    /* %M%D%h%m%s_pid_ */
+
+    sprintf(logtag, "%c%c%c%c%05d",
+	    taspid_encodechars[ tt->tm_mday-1 ],
+	    taspid_encodechars[ tt->tm_hour   ],
+	    taspid_encodechars[ tt->tm_min    ],
+	    taspid_encodechars[ tt->tm_sec    ],
+	    pid);
 
     if (logfp != NULL)
 	fclose(logfp);
@@ -304,6 +324,7 @@ int insecure;
 			"%s: cannot open logfile \"%s\": %s\n",
 			progname, fname, strerror(errno));
 	} else {
+	    fcntl(fd, F_SETFD, fcntl(fd, F_GETFD, 0) | FD_CLOEXEC);
 	    logfp = fdopen(fd, "a");
 	    setvbuf(logfp, NULL, _IOFBF, BUFSIZ);
 	    /* Line-buffered */
@@ -381,15 +402,15 @@ char **argv;
 #ifndef __STDC__
 #if defined(AF_INET6) && defined(INET6)
 #ifdef USE_TRANSLATION
-		       "?46aBC:d:ighl:np:I:L:M:P:R:s:S:VvwX8"
+		       "?46aBC:d:ighl:np:I:L:M:P:R:s:S:T:VvwX8"
 #else /* xlate */
-		       "?46aBC:d:ighl:np:I:L:M:P:R:s:S:Vvw"
+		       "?46aBC:d:ighl:np:I:L:M:P:R:s:S:T:Vvw"
 #endif /* xlate */
 #else /* INET6 */
 #ifdef USE_TRANSLATION
-		       "?4aBC:d:ighl:np:I:L:M:P:R:s:S:VvwX8"
+		       "?4aBC:d:ighl:np:I:L:M:P:R:s:S:T:VvwX8"
 #else
-		       "?4aBC:d:ighl:np:I:L:M:P:R:s:S:Vvw"
+		       "?4aBC:d:ighl:np:I:L:M:P:R:s:S:T:Vvw"
 #endif /* xlate */
 #endif /* INET6 */
 #else /* __STDC__ */
@@ -400,7 +421,7 @@ char **argv;
 #endif
 		       "aBC:d:ighl:n"
 		       "p:"
-		       "I:L:M:P:R:s:S:Vvw"
+		       "I:L:M:P:R:s:S:T:Vvw"
 #ifdef USE_TRANSLATION
 		       "X8"
 #endif /* USE_TRANSLATION */
@@ -433,25 +454,57 @@ char **argv;
 	case 'd':
 	    debug = atoi(optarg);
 	    break;
-	case 'M':
-	    maxsize = atol(optarg);
-	    if (maxsize < 0)
-		maxsize = 0;
-	    break;
-	case 'i':		/* interactive */
-	    daemon_flg = 0;
-	    break;
-	case 'v':
-	    verbose = 1;	/* in conjunction with -i */
-	    break;
 	case 'g':		/* gullible */
 	    skeptical = 0;
 	    break;
 	case 'h':		/* checkhelo */
 	    checkhelo = 1;
 	    break;
+	case 'i':		/* interactive */
+	    daemon_flg = 0;
+	    break;
+	case 'I':		/* PID file */
+	    pidfile = optarg;
+	    pidfile_set = 1;
+	    break;
 	case 'l':		/* log file(prefix) */
+
+	    if (strcmp(optarg,"SYSLOG")==0) {
+	      logfp_to_syslog = 1;
+	      break;
+	    }
+
 	    logfile = optarg;
+
+	    break;
+	case 'L':		/* Max LoadAverage */
+	    maxloadavg = atoi(optarg);
+	    if (maxloadavg < 1)
+		maxloadavg = 10;	/* Humph.. */
+	    break;
+	case 'M':
+	    maxsize = atol(optarg);
+	    if (maxsize < 0)
+		maxsize = 0;
+	    break;
+	case 'n':		/* running under inetd */
+	    inetd = 1;
+	    break;
+	case 'p':
+	    bindport = htons(atoi(optarg));
+	    bindport_set = 1;
+	    break;
+	case 'P':
+	    postoffice = strdup(optarg);
+	    break;
+	case 'R':		/* router binary used for verification */
+	    routerprog = strdup(optarg);
+	    break;
+	case 's':		/* checking style */
+	    if (strcmp(optarg,"strict")==0)
+	      strict_protocol = 1;
+	    else
+	      style = strdup(optarg);
 	    break;
 	case 'S':		/* Log-suffix style */
 	    logstyle = 0;
@@ -460,33 +513,48 @@ char **argv;
 	    else if (cistrcmp(optarg, "local") == 0)
 		logstyle = 1;
 	    break;
-	case 'n':		/* running under inetd */
-	    inetd = 1;
+	case 'T':
+	  /* Enter in interactive mode claimed foreign source IPv4/IPv6
+	     address, and then proceed to handle policy analysis as in
+	     normal operational case. */
+
+	    memset(&testaddr, 0, sizeof(testaddr));
+	    testaddr_set = 1;
+#if defined(AF_INET6) && defined(INET6)
+	    if (cistrncmp(optarg,"[ipv6 ",6) == 0 ||
+		cistrncmp(optarg,"[ipv6:",6) == 0 ||
+		cistrncmp(optarg,"[ipv6.",6) == 0) {
+	      char *s = strchr(optarg,']');
+	      if (s) *s = 0;
+	      if (inet_pton(AF_INET6, optarg+6, &testaddr.v6.sin6_addr) < 1) {
+		/* False IPv6 number literal */
+		/* ... then we don't set the IP address... */
+		testaddr_set = 0;
+		fprintf(stderr,"smtpserver: -T option argument is not valid IPv6 address: [ipv6.hhhh:hhhh:hhhh:hhhh:hhhh:hhhh:1.2.3.4]\n");
+		++errflg;
+	      }
+	      testaddr.v6.sin6_family = AF_INET6;
+	    } else
+#endif
+	      if (*optarg == '[') {
+		char *s = strchr(optarg,']');
+		if (s) *s = 0;
+		if (inet_pton(AF_INET, optarg+1, &testaddr.v4.sin_addr) < 1) {
+		  /* False IP(v4) number literal */
+		  /* ... then we don't set the IP address... */
+		  testaddr_set = 0;
+		  fprintf(stderr,"smtpserver: -T option argument is not valid IPv4 address: [1.2.3.4]\n");
+		  ++errflg;
+		}
+		testaddr.v4.sin_family = AF_INET;
+	      } else {
+		testaddr_set = 0;
+		fprintf(stderr,"smtpserver: -T option argument must be wrapped inside brackets: [1.2.3.4]\n");
+		++errflg;
+	      }
 	    break;
-	case 's':		/* checking style */
-	    if (strcmp(optarg,"strict")==0)
-	      strict_protocol = 1;
-	    else
-	      style = strdup(optarg);
-	    break;
-	case 'I':		/* PID file */
-	    pidfile = optarg;
-	    pidfile_set = 1;
-	    break;
-	case 'L':		/* Max LoadAverage */
-	    maxloadavg = atoi(optarg);
-	    if (maxloadavg < 1)
-		maxloadavg = 10;	/* Humph.. */
-	    break;
-	case 'p':
-	    bindport = htons(atoi(optarg));
-	    bindport_set = 1;
-	    break;
-	case 'R':		/* router binary used for verification */
-	    routerprog = strdup(optarg);
-	    break;
-	case 'P':
-	    postoffice = strdup(optarg);
+	case 'v':
+	    verbose = 1;	/* in conjunction with -i */
 	    break;
 	case 'V':
 	    prversion("smtpserver");
@@ -528,8 +596,8 @@ char **argv;
 		"Usage: %s [-46aBignVvw]\
  [-C cfgfile] [-s xx] [-L maxLoadAvg]\
  [-M SMTPmaxsize] [-R rtrprog] [-p port#]\
- [-P postoffice] [-l logfile] [-S 'local'|'remote']\
- [-I pidfile]\n"
+ [-P postoffice] [-l SYSLOG] [-l logfile] [-S 'local'|'remote']\
+ [-I pidfile] [-T test-net-addr]\n"
 #else /* __STDC__ */
 		"Usage: %s [-4"
 #if defined(AF_INET6) && defined(INET6)
@@ -542,7 +610,7 @@ char **argv;
 		"] [-C cfgfile] [-s xx] [-L maxLoadAvg]"
 		" [-M SMTPmaxsize] [-R rtrprog] [-p port#]"
 		" [-P postoffice] [-l logfile] [-S 'local'|'remote']"
-		" [-I pidfile]\n"
+		" [-I pidfile] [-T test-net-addr]\n"
 #endif /* __STDC__ */
 		, progname);
 	exit(1);
@@ -578,19 +646,33 @@ char **argv;
 
       raddrlen = sizeof(SS.raddr);
       memset(&SS.raddr, 0, raddrlen);
-      if (getpeername(SS.inputfd, (struct sockaddr *) &SS.raddr, &raddrlen))
-	netconnected_flg = 0;
-      else {
+      if (getpeername(SS.inputfd, (struct sockaddr *) &SS.raddr, &raddrlen)) {
+	if (testaddr_set) {
+	  netconnected_flg = 1;
+	  memcpy(&SS.raddr, &testaddr, sizeof(testaddr));
+	}
+      } else {
 	/* Got a peer name (it is a socket) */
 	netconnected_flg = 1;
 	if (SS.raddr.v4.sin_family != AF_INET
 #ifdef AF_INET6
 	    && SS.raddr.v4.sin_family != AF_INET6
 #endif
-	    )
+	    ) {
 	  /* well, but somebody uses socketpair(2)  which is
 	     an AF_UNIX thing and sort of full-duplex pipe(2)... */
 	  netconnected_flg = 0;
+	}
+	if (netconnected_flg) {
+	  /* Lets figure-out who we are this time around -- we may be on
+	     a machine with multiple identities per multiple interfaces,
+	     or via virtual IP-numbers, or ... */
+	  localsocksize = sizeof(SS.localsock);
+	  if (getsockname(FILENO(stdin), (struct sockaddr *) &SS.localsock,
+			  &localsocksize) != 0) {
+	    /* XX: ERROR! */
+	  }
+	}
       }
 
       strcpy(SS.rhostname, "stdin");
@@ -713,14 +795,15 @@ char **argv;
 	    close(s);
 	  }
 	}
-	if (use_ipv6) {
+	if (force_ipv4) {
+	  s = socket(PF_INET, SOCK_STREAM, 0 /* IPPROTO_IP   */ );
+	  use_ipv6 = 0;
+	} else if (use_ipv6) {
 	  s = socket(PF_INET6, SOCK_STREAM, 0 /* IPPROTO_IPV6 */ );
-#if 0
 	  if (s < 0) {	/* Fallback to the IPv4 mode .. */
 	    s = socket(PF_INET, SOCK_STREAM, 0 /* IPPROTO_IP   */ );
 	    use_ipv6 = 0;
 	  }
-#endif
 	} else
 	  s = socket(PF_INET, SOCK_STREAM, 0 /* IPPROTO_IP   */ );
 #else
@@ -862,6 +945,7 @@ char **argv;
 	  char *cp;
 	  time(&now);
 	  cp = rfc822date(&now);
+	  zsyslog((LOG_INFO, "server started."));
 	  fprintf(logfp, "00000#\tstarted server pid %d at %s", pid, cp);
 	  /*fprintf(logfp,"00000#\tfileno(logfp) = %d\n",fileno(logfp)); */
 	  fclose(logfp);
@@ -911,6 +995,7 @@ char **argv;
 	    fprintf(stderr, "%s: accept(): %s; %s",
 		    progname, strerror(err), rfc822date(&now));
 	    openlogfp(&SS, daemon_flg);
+	    zsyslog((LOG_INFO, "accept() error=%d (%s)", err, strerror(err)));
 	    if (logfp) {
 	      fprintf(logfp, "000000#\taccept(): %s; %s",
 		      strerror(err), (char *) rfc822date(&now));
@@ -1045,7 +1130,7 @@ char **argv;
 	      strcpy(msg, "450 Come again latter\r\n");
 	      len = strlen(msg);
 	      write(msgfd, msg, len);
-	      close(0); close(1);
+	      close(0); close(1); close(2);
 #if 1
 	      sleep(2);	/* Not so fast!  We need to do this to
 			   avoid (as much as possible) the child
@@ -1067,7 +1152,7 @@ char **argv;
 	      strcpy(msg, "450 Come again latter\r\n");
 	      len = strlen(msg);
 	      write(msgfd, msg, len);
-	      close(0); close(1);
+	      close(0); close(1); close(2);
 #if 1
 	      sleep(2);	/* Not so fast!  We need to do this to
 			   avoid (as much as possible) the child
@@ -1080,10 +1165,12 @@ char **argv;
 	    smtpserver(&SS, 1);
 	    /* Expediated filehandle closes before
 	       the mandatory sleep(2) below. */
-	    close(0); close(1);
+	    close(0); close(1); close(2);
 
 	    if (routerpid > 0)
 	      killr(&SS, routerpid);
+	    if (contentpolicypid > 0)
+	      killr(&SS, contentpolicypid);
 
 	    if (netconnected_flg)
 	      sleep(2);
@@ -1093,6 +1180,7 @@ char **argv;
 	/* Stand-alone server, kill the pidfile at the exit! */
 	killpidfile(pidfile);
 	openlogfp(&SS, daemon_flg);
+	zsyslog((LOG_INFO, "killed server."));
 	if (logfp != NULL) {
 	  char *cp;
 	  time(&now);
@@ -1104,6 +1192,8 @@ char **argv;
       }
     if (routerpid > 0)
 	killr(&SS, routerpid);
+    if (contentpolicypid > 0)
+      killr(&SS, contentpolicypid);
     if (netconnected_flg)
       sleep(2);
     exit(0);
@@ -1218,6 +1308,7 @@ int sig;
   mustexit = 1;
 }
 
+
 static RETSIGTYPE
  reaper(sig)
 int sig;
@@ -1241,6 +1332,10 @@ int sig;
 	  router_status = status;
 	  routerpid = -1;
 	}
+	if (lpid == contentpolicypid && contentpolicypid > 0) {
+	  contentpolicypid = -1;
+	}
+
 	childreap(lpid);
     }
     SIGNAL_HANDLE(SIGCHLD, reaper);
@@ -1252,10 +1347,10 @@ long tell;
 const char *msg;
 {
     zsyslog((LOG_ERR,
-	     "aborted (%ld bytes) from %s/%d: %s",
-	     tell, SS->rhostname, SS->rport, msg));
+	     "%s - aborted (%ld bytes) from %s/%d: %s",
+	     logtag, tell, SS->rhostname, SS->rport, msg));
     if (logfp != NULL) {
-	fprintf(logfp, "%s-\taborted (%ld bytes): %s\n", logtag, tell, msg);
+	fprintf(logfp, "%s - aborted (%ld bytes): %s\n", logtag, tell, msg);
 	fflush(logfp);
     }
 }
@@ -1369,12 +1464,7 @@ int buflen, *rcp;
 	int i = -1, rc = -1;
 
 	if (!pipeliningok || !s_hasinput(SS))
-	    typeflush(SS);
-	else
-	    /* if (verbose) */
-	if (logfp)
-	  type(NULL,0,NULL,
-	       "-- pipeline input exists %d bytes", s_hasinput(SS));
+	  typeflush(SS);
 
 	/* Alarm processing on the SMTP protocol channel */
 	alarm(SMTP_COMMAND_ALARM_IVAL);
@@ -1479,7 +1569,7 @@ int insecure;
     if (!netconnected_flg)
       strict_protocol = 0;
 
-    rc = setjmp(jmpalarm);
+    rc = sigsetjmp(jmpalarm,1);
     if (rc != 0) {
 	/* Oooo...  We are returning here via  longjmp(),
 	   which means we just got a timeout (SIGALRM),
@@ -1503,6 +1593,8 @@ int insecure;
 
 	if (routerpid > 0)
 	    killr(SS, routerpid);
+	if (contentpolicypid > 0)
+	  killr(SS, contentpolicypid);
 	exit(0);
     }
     report(SS, "(connected)");
@@ -1560,6 +1652,7 @@ int insecure;
 	}
     } else {
 	strcpy(SS->whoson_data,"NOT-CHECKED");
+	SS->whoson_result = -1;
     }
     policystatus     = policyinit(&policydb, &SS->policystate,
 				  SS->whoson_result);
@@ -1584,6 +1677,7 @@ int insecure;
     /* re-opening the log ?? */
     zopenlog("smtpserver", LOG_PID, LOG_MAIL);
 
+#if 0 /* NO MORE TCP-WRAPPER AT SMTPSERVER -- USE POLICY CODE! */
 #ifdef HAVE_TCPD_H		/* TCP-Wrapper code */
     if (use_tcpwrapper && netconnected_flg &&
 	wantconn(SS->inputfd, "smtp-receiver") == 0) {
@@ -1595,6 +1689,7 @@ int insecure;
 	sleep(2);
 	exit(0);
     }
+#endif
 #endif
 
     if (SS->reject_net) {
@@ -1640,7 +1735,7 @@ int insecure;
 	SS->state = MailOrHello;
 
     cfinfo = NULL;
-    if (logfp != NULL) {
+    {
 	char *s = policymsg(policydb, &SS->policystate);
 	if (insecure)
 	  type(NULL,0,NULL,"remote from %s:%d", SS->ihostaddr, SS->rport);
@@ -1649,21 +1744,32 @@ int insecure;
 	if (SS->policyresult != 0 || s != NULL)
 	  type(NULL,0,NULL,"-- policyresult=%d initial policy msg: %s",
 	       SS->policyresult, (s ? s : "<NONE!>"));
-	fflush(logfp);
+	if (logfp)
+	  fflush(logfp);
     }
     while (1) {
 
 	char buf[SMTPLINESIZE];	/* limits size of SMTP commands...
 				   On the other hand, limit is asked
 				   to be only 1000 chars, not 8k.. */
-	char *eobuf, c, co;
+	char *eobuf;
+	char c, co;
 	int i;
 
+	if (always_flush_replies)
+	  typeflush(SS);
+
 	i = s_gets(SS, buf, sizeof(buf), &rc, &co, &c );
 
 	if (i <= 0)	/* EOF ??? */
 	  break;
 
+	if (s_hasinput(SS))
+	  if (logfp || logfp_to_syslog)
+	    type(NULL,0,NULL,
+		 "-- pipeline input exists %d bytes", s_hasinput(SS));
+
+
 	eobuf = &buf[i-1];	/* Buf end ptr.. */
 
 	/* Chop the trailing spaces */
@@ -1680,7 +1786,10 @@ int insecure;
 	}
 				   
 
-	if (logfp != NULL) {
+	if (logfp_to_syslog)
+	  zsyslog((LOG_DEBUG, "%s r %s", logtag, buf));
+
+	if (logfp) {
 	    fprintf(logfp, "%sr\t%s\n", logtag, buf);
 	    fflush(logfp);
 	}
@@ -1720,14 +1829,12 @@ int insecure;
 	    fprintf(stdout, "%s\n", buf);	/* XX: trace.. */
 	report(SS, "%.100s", buf);
 
-	for (cp = buf; isascii(*cp) && isalpha(*cp); ++cp)
+	for (cp = buf; (c = *cp) && isascii(c & 0xFF) && isalnum(c & 0xFF); ++cp)
 	    continue;
-	if (cp > buf + 8) {	/* "DEBUG" is longest of them.. */
-	    smtp_tarpit(SS);
-	    type(SS, 550, m552, "Syntax error");
-	    typeflush(SS);
-	    continue;
-	}
+
+	if (cp > buf + 8)	/* "DEBUG" is longest of them.. */
+	  goto unknown_command;
+
 	c = *cp;
 	if (c != '\0')
 	    *cp = '\0';
@@ -1741,6 +1848,14 @@ int insecure;
 
 	unknown_command:
 
+	    ++SS->unknown_cmd_count;
+
+	    if (SS->unknown_cmd_count >= unknown_cmd_limit) {
+	      type(SS, 550, m552, "One too many unknown command '%s'", buf);
+	      typeflush(SS);
+	      break;
+	    }
+
 	    smtp_tarpit(SS);
 
 	    type(SS, 550, m552, "Unknown command '%s'", buf);
@@ -1818,12 +1933,18 @@ int insecure;
 	}
 
 	switch (SS->carp->cmd) {
+	case Silent:
+	    /* We are SILENT - no response at all! */
+	    break;
 	case Null:
 	    type(SS, 550, m550, "panic!");
 	    typeflush(SS);
 	    break;
 #ifdef HAVE_OPENSSL
 	case StartTLS:
+#if 0 /* A debug thing.. */
+	    always_flush_replies = 1;
+#endif
 	    smtp_starttls(SS, buf, cp);
 	    break;
 #endif /* - HAVE_OPENSSL */
@@ -1851,6 +1972,7 @@ int insecure;
 	    smtp_rcpt(SS, buf, cp);
 	    break;
 	case Data:
+
 	    if (smtp_data(SS, buf, cp) < 0) {
 #ifdef HAVE_OPENSSL
 	      Z_cleanup(SS);
@@ -1859,6 +1981,7 @@ int insecure;
 	    }
 	    break;
 	case BData:
+
 	    if (smtp_bdata(SS, buf, cp) < 0) {
 #ifdef HAVE_OPENSSL
 	      Z_cleanup(SS);
@@ -2178,6 +2301,10 @@ const char *status, *fmt, *s1, *s2, *s3, *s4, *s5, *s6;
     if (buflen+4 > sizeof(buf)) {
       /* XXX: Buffer overflow ??!! Signal about it, and crash! */
     }
+
+    if (logfp_to_syslog)
+      zsyslog((LOG_DEBUG,"%s %c %s", logtag, (SS ? 'w' : '#'), buf));
+
     if (logfp != NULL) {
 	fprintf(logfp, "%s%c\t%s\n", logtag, (SS ? 'w' : '#'), buf);
 	fflush(logfp);
@@ -2201,6 +2328,8 @@ type220headers(SS, identflg, xlatelang, curtime)
      const char *curtime;
 {
     char *s, **hh = hdr220lines;
+    char linebuf[8000];
+    char *l, *le;
 
     /* Below use of  fprintf()  for SS->outfp  channel is for
        ensuring that  setvbuf( _IOFBF ) is honoured always.
@@ -2210,9 +2339,8 @@ type220headers(SS, identflg, xlatelang, curtime)
     for (; *hh ; ++hh) {
       char c = (hh[1] == NULL) ? ' ' : '-';
       
-      fprintf(SS->outfp, "%03d%c", 220, c);
-      if (logfp != NULL)
-	fprintf(logfp, "%sw\t%03d%c", logtag, 220, c);
+      le = linebuf + sizeof(linebuf) -1;
+      l  = linebuf;
 
       /* The format meta-tags:
        *
@@ -2225,49 +2353,64 @@ type220headers(SS, identflg, xlatelang, curtime)
        */
 
       s = *hh;
-      while (*s) {
+      while (*s && l < le) {
 	if (*s == '%') {
+	  int freespc = le-l;
+	  int len;
+
 	  ++s;
 	  switch (*s) {
 	  case '%':
-	    fprintf(SS->outfp,"%%");
-	    if (logfp) putc('%',logfp);
+	    *l++ = '%';
 	    break;
 	  case 'H':
-	    fprintf(SS->outfp,"%s",SS->myhostname);
-	    if (logfp) fputs(SS->myhostname,logfp);
+	    len = strlen(SS->myhostname);
+	    memcpy(l, SS->myhostname, freespc < len ? freespc : len);
+	    l += len;
 	    break;
 	  case 'I':
 	    if (identflg) {
-	      fprintf(SS->outfp,"+IDENT");
-	      if (logfp) fputs("+IDENT",logfp);
+	      len = 6;
+	      memcpy(l, "+IDENT", freespc < len ? freespc : len);
+	      l += len;
 	    }
 	    break;
 	  case 'V':
-	    fprintf(SS->outfp,"%s",VersionNumb);
-	    if (logfp) fputs(VersionNumb,logfp);
+	    len = strlen(VersionNumb);
+	    memcpy(l, VersionNumb, freespc < len ? freespc : len);
+	    l += len;
 	    break;
 	  case 'T':
-	    fprintf(SS->outfp,"%s",curtime);
-	    if (logfp) fputs(curtime,logfp);
+	    len = strlen(curtime);
+	    memcpy(l, curtime, freespc < len ? freespc : len);
+	    l += len;
 	    break;
 	  case 'X':
 	    if (!xlatelang) xlatelang = "";
-	    fprintf(SS->outfp,"%s",xlatelang);
-	    if (logfp) fputs(xlatelang,logfp);
+	    len = strlen(xlatelang);
+	    memcpy(l, xlatelang, freespc < len ? freespc : len);
+	    l += len;
 	    break;
 	  default:
 	    /* Duh ?? */
 	    break;
 	  }
 	} else {
-	  fprintf(SS->outfp,"%c",*s);
-	  if (logfp) putc(*s, logfp);
+	  *l++ = *s;
 	}
 	if (*s) ++s;
       }
-      fprintf(SS->outfp,"\r\n");
-      if (logfp) fputc('\n',logfp);
+      if (l < le)
+	*l = 0;
+      *le = 0;
+
+      fprintf(SS->outfp, "220%c%s\r\n", c, linebuf);
+
+      if (logfp_to_syslog)
+	zsyslog((LOG_DEBUG, "%s w 220%c%s", logtag, c, linebuf));
+      if (logfp)
+	fprintf(logfp, "%sw\t220%c%s\n", logtag, c, linebuf);
+
     }
     fflush(SS->outfp);
     if (logfp) fflush(logfp);
@@ -2341,7 +2484,9 @@ va_dcl
 
       buflen = bp - buf;
 
-      if (logfp != NULL)
+      if (logfp_to_syslog)
+	zsyslog((LOG_DEBUG, "%s w %s", logtag, buf));
+      if (logfp)
 	fprintf(logfp, "%sw\t%s\n", logtag, buf);
 
       strcpy(bp, "\r\n");
