@@ -1,7 +1,7 @@
 /*
  *  Globals of the ZMailer  smtp-server
  *
- *  Matti Aarnio <mea@nic.funet.fi> 1997-1999
+ *  Matti Aarnio <mea@nic.funet.fi> 1997-2000
  */
 
 
@@ -45,6 +45,7 @@
 #include <sys/time.h>
 #include <sys/stat.h>
 #include <sys/file.h>
+#include <sysexits.h>
 #ifdef HAVE_FCNTL_H
 #include <fcntl.h>
 #endif
@@ -80,6 +81,9 @@ extern int wait();
 #ifdef HAVE_NETDB_H
 #include <netdb.h>
 #endif
+#ifndef EAI_AGAIN
+# include "netdb6.h"
+#endif
 
 #include <sys/socket.h>
 #include <netinet/in.h>
@@ -94,6 +98,7 @@ extern int wait();
 #endif
 #endif
 #endif
+#include <arpa/inet.h>
 
 #include "libc.h"
 #include "libz.h"
@@ -143,7 +148,8 @@ typedef enum {
 #ifdef HAVE_OPENSSL
     StartTLS,
 #endif /* - HAVE_OPENSSL */
-    Hello2, Mail2, Send2, Verify2	/* 8-bit extensions */
+    Hello2, Mail2, Send2, Verify2,	/* 8-bit extensions */
+    Silent		/* One particular client error trap.. */
 } Command;
 
 
@@ -210,6 +216,8 @@ typedef struct {
     char whoson_data[128];
 #endif
 
+    int unknown_cmd_count;
+
 } SmtpState;
 
 #define STYLE(i,c)	(strchr(((i)==NULL ? style : (i)->flags), (c)) != NULL)
@@ -252,6 +260,11 @@ extern int ehlo_ok;
 extern int etrn_ok;
 extern int starttls_ok;
 extern int msa_mode;
+#define MAX_ETRN_CLUSTER_IDX 40
+typedef struct {
+  char *nodename; char *username; char *password;
+} etrn_cluster_ent;
+extern etrn_cluster_ent etrn_cluster[];
 extern char *tls_cert_file, *tls_key_file, *tls_CAfile, *tls_CApath;
 extern int tls_loglevel, tls_enforce_tls, tls_ccert_vd, tls_use_scache;
 extern int tls_ask_cert, tls_req_cert, tls_scache_timeout;
@@ -259,15 +272,19 @@ extern int log_rcvd_whoson, log_rcvd_ident, log_rcvd_authuser;
 extern int log_rcvd_tls_mode, log_rcvd_tls_peer;
 extern int auth_login_without_tls;
 extern char *smtpauth_via_pipe;
+extern char *contentfilter;
 extern int strict_protocol;
 extern int rcptlimitcnt;
 extern int enable_router;
 extern int use_tcpwrapper;
 extern int configuration_ok;
+extern int unknown_cmd_limit;
+extern int sum_sizeoption_value;
 
-extern int bindaddr_set, bindport_set;
+extern int bindaddr_set, bindport_set, testaddr_set;
 extern u_short   bindport;
 extern Usockaddr bindaddr;
+extern Usockaddr testaddr;
 
 extern const char *progname;
 extern int debug, skeptical, checkhelo, ident_flag, verbose;
@@ -299,6 +316,7 @@ extern void reporterr __((SmtpState *, long, const char *));
 extern const char *Copyright;
 extern const char *Copyright2;
 extern FILE *logfp;
+extern int   logfp_to_syslog;
 extern int pid;
 
 extern char *routerprog;
@@ -419,9 +437,12 @@ extern int  Z_pending __(( SmtpState * ));
 extern int  Z_write   __(( SmtpState *, const void *, int ));
 extern int  Z_read    __(( SmtpState *, void *, int ));
 
+#if 0 /* NOT ANYMORE! */
 #ifdef HAVE_TCPD_H		/* The hall-mark of having tcp-wrapper things around */
 extern int wantconn __((int sock, char *prgname));
 #endif
+#endif
+
 extern char *rfc822date __((time_t *));
 
 #ifdef HAVE_STDARG_H		/* Fwd declaration */
