@@ -181,15 +181,18 @@ ctladdr(cp)
 	if (ap == NULL)
 		return NULL;
 	ap->link = NULL;
+
 	/* While space: */
-	while (*cp != 0 && (*cp == ' ' || *cp == '\t')) ++cp;
+	while (*cp == ' ' || *cp == '\t') ++cp;
+
 	/* CHANNEL: */
 	ap->channel = cp;
-	/* While not space: */
 	cp = skip821address(cp);
-	*cp++ = '\0';
+	if (*cp) *cp++ = '\0';
+
 	/* While space: */
-	while (*cp != 0 && (*cp == ' ' || *cp == '\t')) ++cp;
+	while (*cp == ' ' || *cp == '\t') ++cp;
+
 	/* HOST: */
 	ap->host = cp;
 	ap->routermxes = NULL;
@@ -198,14 +201,18 @@ ctladdr(cp)
 	else {
 	  /* While not space: */
 	  cp = skip821address(cp);
-	  *cp++ = '\0';
+	  if (*cp) *cp++ = '\0';
 	}
+
 	/* While space: */
-	while (*cp != 0 && (*cp == ' ' || *cp == '\t')) ++cp;
+	while (*cp == ' ' || *cp == '\t') ++cp;
+
+	/* USER: */
 	ap->user = cp;
-	/* the user value is allowed to have embedded whitespace */
 	cp = skip821address(cp);
-	*cp++ = '\0';
+	if (*cp) *cp++ = '\0';
+
+	/* PRIVILEGE: */
 	ap->misc = cp;
 	return ap;
 }
@@ -620,6 +627,11 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 	  }
 	}
 
+	/* Sometimes we bail out before terminating NULLs are added..
+	   propably before anything is added. */
+	msgheaders   [headers_cnt] = NULL;
+	msgheaderscvt[headers_cnt] = NULL;
+
 	d.msgheaders    = msgheaders;		/* Original headers	*/
 	d.msgheaderscvt = msgheaderscvt;	/* Modified set		*/
 
@@ -643,7 +655,7 @@ ctlopen(file, channel, host, exitflagp, selectaddr, saparam, matchrouter, mrpara
 		       "message file is missing(!) -- possibly due to delivery scheduler restart.  Consider resending your message");
 	  }
 	  errno = e;
-	  warning("Cannot open message file \"%s\"! (%m)", mfpath);
+	  warning("Cannot open message file \"%s\"! (errno=%d)", mfpath, errno);
 #ifndef USE_ALLOCA
 	  free(mfpath);
 #endif
