@@ -5,13 +5,17 @@
  *	Universitaetsplatz 3-4
  *	D-03044 Cottbus, Germany
  *
- * Adaptation to ZMailer is by Matti Aarnio <mea@nic.funet.fi> (c) 1999
+ * Adaptation to ZMailer is by Matti Aarnio <mea@nic.funet.fi> (c) 1999-2000
  */
 
 #include "smtp.h"
 
+extern int timeout_tcpw;
+
 #ifdef HAVE_OPENSSL
 
+static FILE *vlog = NULL;
+
 static const char MAIL_TLS_CLNT_CACHE[] = "TLSclntcache";
 static const int id_maxlength = 32;	/* Max ID length in bytes */
 
@@ -25,24 +29,26 @@ int	tls_scache_timeout = 3600;	/* One hour */
 int	tls_use_scache     = 0;
 
 #define CCERT_BUFSIZ 256
-static char peer_CN[CCERT_BUFSIZ];
-static char issuer_CN[CCERT_BUFSIZ];
+char tls_peer_cert_name        [CCERT_BUFSIZ];
+char tls_peer_cert_issuer_name [CCERT_BUFSIZ];
 
 static unsigned char peername_md5[MD5_DIGEST_LENGTH];
 
 extern int demand_TLS_mode;
+extern int tls_available;
 
-int	tls_peer_verified = 0;
+int	tls_peer_verified  = 0;
+int	tls_use_read_ahead = 1;
 
-char   *tls_CAfile = NULL;
-char   *tls_CApath = NULL;
+char   *tls_CAfile    = NULL;
+char   *tls_CApath    = NULL;
 char   *tls_cert_file = NULL;
 char   *tls_key_file  = NULL;
 
-char   *tls_protocol = NULL;
-const char   *tls_cipher_name = NULL;
-int	tls_cipher_usebits = 0;
-int	tls_cipher_algbits = 0;
+const char *tls_protocol    = NULL;
+const char *tls_cipher_name = NULL;
+int	tls_cipher_usebits  = 0;
+int	tls_cipher_algbits  = 0;
 
 int	tls_loglevel = 0;
 
@@ -63,6 +69,7 @@ msg_info(va_alist)
 #endif
 {
 	va_list	ap;
+	FILE *fp;
 #ifdef HAVE_STDARG_H
 	va_start(ap,fmt);
 #else
@@ -72,15 +79,27 @@ msg_info(va_alist)
 	SS  = va_arg(ap, SmtpState *);
 	fmt = va_arg(ap, char *);
 #endif
-	if (!logfp) return;
 
+	if (vlog) {
+	  fp = vlog;
+	  fprintf(fp, "# ");
+	} else if (logfp) {
+	  fp = logfp;
+	  fprintf(fp, "%s#\t", logtag());
+	} else {
+	  fp = stderr; /* No LOGFP, to STDERR with DBGdiag prefix.. */
+	  fprintf(fp, "# ");
+	}
 
 #ifdef	HAVE_VPRINTF
-	vfprintf(logfp, fmt, ap);
+	vfprintf(fp, fmt, ap);
 #else	/* !HAVE_VPRINTF */
  ERROR:ERROR:ERROR:No 
 #endif	/* HAVE_VPRINTF */
 
+	fprintf(fp,"\n");
+	fflush(fp);
+
 	va_end(ap);
 }
 
@@ -97,89 +116,6 @@ void mail_queue_path(buf, subdir, filename)
 }
 
 
- /*
-  * Wrapper routines around read/write and read_wait/write_wait.
-  * If TLS is active, we must use the appropriate SSL function
-  * instead of the direct one.
-  *
-  * We do live quite comfortable here, as smtp_stream can only handle
-  * one connection at a time, so we also only take to care of one
-  * filedescriptor, which is saved as "tls_fd".
-  *
-  * You may note, that tls_read_wait() and tls_write_wait()
-  * seem to be of no real use here. The reason is, that there is no
-  * equivalent to "select()" available for the SSL connection. But I
-  * already have prepared the wrapper functions, so once this equivalent
-  * is available, we can immediately use it.
-  *
-  * We explicitly leave out the select() calls in TLS mode, as there
-  * is buffering included in the SSL_* routines and we don't want to
-  * have it hanging! Consider that there are still bytes in the SSL buffer,
-  * but no new bytes arrive at the interface, then select() on the read
-  * channel would wait erronously.
-  *
-  * The SSL_read() and SSL_write() calls are blocking anyway, so we can
-  * live without select() at this time.
-  */
-
-int     tls_read(SS, fd, buf, count)
-     SmtpState *SS;
-     int fd;
-     void *buf;
-     size_t count;
-{
-    int     i;
-    int     ret;
-    char    mybuf[40];
-    char   *mybuf2;
-
-    if (SS->sslmode) {
-	ret = SSL_read(SS->ssl, buf, count);
-	if (tls_loglevel >= 4) {
-	    mybuf2 = (char *) buf;
-	    if (ret > 0) {
-		i = 0;
-		while ((i < 39) && (i < ret) && (mybuf2[i] != 0)) {
-		    mybuf[i] = mybuf2[i];
-		    i++;
-		}
-		mybuf[i] = '\0';
-		msg_info(SS, "Read %d chars: %s", ret, mybuf);
-	    }
-	}
-	return (ret);
-    } else
-	return (read(fd, buf, count));
-}
-
-int     tls_write(SS, fd, buf, count)
-     SmtpState *SS;
-     int fd;
-     void *buf;
-     size_t count;
-{
-    int     i;
-    char    mybuf[40];
-    char   *mybuf2;
-
-    if (SS->sslmode) {
-      if (tls_loglevel >= 4) {
-	mybuf2 = (char *) buf;
-	if (count > 0) {
-	  i = 0;
-	  while ((i < 39) && (i < count) && (mybuf2[i] != 0)) {
-	    mybuf[i] = mybuf2[i];
-	    i++;
-	  }
-	  mybuf[i] = '\0';
-	  msg_info(SS, "Write %d chars: %s", count, mybuf);
-	}
-      }
-      return (SSL_write(SS->ssl, buf, count));
-    } else
-      return (write(fd, buf, count));
-}
-
 /* skeleton taken from OpenSSL crypto/err/err_prn.c */
 
 static void tls_print_errors(void)
@@ -238,6 +174,7 @@ static int set_cert_stuff(SSL_CTX * ctx, char *cert_file, char *key_file)
     return (1);
 }
 
+
 /* taken from OpenSSL apps/s_cb.c */
 
 static RSA *tmp_rsa_cb(SSL * s, int export, int keylength)
@@ -341,7 +278,7 @@ static void apps_ssl_info_callback(SSL * s, int where, int ret)
 #define TRUNCATE
 #define DUMP_WIDTH	16
 
-static int tls_dump(const char *s, int len)
+int tls_dump(const char *s, int len)
 {
     int     ret = 0;
     char    buf[160 + 1];
@@ -398,7 +335,7 @@ static int tls_dump(const char *s, int len)
     }
 #ifdef TRUNCATE
     if (trunc > 0) {
-	sprintf(buf, "%04x - <SPACES/NULS>\n", len + trunc);
+	sprintf(buf, "%04x - <SPACES/NULS>", len + trunc);
 	msg_info(NULL, "%s", buf);
 	ret += strlen(buf);
     }
@@ -457,7 +394,7 @@ static SSL_SESSION *load_clnt_session(unsigned char *SessionID, int length,
     if (tls_loglevel >= 3)
 	msg_info(NULL, "Trying to reload Session from disc: %s", idstring);
 
-    // FIXME: xxx
+    /* FIXME: xxx */
     buf = (char *)malloc(100 + 2 * uselength + 1);
     mail_queue_path(buf, MAIL_TLS_CLNT_CACHE, idstring);
 
@@ -605,7 +542,7 @@ static void save_clnt_session(SSL_SESSION *session, unsigned char *HostID,
      */
     if ((fd = open(temp, O_WRONLY | O_CREAT | O_EXCL, 0600)) >= 0) {
       if ((fp = fdopen(fd, "w")) == 0) {
-	msg_info(NULL,"%s: could not fdopen %s: %s", myname, temp,
+	msg_info(NULL, "%s: could not fdopen %s: %s", myname, temp,
 		 strerror(errno));
 	return;
       }
@@ -637,7 +574,7 @@ int     tls_init_clientengine(SS, cfgpath)
      SmtpState *SS;
      char *cfgpath;
 {
-  int     off = 0;
+  int     off = 0, linenum = 0;
   int     verify_flags = SSL_VERIFY_NONE;
   char   *CApath;
   char   *CAfile;
@@ -645,47 +582,69 @@ int     tls_init_clientengine(SS, cfgpath)
   char   *c_key_file;
   int	  verifydepth = 0;
   FILE   *fp;
-  char	  buf[1024], *s, *a1, *a2;
+  char	  buf[1024], *n, *a1;
+  unsigned char *s;
+
+  vlog = SS->verboselog;
 
   if (SS->sslmode)
     return (0);				/* already running */
 
   fp = fopen(cfgpath,"r");
   if (!fp) {
-    msg_info(SS, "Can't read TLS config file: '%s'\n",cfgpath);
+    msg_info(SS, "Can't read TLS config file: '%s'",cfgpath);
     return -1;
   }
   while (!feof(fp) && !ferror(fp)) {
     if (!fgets(buf, sizeof(buf), fp))
       break;
-    s = strchr(buf, '\n');
+    ++linenum;
+    s = (void*) strchr(buf, '\n');
     if (s) *s = 0;
-    s = buf;
-    while (*s == ' ' || *s == '\t') ++s;
+    s = (void*) buf;
+
+#define SKIPWHILE(X,Y)  while (*Y != '\0' && isascii(*Y) && X(*Y)) { ++Y; }
+
+    SKIPWHILE(isspace, s);
     if (!*s || *s == '#' || *s == ';')
-      continue;
-    a1 = strtok(s, " \t");
-    a2 = strtok(a1, " \t");
-    if        (strcasecmp(s, "tls-cert-file") == 0 && a1) {
+      continue; /* First non-whitespace char is comment start (or EOL) */
+
+    SKIPWHILE( isspace, s);
+    n = s;
+    SKIPWHILE(!isspace, s);
+    if (*s) *s++ = 0;
+
+    SKIPWHILE( isspace, s);
+    a1 = s;
+    SKIPWHILE(!isspace, s);
+    if (*s) *s++ = 0;
+
+
+    if        (strcasecmp(n, "tls-cert-file") == 0 && a1) {
       tls_cert_file = strdup(a1);
-    } else if (strcasecmp(s, "tls-key-file") == 0  && a1) {
+    } else if (strcasecmp(n, "tls-key-file") == 0  && a1) {
       tls_key_file = strdup(a1);
-    } else if (strcasecmp(s, "tls-CAfile") == 0     && a1) {
+    } else if (strcasecmp(n, "tls-CAfile") == 0     && a1) {
       tls_CAfile = strdup(a1);
-    } else if (strcasecmp(s, "tls-CApath") == 0     && a1) {
+    } else if (strcasecmp(n, "tls-CApath") == 0     && a1) {
       tls_CApath = strdup(a1);
-    } else if (strcasecmp(s, "tls-loglevel") == 0   && a1) {
+    } else if (strcasecmp(n, "tls-loglevel") == 0   && a1) {
       tls_loglevel = atol(a1);
       if (tls_loglevel < 0) tls_loglevel = 0;
       if (tls_loglevel > 4) tls_loglevel = 4;
-    } else if (strcasecmp(s, "tls-use-scache") == 0) {
+    } else if (strcasecmp(n, "tls-use-scache") == 0) {
       tls_use_scache = 1;
-    } else if (strcasecmp(s, "tls-scache-timeout") == 0 && a1) {
+    } else if (strcasecmp(n, "tls-scache-timeout") == 0 && a1) {
       tls_scache_timeout = atol(a1);
       if (tls_loglevel < 0) tls_loglevel = 0;
       if (tls_loglevel > 4) tls_loglevel = 4;
-    } else if (strcasecmp(s, "demand-tls-mode") == 0) {
+    } else if (strcasecmp(n, "demand-tls-mode") == 0) {
       demand_TLS_mode = 1;
+    } else if (strcasecmp(n, "no-tls-readahead") == 0) {
+      tls_use_read_ahead = 0;
+    } else {
+      sfprintf(sfstderr,"# TLS config file, line %d verb: '%s' unknown or missing parameters!\n",
+	       linenum, n);
     }
   }
   fclose(fp);
@@ -764,14 +723,15 @@ int     tls_init_clientengine(SS, cfgpath)
    * another copy of the CApath directory for chroot-jail. On the other
    * hand, the file is not really readable.
    */ 
-  if (strlen(tls_CAfile) == 0)
+  if (!tls_CAfile || strlen(tls_CAfile) == 0)
     CAfile = NULL;
   else
     CAfile = tls_CAfile;
-  if (strlen(tls_CApath) == 0)
+  if (!tls_CApath || strlen(tls_CApath) == 0)
     CApath = NULL;
   else
     CApath = tls_CApath;
+#if 1
   if (CAfile || CApath)
     if ((!SSL_CTX_load_verify_locations(SS->ctx, CAfile, CApath)) ||
 	(!SSL_CTX_set_default_verify_paths(SS->ctx))) {
@@ -779,20 +739,24 @@ int     tls_init_clientengine(SS, cfgpath)
       tls_print_errors();
       return (-1);
     }
+#endif
 
-  if (strlen(tls_cert_file) == 0)
+  if (!tls_cert_file || strlen(tls_cert_file) == 0)
     c_cert_file = NULL;
   else
     c_cert_file = tls_cert_file;
-  if (strlen(tls_key_file) == 0)
+  if (!tls_key_file || strlen(tls_key_file) == 0)
     c_key_file = NULL;
   else
     c_key_file = tls_key_file;
+
   if (c_cert_file || c_key_file)
     if (!set_cert_stuff(SS->ctx, c_cert_file, c_key_file)) {
+#if 0
       msg_info(SS, "TLS engine: cannot load cert/key data");
       tls_print_errors();
       return (-1);
+#endif
     }
 
   /*
@@ -810,8 +774,6 @@ int     tls_init_clientengine(SS, cfgpath)
   verify_depth = verifydepth;
   SSL_CTX_set_verify(SS->ctx, verify_flags, verify_callback);
 
-
-  SS->sslmode = 1;
   return (0);
 }
 
@@ -833,8 +795,11 @@ int     tls_start_clienttls(SS,peername)
     int     verify_result;
     unsigned char *old_session_id;
 
-    if (!SS->sslmode) {		/* should never happen */
+    vlog = SS->verboselog;
+
+    if (!tls_available) {		/* should never happen */
 	msg_info(SS, "tls_engine not running");
+	alarm(0);
 	return (-1);
     }
     if (tls_loglevel >= 1)
@@ -850,6 +815,7 @@ int     tls_start_clienttls(SS,peername)
     else if ((SS->ssl = (SSL *) SSL_new(SS->ctx)) == NULL) {
 	msg_info(SS, "Could not allocate 'con' with SSL_new()");
 	tls_print_errors();
+	alarm(0);
 	return (-1);
     }
     old_session_id = NULL;	/* make sure no old info is kept */
@@ -860,6 +826,7 @@ int     tls_start_clienttls(SS,peername)
     if (!SSL_set_fd(SS->ssl, sffileno(SS->smtpfp))) {
 	msg_info(SS, "SSL_set_fd failed");
 	tls_print_errors();
+	alarm(0);
 	return (-1);
     }
 
@@ -883,14 +850,14 @@ int     tls_start_clienttls(SS,peername)
 	   
 	}
     }
-
+#if 0
     /*
      * Initialize the SSL connection to connect state. This should not be
      * necessary anymore since 0.9.3, but the call is still in the library
      * and maintaining compatibility never hurts.
      */
     SSL_set_connect_state(SS->ssl);
-
+#endif
     /*
      * If the debug level selected is high enough, all of the data is
      * dumped: 3 will dump the SSL negotiation, 4 will dump everything.
@@ -917,21 +884,102 @@ int     tls_start_clienttls(SS,peername)
      * everything that might be there. A session has to be removed anyway,
      * because RFC2246 requires it. 
      */
-    if ((sts = SSL_connect(SS->ssl)) <= 0) {
-	msg_info(SS, "SSL_connect error %d", sts);
-	tls_print_errors();
-	session = SSL_get_session(SS->ssl);
-	if (session) {
+    for (;;) {
+	int sslerr, rc, i;
+	BIO *wbio, *rbio;
+	fd_set rdset, wrset;
+	struct timeval tv;
+
+    ssl_connect_retry:;
+
+	wbio = SSL_get_wbio(SS->ssl);
+	rbio = SSL_get_rbio(SS->ssl);
+
+	sts = SSL_connect(SS->ssl);
+	sslerr = SSL_get_error(SS->ssl, sts);
+
+	switch (sslerr) {
+
+	case SSL_ERROR_WANT_READ:
+	    SS->wantreadwrite = -1;
+	    sslerr = EAGAIN;
+	    break;
+	case SSL_ERROR_WANT_WRITE:
+	    SS->wantreadwrite =  1;
+	    sslerr = EAGAIN;
+	    break;
+
+	case SSL_ERROR_WANT_X509_LOOKUP:
+	    goto ssl_connect_retry;
+	    break;
+
+	case SSL_ERROR_NONE:
+	    goto ssl_connect_done;
+	    break;
+
+	default:
+	    SS->wantreadwrite =  0;
+	    break;
+	}
+
+	if (BIO_should_read(rbio))
+	    SS->wantreadwrite = -1;
+	else if (BIO_should_write(wbio))
+	    SS->wantreadwrite =  1;
+
+	if (! SS->wantreadwrite) {
+	  /* Not proper retry by read or write! */
+
+	ssl_connect_error_bailout:;
+
+	  msg_info(SS, "SSL_connect error %d/%d", sts, sslerr);
+	  tls_print_errors();
+	  session = SSL_get_session(SS->ssl);
+	  if (session) {
 	    remove_clnt_session(session->session_id,
 			        session->session_id_length);
 	    SSL_CTX_remove_session(SS->ctx, session);
 	    msg_info(SS, "SSL session removed");
+	  }
+	  SSL_free(SS->ssl);
+	  SS->ssl = NULL;
+	  alarm(0);
+	  return (-1);
 	}
-	SSL_free(SS->ssl);
-	SS->ssl = NULL;
-	return (-1);
+
+	i = SSL_get_fd(SS->ssl);
+	_Z_FD_ZERO(wrset);
+	_Z_FD_ZERO(rdset);
+
+	if (SS->wantreadwrite < 0)
+	  _Z_FD_SET(i, rdset); /* READ WANTED */
+	else if (SS->wantreadwrite > 0)
+	  _Z_FD_SET(i, wrset); /* WRITE WANTED */
+
+	tv.tv_sec = timeout_tcpw;
+	tv.tv_usec = 0;
+
+	rc = select(i+1, &rdset, &wrset, NULL, &tv);
+	sslerr = errno;
+
+	if (rc == 0) {
+	  /* TIMEOUT! */
+	  sslerr = ETIMEDOUT;
+	  goto ssl_connect_error_bailout;
+	}
+
+	if (rc < 0) {
+	  if (sslerr == EINTR || sslerr == EAGAIN)
+	    continue;
+
+	  /* Bug time ?! */
+	  goto ssl_connect_error_bailout;
+	}
+	/* Default is then success for either read, or write.. */
     }
 
+ ssl_connect_done:;
+
     /*
      * Now we must save the new session to disk, if necessary. If we had
      * an old session, its ID was saved in "old_session_id" for comparison.
@@ -971,12 +1019,13 @@ int     tls_start_clienttls(SS,peername)
      */
     peer = SSL_get_peer_certificate(SS->ssl);
     if (peer != NULL) {
-	X509_NAME_get_text_by_NID(X509_get_subject_name(peer),
-				  NID_commonName, peer_CN, CCERT_BUFSIZ);
-	X509_NAME_get_text_by_NID(X509_get_issuer_name(peer),
-				  NID_commonName, issuer_CN, CCERT_BUFSIZ);
+	X509_NAME_oneline(X509_get_subject_name(peer), tls_peer_cert_name,
+			  CCERT_BUFSIZ);
+	X509_NAME_oneline(X509_get_issuer_name(peer),
+			  tls_peer_cert_issuer_name, CCERT_BUFSIZ);
 	if (tls_loglevel >= 3)
-	    msg_info(SS, "subject_CN=%s, issuer_CN=%s", peer_CN, issuer_CN);
+	    msg_info(SS, "subject=%s, issuer=%s",
+		     tls_peer_cert_name, tls_peer_cert_issuer_name);
 	X509_free(peer);
     }
 
@@ -992,7 +1041,11 @@ int     tls_start_clienttls(SS,peername)
 	     tls_protocol, tls_cipher_name,
 	     tls_cipher_usebits, tls_cipher_algbits);
 
-    SSL_set_read_ahead(SS->ssl, 1); /* Improves performance */
+    if (tls_use_read_ahead)
+      SSL_set_read_ahead(SS->ssl, 1); /* Improves performance */
+
+    /* Mark the mode! */
+    SS->sslmode = 1;
 
     return (0);
 }
@@ -1013,12 +1066,14 @@ int     tls_stop_clienttls(SS, failure)
      SmtpState *SS;
      int failure;
 {
-    SSL_SESSION *session;
+	SSL_SESSION *session;
 
-    if (SS->sslmode) {
-	session = SSL_get_session(SS->ssl);
-	SSL_shutdown(SS->ssl);
-	if (session) {
+	vlog = SS->verboselog;
+
+	if (SS->sslmode) {
+	  session = SSL_get_session(SS->ssl);
+	  SSL_shutdown(SS->ssl);
+	  if (session) {
 	    if (failure) {
 	      remove_clnt_session(peername_md5, MD5_DIGEST_LENGTH);
 	      msg_info(SS, "SSL session removed");
@@ -1026,17 +1081,17 @@ int     tls_stop_clienttls(SS, failure)
 	    SSL_CTX_remove_session(SS->ctx, session);
 	    SSL_free(SS->ssl);
 	    SS->ssl = NULL;
+	  }
+	  SSL_CTX_flush_sessions(SS->ctx,time(NULL));
+
+	  tls_peer_verified = 0;
+	  tls_protocol = NULL;
+	  tls_cipher_name = NULL;
+	  tls_cipher_usebits = 0;
+	  tls_cipher_algbits = 0;
 	}
-	SSL_CTX_flush_sessions(SS->ctx,time(NULL));
 
-	tls_peer_verified = 0;
-	tls_protocol = NULL;
-	tls_cipher_name = NULL;
-	tls_cipher_usebits = 0;
-	tls_cipher_algbits = 0;
-    }
-
-    return (0);
+	return (0);
 }
 #endif /* - HAVE_OPENSSL */
 
@@ -1048,12 +1103,7 @@ int     tls_stop_clienttls(SS, failure)
      - While writing data, a block
        at the time: 3 minutes  (How large a block ?)
      - From "." to "250 OK": 10 minutes
-
-   I think we simplify:  5 minutes each, except "."
-   to "250 ok" which is 60 (!) minutes. (sendmail's
-   default value.)
-   Block-size is 1kB.   4-Feb-95: [mea@utu.fi]
-
+       (We use 60 minutes here - sendmail's default)
  */
 
 extern int timeout;  /* That is the global setting.. */
@@ -1071,126 +1121,176 @@ ssize_t smtp_sfwrite(sfp, vp, len, discp)
      size_t len;
      Sfdisc_t *discp;
 {
-    struct smtpdisc *sd = (struct smtpdisc *)discp;
-    SmtpState *SS = (SmtpState *)sd->SS;
+	struct smtpdisc *sd = (struct smtpdisc *)discp;
+	SmtpState *SS = (SmtpState *)sd->SS;
 
-    const char * p = (const char *)vp;
-    int r, rr, e, i;
-
-    rr = -1; /* No successfull write */
-    e = 0;
-
-#if 1 /* Remove after debug tests */
-    if (SS->verboselog)
-      fprintf(SS->verboselog,
-	      " smtp_sfwrite() to write %d bytes\n", (int)len);
-#endif
-
-    while (len > 0) {
+	const char * p = (const char *)vp;
+	int r, rr, e, i;
 
 #ifdef HAVE_OPENSSL
-      if (SS->sslmode) {
-	r = SSL_write(SS->ssl, p, len);
-	e = errno; /* FIXME: Some SSL function ??? */
-	if (r < 0) {
-	  e = SSL_get_error(SS->ssl, r);
-	  if (e == SSL_ERROR_WANT_WRITE) {
-	    /* Right, so we want to wait a bit, and retry.. */
-	    e = EAGAIN;
-	  } else {
-	    /* XXX: Err... What ??? */
-	    e = ETIMEDOUT; /* not precisely.. */
-	    gotalarm = 1;  /* Well, sort of.. */
-	    break;
-	  }
-	}
-      } else
+	vlog = SS->verboselog;
 #endif /* - HAVE_OPENSSL */
-	{
-	  r = write(sffileno(sfp), p, len);
-	  e = errno;
-	}
 
-      if (r >= 0) {
-	if (rr < 0) rr = 0;	/* something successfull. init this!	*/
-	rr  += r;		/* Accumulate writeout accounting	*/
-	p   += r;		/* move pointer				*/
-	len -= r;		/* count down the length to be written	*/
-	continue;
-      }
+	rr = -1; /* No successfull write */
+	e = errno; /* Whatever the previous one was.. */
 
-      /* Hmm..  Write bounced for some reason */
-      switch (e) {
-      case EAGAIN:
+#if 1 /* Remove after debug tests */
+	if (SS->verboselog)
+	  fprintf(SS->verboselog,
+		  " smtp_sfwrite() to write %d bytes\n", (int)len);
+#endif
+
+	/* Don't even consider writing, if the stream has error status.. */
+	errno = EIO;
+	if (sferror(sfp)) return -1;
+	if (sffileno(sfp) < 0) return -1; /* Write-FD killed! */
+
+	/* If 'len' is zero, return zero.. */
+	/* (I have a feeling such writes are sometimes asked for..) */
+	errno = 0;
+	if (len == 0) return 0;
+
+	errno = e;
+
+	while (len > 0 && !sferror(sfp) && sffileno(sfp) >= 0) {
+
+#ifdef HAVE_OPENSSL
+	  if (SS->sslmode) {
+	    r = SSL_write(SS->ssl, p, len);
+	    e = SSL_get_error(SS->ssl, r);
+	    switch (e) {
+	    case SSL_ERROR_WANT_READ:
+	      SS->wantreadwrite = -1;
+	      e = EAGAIN;
+	      break;
+	    case SSL_ERROR_WANT_WRITE:
+	      SS->wantreadwrite = 1;
+	      e = EAGAIN;
+	      break;
+	    default:
+	      SS->wantreadwrite = 0;
+	      break;
+	    }
+	  } else
+#endif /* - HAVE_OPENSSL */
+	    {
+	      r = write(sffileno(sfp), p, len);
+	      e = errno;
+	    }
+
+	  if (r >= 0) {
+	    if (rr < 0) rr = 0;	/* something successfull. init this!	*/
+	    rr  += r;		/* Accumulate writeout accounting	*/
+	    p   += r;		/* move pointer				*/
+	    len -= r;		/* count down the length to be written	*/
+	    continue;
+	  }
+
+	  /* Hmm..  Write bounced for some reason */
+	  switch (e) {
+	  case EAGAIN:
 #ifdef EWOULDBLOCK
 #if EWOULDBLOCK != EAGAIN
-      case EWOULDBLOCK:
+	  case EWOULDBLOCK:
 #endif
 #endif
-	{
-	  /* Write blocked, lets select (and sleep) for write.. */
-	  struct timeval tv, t0;
-	  fd_set wrset;
+	    {
+	      /* Write blocked, lets select (and sleep) for write.. */
+	      struct timeval tv, t0;
+	      fd_set wrset, rdset;
 
 #if 1 /* Remove after debug tests */
-	  if (SS->verboselog)
-	    gettimeofday(&t0, NULL);
+	      if (SS->verboselog)
+		gettimeofday(&t0, NULL);
 #endif
 
-	  i = sffileno(sfp);
-	  _Z_FD_ZERO(wrset);
-	  _Z_FD_SET(i, wrset);
-	  tv.tv_sec = timeout;
-	  tv.tv_usec = 0;
+	      i = sffileno(sfp);
+	      _Z_FD_ZERO(wrset);
+	      _Z_FD_ZERO(rdset);
 
-	  r = select(i+1, NULL, &wrset, NULL, &tv);
-	  e = errno;
+#ifdef HAVE_OPENSSL
+	      if (SS->sslmode) {
+		if (SS->wantreadwrite < 0)
+		  _Z_FD_SET(i, rdset); /* READ WANTED */
+		else if (SS->wantreadwrite > 0)
+		  _Z_FD_SET(i, wrset); /* WRITE WANTED */
+		else
+		  ; /* FIXME! What???  No reading, nor writing ??? */
+	      } else
+#endif /* - HAVE_OPENSSL */
+		{
+		  _Z_FD_SET(i, wrset);
+		}
+
+	      tv.tv_sec = timeout_tcpw;
+	      tv.tv_usec = 0;
+
+	      r = select(i+1, &rdset, &wrset, NULL, &tv);
+	      e = errno;
 
 #if 1 /* Remove after debug tests */
-	  if (SS->verboselog) {
-	    struct timeval t2;
+	      if (SS->verboselog) {
+		struct timeval t2;
 
-	    gettimeofday(&t2, NULL);
+		gettimeofday(&t2, NULL);
 
-	    t2.tv_usec -= t0.tv_usec;
-	    if (t2.tv_usec < 0) {
-	      t2.tv_usec += 1000000;
-	      t2.tv_sec  -= 1;
+		t2.tv_usec -= t0.tv_usec;
+		if (t2.tv_usec < 0) {
+		  t2.tv_usec += 1000000;
+		  t2.tv_sec  -= 1;
+		}
+		t2.tv_sec -= t0.tv_sec;
+		fprintf(SS->verboselog,
+			" smtp_sfwrite() did select; rc=%d errno=%d len=%d dt=%d.%06d\n",
+			r, e, (int)len, (int)t2.tv_sec, (int)t2.tv_usec);
+	      }
+#endif
+
+	      if (r > 0)
+		/* Ready to write! */
+		break;
+
+	      if (r == 0) {
+		/* TIMEOUT!  Uarrgh!! */
+		gotalarm = 1; 
+		sfp->flags |= SF_ERROR; /* Ensure the error treatment.. */
+
+		/* Actually KILL the outbound stream here! */
+
+		if (sffileno(SS->smtpfp) >= 0) {
+		  /* Error on write stream, write is thus from now on
+		     FORBIDDEN!  We do a write direction shutdown on
+		     the socket, and only listen for replies from now on... */
+		  shutdown(sffileno(SS->smtpfp), 1);
+		  /* Absolutely NO SFIO SYNC AT THIS POINT! */
+		  zsfsetfd(SS->smtpfp, -1);
+		}
+
+		e = ETIMEDOUT;
+
+zsyslog((LOG_ERR,
+	 "ERROR: SMTP socket write timeout; leftover=%d; IP=[%s] mx=%d/%d\n",
+	 len, SS->ipaddress, SS->firstmx, SS->mxcount));
+
+		break;
+	      }
+
+	      /* Khrm... */
+	      /* Select error status is sent out */
 	    }
-	    t2.tv_sec -= t0.tv_sec;
-	    fprintf(SS->verboselog,
-		    " smtp_sfwrite() did select; rc=%d errno=%d len=%d dt=%d.%06d\n",
-		    r, e, (int)len, (int)t2.tv_sec, (int)t2.tv_usec);
-	  }
-#endif
-
-	  if (r > 0)
-	    /* Ready to write! */
-	    continue;
-
-	  if (r == 0) {
-	    /* TIMEOUT!  Uarrgh!! */
-	    rr = -1; /* Zap the write-count! We are to FAIL! */
-	    e = ETIMEDOUT;
+	    break;
+	  default:
+	    /* Any other errno.. */
 	    break;
 	  }
 
-	  /* Khrm... */
-	  /* Select error status is sent out */
-	}
-	break;
-      default:
-	/* Any other errno.. */
-	break;
-      }
+	  /* If STILL a error, break out -- will retry EINTR at  Sfio  library. */
+	  if (r < 0) break;
 
-      /* If STILL a error, break out -- will retry EINTR at  Sfio  library. */
-      if (r < 0) break;
+	} /* End of while(len > 0) loop */
 
-    } /* End of while(len > 0) loop */
-
-  errno = e;
-  return rr;
+	errno = e;
+	return rr;
 }
 
 /*
@@ -1202,23 +1302,49 @@ int smtp_nbread(SS, buf, spc)
      void *buf;
      int spc;
 {
-  int r, e;
-  int infd = sffileno(SS->smtpfp);
-
+	int r, e;
+	int infd = SS->smtpfd;
 
 #ifdef HAVE_OPENSSL
-  if (SS->sslmode) {
-    r = SSL_read(SS->ssl, buf, spc);
-    e = SSL_get_error(SS->ssl, r);
-    if (e == SSL_ERROR_WANT_READ) {
-      e = EAGAIN;
-    } else
-      e = EINTR;
-    errno = e;
-  } else
+	vlog = SS->verboselog;
+
+	if (SS->sslmode) {
+	  r = SSL_read(SS->ssl, buf, spc);
+	  e = SSL_get_error(SS->ssl, r);
+	  switch (e) {
+	  case SSL_ERROR_WANT_READ:
+	    SS->wantreadwrite = -1;
+	    e = EAGAIN;
+	    break;
+	  case SSL_ERROR_WANT_WRITE:
+	    SS->wantreadwrite =  1;
+	    e = EAGAIN;
+	    break;
+	  default:
+	    SS->wantreadwrite =  0;
+	    break;
+	  }
+	  if (tls_loglevel >= 3) {
+	    msg_info(NULL,"smtp_nbread() rc=%d errno=%d", r, e);
+	    if (r > 0) tls_dump(buf, r);
+	  }
+	} else
 #endif /* - HAVE_OPENSSL */
-    /* Normal read(2) */
-    r = read(infd, buf, spc);
+	  {
+	    /* Normal read(2) */
+	    r = read(infd, buf, spc);
+	    e = errno;
+	  }
   
-  return r;
+	errno = e;
+	return r;
+}
+
+int zsfsetfd(fp, fd)
+     Sfio_t *fp;
+     int fd;
+{
+  /* This is *NOT* the SFIO's sfsetfd() -- we do no sfsync() at any point.. */
+  fp->file = fd;
+  return fd;
 }
